/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* absolute to relative conversion */

struct mxk_dim{
  int d_abs;
  int d_rel;
  int d_max;
  int d_min;
  int d_current;

  int d_big;
  int d_small;
  int d_base;
};

struct mxk_ar{
  int a_size;

  struct mxk_dim *a_dims;

  int a_brake;
  int a_reset;

  int a_arm;
  int a_pass;

  int a_pit; /* bottom fraction at which no scaling happens */
  int a_top; /* top fraction at which redline scale happens */

  int a_scale; /* mouse scale factor */
  int a_redline; /* max scale factor at edge */

  int a_slow; /* longest move interval */
  int a_fast; /* shortest move interval */

  struct mxk_client *a_vector;
  int a_count;
};

struct mxk_ar *mxk_create_ar()
{
  struct mxk_ar *ma;

  ma = malloc(sizeof(struct mxk_ar));
  if(ma == NULL){
    return NULL;
  }

  ma->a_dims = NULL;
  ma->a_size = 0;

  ma->a_reset = -1;
  ma->a_brake = -1;

  ma->a_pass = 1;
  ma->a_arm = 0;

  ma->a_vector = NULL;
  ma->a_count = 0;

  ma->a_pit = 4;
  ma->a_top = 5;
  ma->a_scale = 14;
  ma->a_redline = 64;

  ma->a_slow = 80000;
  ma->a_fast = 20000;

  return ma;
}

void mxk_destroy_ar(struct mxk_ar *ma)
{
  struct mxk_dim *md;
  int i;

  if(ma == NULL){
    return;
  }

  for(i = 0; i < ma->a_size; i++){
    md = &(ma->a_dims[i]);
    /* do something */
  }

  if(ma->a_dims){
    free(ma->a_dims);
    ma->a_dims = NULL;
  }

  mxk_shutdown_client(ma->a_vector, ma->a_count);
#ifdef DEBUG
  ma->a_vector = NULL;
  ma->a_count = 0;
#endif

  free(ma);
}

int mxk_axis_ar(struct mxk_ar *ma, int index, char *name)
{
  struct mxk_dim *md;
  char *abs, *rel;
  int a, r, i;

  abs = name;
  rel = strchr(name, ':');
  if(rel){
    *rel = '\0';
    rel++;
  }

  a = name_to_code(mxk_abs_codes, abs);
  if(a < 0){
    return -1;
  }
  r = name_to_code(mxk_rel_codes, rel ? rel : abs);
  if(r < 0){
    return -1;
  }

  if(index >= ma->a_size){
    md = realloc(ma->a_dims, sizeof(struct mxk_dim) * (index + 1));
    if(md == NULL){
      return -1;
    }

    ma->a_dims = md;

    for(i = ma->a_size; i <= index; i++){
      md = &(ma->a_dims[i]);

      md->d_abs = -1;
      md->d_rel = -1;

      md->d_max = INT_MIN;
      md->d_min = INT_MAX;

      md->d_big = INT_MIN;
      md->d_small = INT_MAX;

      md->d_current = 0;
      md->d_base = 0;
    }

    ma->a_size = index + 1;
  }

  md = &(ma->a_dims[index]);

  md->d_abs = a;
  md->d_rel = r;

  return 0;
}

/**********************************************************************/

int mxk_pass_ar(struct list_loop *ls, struct task_loop *tx, int pass, char *args)
{
  struct mxk_ar *ma;

#ifdef DEBUG
  fprintf(stderr,  "pass: setting bypass mode to %d\n", pass);
#endif

  ma = user_loop(ls, tx);
  ma->a_pass = pass;

  return 0;
}

int mxk_brake_ar(struct list_loop *ls, struct task_loop *tx, int valid, char *args)
{
  int idx;
  struct mxk_ar *ma;
  char buffer[MXK_NAME_SIZE];
  int key;

  if(valid){

    idx = start_word(args, 0);
    if(idx < 0){
      return -1;
    }

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key < 0){
      return -1;
    }
  } else {
    key = (-1);
  }

  ma = user_loop(ls, tx);
  ma->a_brake = key;

  return 0;
}

int mxk_split_ar(struct list_loop *ls, struct task_loop *tx, int flag, char *args)
{
  int idx;
  struct mxk_ar *ma;
  char *end;
  int top, pit;

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  pit = strtoul(args + idx, &end, 10);
  if((pit <= 0) || (*end != ' ')){
#ifdef DEBUG
    fprintf(stderr, "absrel: invalid pit: %d, end=<%s>\n", pit, end);
#endif
    return -1;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    return -1;
  }

  top = strtoul(args + idx, &end, 10);
  if((top <= 0) || (*end != ' ')){
#ifdef DEBUG
    fprintf(stderr, "absrel: invalid top: %d, end=<%s>\n", top, end);
#endif
    return -1;
  }

  ma = user_loop(ls, tx);

  ma->a_pit = pit;
  ma->a_top = top;

  return 0;
}

/**********************************************************************/

static int mxk_run_ar(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_dim *md;
  struct mxk_ar *ma;

  struct input_event *vp;
  struct link_loop *nk;

  int i, wake, relay, delta, direction, value, range, scale, period, test;
  struct timeval interval;

  mg = global_loop(ls);
  ma = user_loop(ls, tl);

  if(ma == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_ar(ma);
    return 0;
  }

  relay = 0;

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_ABS :
            for(i = 0; i < ma->a_size; i++){
              if(ma->a_dims[i].d_abs == vp->code){
                md = &(ma->a_dims[i]);

                value = vp->value;
                if(value < 0){
                  if((value + md->d_base) < 0){
                    md->d_base = vp->value * (-1);
                    value = 0;
                  } else {
                    value += md->d_base;
                  }
                }

                if(md->d_min > value){
                  md->d_min = value;
                }
                if(md->d_max < value){
                  md->d_max = value;
                }
                md->d_current = value;
#ifdef DEBUG
                fprintf(stderr, "absrel[%d]: motion: (%d,%d) <%d,%d> %d\n", i, md->d_min, md->d_max, md->d_small, md->d_big, md->d_current);
#endif
                i = ma->a_size;
                ma->a_arm = 1;
              }
            }
            break;
          case EV_SYN :
            if(ma->a_arm == 0){
              if(ma->a_pass){
                mxk_send_client(ls, tl, ma->a_vector, ma->a_count, vp->type, vp->code, vp->value);
              }
              break;
            }
            ma->a_arm = 0;
            relay = 1;
            break;
          case EV_KEY :
            if((ma->a_brake == vp->code) && (vp->value == 1)){
#ifdef DEBUG
              fprintf(stderr, "absrel: got brake 0x%x\n", ma->a_brake);
#endif

              for(i = 0; i < ma->a_size; i++){
                md = &(ma->a_dims[i]);
                if(md->d_current < md->d_small){
                  md->d_small = md->d_current;
                }
                if(md->d_current > md->d_big){
                  md->d_big = md->d_current;
                }
              }
              break;
            }
            /* WARNING: partial fall */
          default :
            if(ma->a_pass){
              mxk_send_client(ls, tl, ma->a_vector, ma->a_count, vp->type, vp->code, vp->value);
            }
            break;
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  if(mask & LO_TIME_MASK){
    relay = 1;
  }

  if(relay){
    wake = 0;
    period = ma->a_slow;
    for(i = 0; i < ma->a_size; i++){
      md = &(ma->a_dims[i]);
      if(md->d_big >= md->d_small){ /* box is valid */
#ifdef DEBUG
        fprintf(stderr, "absrel[%d]: %d<=%d (%d) %d<=%d\n", i, md->d_min, md->d_small, md->d_current, md->d_big, md->d_max);
#endif
        direction = 0;
        range = 0;
        delta = 0;
        if(md->d_current < md->d_small){
          range = md->d_small - md->d_min;
          delta = md->d_small - md->d_current;
          direction = -1;
        }
        if(md->d_current > md->d_big){
          range = md->d_max - md->d_big;
          delta = md->d_current - md->d_big;
          direction = 1;
        }
        if(direction){
#ifdef DEBUG
          if(range <= 0){
            fprintf(stderr, "absrel: major logic problem: range %d too small\n", range);
            abort();
          }
#endif
          wake = 1;
          scale = 1;
          if(((range - delta) * ma->a_top) < range){
#ifdef DEBUG
            fprintf(stderr, "absrel: redlining %d\n", ma->a_redline);
#endif
            scale = ma->a_redline;
          } else if ((delta * ma->a_pit) > range){
            scale = 1 + (((delta - 1) * ma->a_scale) / range);
#ifdef DEBUG
            fprintf(stderr, "absrel: enabling scale=%d\n", scale);
            if(scale <= 0){
              fprintf(stderr, "major logic problem: %d too small (delta=%d, range=%d\n", scale, delta, range);
              abort();
            }
#endif
#if 0
            direction *= ((ma->a_scale * limit) - range) / ((ma->a_pit - 1) * range);
#endif
          }
          direction *= scale;

          test = ma->a_fast + (((ma->a_slow - ma->a_fast) * (range - delta)) / range);
          if(test < period){
            period = test;
#ifdef DEBUG
            fprintf(stderr, "new period: %dus\n", period);
#endif
          }

#ifdef DEBUG
          fprintf(stderr, "absrel[%d]: active: (%d) %d/%d\n", i, direction, delta, range);
#endif
          mxk_send_client(ls, tl, ma->a_vector, ma->a_count, EV_REL, md->d_rel, direction);
        }
      } else {
#ifdef DEBUG
        fprintf(stderr, "absrel: box out of range\n");
#endif
      }
    }
    if(wake){
      mxk_send_client(ls, tl, ma->a_vector, ma->a_count, EV_SYN, 0, 0);
      interval.tv_sec = 0;
      interval.tv_usec = ma->a_slow;
      set_time_loop(ls, tl, &interval);
      set_mask_loop(ls, tl, LO_TIME_MASK, 1);
    } else {
      set_mask_loop(ls, tl, LO_TIME_MASK, -1);
    }
  }

  return 0;
}

/*********************************************************************/

int mxk_pass_ar_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "absrel", MXK_MISC_OPTIONAL, 1, mxk_pass_ar);

  return 0;
}

int mxk_ignore_ar_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "absrel", MXK_MISC_OPTIONAL, 0, mxk_pass_ar);

  return 0;
}

int mxk_split_ar_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "absrel", MXK_MISC_REQUIRED, 0, mxk_split_ar);

  return 0;
}

int mxk_brake_ar_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "absrel", MXK_MISC_OPTIONAL, 1, mxk_brake_ar);

  return 0;
}

int mxk_start_ar_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  char buffer[MXK_NAME_SIZE];
  struct control_loop *cl;
  struct mxk_ar *ma;
  struct mxk_global *mg;
  struct task_loop *tx;
  char *end, *value;
  int cd, idx, id, type, axis;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "absrel");
  if(type <= 0){
    print_control(cl, cd, 1, "absrel: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "absrel: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "absrel: unable to acquire first parameter\n");
    return 0;
  }

  value = args + idx;
  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "absrel: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  ma = mxk_create_ar();
  if(ma == NULL){
    print_control(cl, cd, 1, "absrel: unable to allocate state\n");
    return 0;
  }

  axis = 0;
  while((idx = next_word(args, idx)) > 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      print_control(cl, cd, 1, "absrel: missing or malformed axis pair %d\n", axis);
      mxk_destroy_ar(ma);
      return 0;
    }

    if(mxk_axis_ar(ma, axis, buffer)){
      print_control(cl, cd, 1, "absrel: unable to add absolute axis pair <%s>\n", buffer);
      mxk_destroy_ar(ma);
      return 0;
    }
    axis++;
  }

  if(axis <= 0){
    print_control(cl, cd, 1, "absrel: need at least one axis\n");
    mxk_destroy_ar(ma);
    return 0;
  }

  tx = add_tag_loop(ls, -1, ma, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_ar, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "absrel: unable to create task\n");
    mxk_destroy_ar(ma);
    return 0;
  }

  text_loop(ls, tx, "ar:%d axes=%d", id, axis);
  print_control(cl, cd, 1, "absrel: created %d axis task\n", axis);

  return 0;
}

int mxk_list_ar_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_ar *ma;
  struct mxk_dim *md;
  char *abs, *rel;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "absrel");
  if(type <= 0){
    print_control(cl, cd, 1, "absrel: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    ma = user_loop(ls, tx);

    if(ma){

      print_control(cl, cd, 0, "absrel[%d]: unknown: %s, dimensions: %d\n", tag_loop(ls, tx), ma->a_pass ? "bypass" : "ignore", ma->a_size);
      print_control(cl, cd, 0, "absrel[%d]: low=>1 (1/%d,%d/%d)*>%d high=>%d\n", tag_loop(ls, tx), ma->a_pit, ma->a_top - 1, ma->a_top, ma->a_scale, ma->a_redline);
      for(i = 0; i < ma->a_size; i++){
        md = &(ma->a_dims[i]);
        abs = code_to_name(mxk_abs_codes, md->d_abs);
        rel = code_to_name(mxk_rel_codes, md->d_rel);
        if(md->d_big < md->d_small){
          print_control(cl, cd, 0, "axis[%d]: %s->%s uninitialised\n", i, abs ? abs : "undefined", rel ? rel : "undefined");
        } else {
          print_control(cl, cd, 0, "axis[%d]: %s->%s range=(%d,%d) rest=(%d,%d) position=%d\n", i, abs ? abs : "undefined", rel ? rel : "undefined", md->d_min, md->d_max, md->d_small, md->d_big, md->d_current);
        }
      }
      count++;
    } else {
      print_control(cl, cd, 1, "absrel[%d]: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}
/**********************************************************************/

static int mxk_insert_ar(void *absrel, int type, int id)
{
  struct mxk_ar *ma;
  struct mxk_client *mc;

  ma = absrel;

  mc = mxk_insert_client(ma->a_vector, ma->a_count, type, id);
  if(mc == NULL){
    return -1;
  }

  ma->a_vector = mc;
  ma->a_count++;

  return 0;
}

int mxk_startup_ar(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "absrel", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_ar);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_ar(struct mxk_global *mg)
{
}
