/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* map absolute motion (vector) space to keypresses */

struct mxk_point {
  int p_key;
  int *p_axis;
#if 0
  int p_toggle;
#endif
};

struct mxk_axis{
  int a_max;
  int a_min;

  int a_code;

  /* could be scaling factors in here */

  int a_current;
  int a_point;

  int *a_index;
};

struct mxk_vector{
  int v_dim;
  int v_size;

  struct mxk_point *v_points;
  struct mxk_axis *v_axes;

  int v_key;
  int v_zero;
  int v_arm;
  int v_pass;
#if 0
  int v_toggle;
#endif

  struct mxk_client *v_vector;
  int v_count;
};

/***********************************************************************/

void mxk_destroy_ax(struct mxk_vector *mv)
{
  struct mxk_point *mp;
  struct mxk_axis *ma;
  int i;

  if(mv == NULL){
    return;
  }

  mxk_shutdown_client(mv->v_vector, mv->v_count);
#ifdef DEBUG
  mv->v_vector = NULL;
  mv->v_count = 0;
#endif

  if(mv->v_points){
    for(i = 0; i < mv->v_size; i++){
      mp = &(mv->v_points[i]);
      if(mp->p_axis){
        free(mp->p_axis);
        mp->p_axis = NULL;
      }
    }
    free(mv->v_points);
    mv->v_points = NULL;
  }

  if(mv->v_axes){
    for(i = 0; i < mv->v_dim; i++){
      ma = &(mv->v_axes[i]);

      if(ma->a_index){
        free(ma->a_index);
        ma->a_index = NULL;
      }
    }
    free(mv->v_axes);
    mv->v_axes = NULL;
  }

  free(mv);
}

struct mxk_vector *mxk_create_ax()
{
  struct mxk_vector *mv;

  mv = malloc(sizeof(struct mxk_vector));
  if(mv == NULL){
    return NULL;
  }

  mv->v_dim = 0;
  mv->v_size = 0;

  mv->v_key = -1; /* key reserved ? */
  mv->v_zero = 0;
  mv->v_arm = 0;
  mv->v_pass = 1;

  mv->v_axes = NULL;
  mv->v_points = NULL;

  mv->v_vector = NULL;
  mv->v_count = 0;

  return mv;
}

int mxk_axis_ax(struct mxk_vector *mv, int index, char *name)
{
  struct mxk_axis *ma;
  int code, i;

#if 1
  if(mv->v_size){
    return -1;
  }
#endif

  if(index >= mv->v_dim){
    ma = realloc(mv->v_axes, sizeof(struct mxk_axis) * (index + 1));
    if(ma == NULL){
      return -1;
    }

    mv->v_axes = ma;

    for(i = mv->v_dim; i <= index; i++){
      ma = &(mv->v_axes[i]);
      ma->a_max = INT_MIN;
      ma->a_min = INT_MAX;

      ma->a_code = -1;

      ma->a_current = 0;
      ma->a_index = NULL;

      /* TODO: populate a_index allowing runtime adjustments */
    }

    mv->v_dim = index + 1;
  }

  code = name_to_code(mxk_abs_codes, name);
  if(code < 0){
    return -1;
  }

  ma = &(mv->v_axes[index]);
  ma->a_code = code;

  return 0;
}

int mxk_place_point_ax(struct mxk_vector *mv, int index, int point)
{
  struct mxk_axis *ma;

  if(index >= mv->v_dim){
    return -1;
  }

  ma = &(mv->v_axes[index]);
  if(point > ma->a_max){
    ma->a_max = point;
  }

  if(point < ma->a_min){
    ma->a_min = point;
  }

  ma->a_point = point;

  return 0;
}

int mxk_make_point_ax(struct mxk_vector *mv, char *name)
{
  struct mxk_point *tp;
  struct mxk_axis *ma;
  int key, i, j;
  int *pp;

#ifdef DEBUG
  fprintf(stderr, "make point: creating key <%s>\n", name);
#endif

  if(strcmp(name, "none")){
    key = name_to_code(mxk_key_codes, name);
    if(key < 0){
      return -1;
    }
  } else {
    key = -1;
  }

  tp = realloc(mv->v_points, sizeof(struct mxk_point) * (mv->v_size + 1));
  if(tp == NULL){
    return -1;
  }

  mv->v_points = tp;

  tp = &(mv->v_points[mv->v_size]);
  tp->p_key = key;
#if 0
  tp->p_toggle = mv->v_toggle;
#endif
  tp->p_axis = malloc(sizeof(int) * mv->v_dim);
#ifdef DEBUG
  fprintf(stderr, "point: allocated p_axis: %p (%d bytes)\n", tp->p_axis, mv->v_dim * sizeof(int));
#endif
  if(tp->p_axis == NULL){
    return -1;
  }

  for(i = 0; i < mv->v_dim; i++){
    ma = &(mv->v_axes[i]);
    tp->p_axis[i] = ma->a_point;

    pp = realloc(ma->a_index, sizeof(int) * (mv->v_size + 1));
    if(pp == NULL){
      return -1;
    }
#ifdef DEBUG
    fprintf(stderr, "resized index to %d\n", mv->v_size + 1);
#endif

    ma->a_index = pp;

    for(j = mv->v_size; (j > 0) && (mv->v_points[pp[j - 1]].p_axis[i] > tp->p_axis[i]); j--){
#ifdef DEBUG
      fprintf(stderr, "moving [%d]->[%d]\n", j - 1, j);
#endif
      pp[j] = pp[j - 1];
    }

#ifdef DEBUG
    fprintf(stderr, "insert at [%d]\n", j);
#endif
    pp[j] = mv->v_size;
  }

  mv->v_size++;

  return 0;
}

int mxk_zero_return_ax(struct list_loop *ls, struct task_loop *tx, int zero, char *args)
{
  struct mxk_vector *mv;

#ifdef DEBUG
  fprintf(stderr,  "zero: setting return to zero to %d\n", zero);
#endif

  mv = user_loop(ls, tx);
  mv->v_zero = zero;

  return 0;
}

int mxk_pass_ax(struct list_loop *ls, struct task_loop *tx, int pass, char *args)
{
  struct mxk_vector *mv;

#ifdef DEBUG
  fprintf(stderr,  "pass: setting bypass mode to %d\n", pass);
#endif

  mv = user_loop(ls, tx);
  mv->v_pass = pass;

  return 0;
}


/************************************************************************/

static int mxk_run_ax(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_vector *mv;
  struct mxk_global *mg;
  struct mxk_axis *ma;
  struct mxk_point *mp;
  struct input_event *vp;
  struct link_loop *nk;
  int i, j, k;
  unsigned long best, sum, delta, mul;
  int key, pos;

  mg = global_loop(ls);
  mv = user_loop(ls, tl);

  if(mv == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_ax(mv);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_ABS :
            for(i = 0; i < mv->v_dim; i++){
              if(mv->v_axes[i].a_code == vp->code){
                ma = &(mv->v_axes[i]);
                if(ma->a_min > vp->value){
                  ma->a_min = vp->value;
                }
                if(ma->a_max < vp->value){
                  ma->a_max = vp->value;
                }
                ma->a_current = vp->value;
                i = mv->v_dim;
                mv->v_arm = 1;
              }
            }
            break;
          case EV_SYN :
            if(mv->v_arm == 0){
              if(mv->v_pass){
                mxk_send_client(ls, tl, mv->v_vector, mv->v_count, vp->type, vp->code, vp->value);
              }
              break;
            }
            mv->v_arm = 0;
            best = UINT_MAX;
            key = (-1);
            /* TODO: some really clever optimisation, prolly not worth it */
            for(i = 0; i < mv->v_size; i++){
              sum = 0;
              mp = &(mv->v_points[i]);
              for(j = 0; j < mv->v_dim; j++){
                mul = 0;
                if(mv->v_axes[j].a_max > mv->v_axes[j].a_min){
                  pos = mv->v_axes[j].a_current;
                  delta = (pos > mp->p_axis[j]) ? (pos - mp->p_axis[j]) : (mp->p_axis[j] - pos);
                  mul = delta;
                  for(k = 1; k < mv->v_dim; k++){
                    mul *= delta;
                  }
#ifdef DEBUG
fprintf(stderr, "ax: key[%d], dim[%d]: (key=%d, cur=%d) delta=%lu, mul=%lu\n", i, j, mp->p_axis[j], pos, delta, mul);
#endif
                }
                sum += mul;
              }
              if(sum < best){
                key = mp->p_key;
#ifdef DEBUG
                fprintf(stderr, "ax: updating old best (was %lu), key now %d\n", best, key);
#endif
                best = sum;
              }
            }
            if((mv->v_zero && ((mv->v_key < 0) || (key < 0))) || ((mv->v_zero == 0) && (key != mv->v_key))){
#ifdef DEBUG
              fprintf(stderr, "ax: change: updating (old=%d, new=%d)\n", mv->v_key, key);
#endif
              if(mv->v_key >= 0){ /* release old */
                mxk_send_client(ls, tl, mv->v_vector, mv->v_count, EV_KEY, mv->v_key, 0);
              }
              mv->v_key = key;
              if(mv->v_key >= 0){ /* press new */
                mxk_send_client(ls, tl, mv->v_vector, mv->v_count, EV_KEY, mv->v_key, 1);
              }
              mxk_send_client(ls, tl, mv->v_vector, mv->v_count, EV_SYN, 0, 0);
            }
            break;
          default :
            if(mv->v_pass){
              mxk_send_client(ls, tl, mv->v_vector, mv->v_count, vp->type, vp->code, vp->value);
            }
            break;
        }
#if 0
        mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, key, 1);
#endif
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;

}
/***********************************************************************/

int mxk_start_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  char buffer[MXK_NAME_SIZE];
  struct control_loop *cl;
  struct mxk_vector *mv;
  struct mxk_global *mg;
  struct task_loop *tx;
  char *end, *value;
  int cd, idx, id, type, axis;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "vector");
  if(type <= 0){
    print_control(cl, cd, 1, "vector: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "vector: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "vector: unable to acquire first parameter\n");
    return 0;
  }

  value = args + idx;
  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "vector: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  mv = mxk_create_ax();
  if(mv == NULL){
    print_control(cl, cd, 1, "vector: unable to allocate state\n");
    return 0;
  }

  axis = 0;
  while((idx = next_word(args, idx)) > 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      print_control(cl, cd, 1, "vector: missing or malformed axis number %d\n", axis);
      mxk_destroy_ax(mv);
      return 0;
    }

    if(mxk_axis_ax(mv, axis, buffer)){
      print_control(cl, cd, 1, "vector: unable to add absolute axis <%s>\n", buffer);
      mxk_destroy_ax(mv);
      return 0;
    }
    axis++;
  }

  if(axis <= 0){
    print_control(cl, cd, 1, "vector: need at least one axis\n");
    mxk_destroy_ax(mv);
    return 0;
  }

  tx = add_tag_loop(ls, -1, mv, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_ax, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "vector: unable to create task\n");
    mxk_destroy_ax(mv);
    return 0;
  }

  text_loop(ls, tx, "vector:%d axes=%d", id, axis);
  print_control(cl, cd, 1, "vector: created %d axis task\n", axis);

  return 0;
}

int mxk_point_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  char buffer[MXK_NAME_SIZE];
  struct control_loop *cl;
  struct mxk_vector *mv;
  struct mxk_global *mg;
  struct task_loop *tx;
  char *end, *value;
  int cd, idx, id, type, pos, pdx, index, count;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "vector");
  if(type <= 0){
    print_control(cl, cd, 1, "vector: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "vector: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "vector: unable to acquire first parameter\n");
    return 0;
  }

  value = args + idx;
  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "vector: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    print_control(cl, cd, 1, "vector: unable to acquire key name\n");
    return 0;
  }

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, cd, 1, "vector: missing or malformed key name\n");
    return 0;
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    mv = user_loop(ls, tx);
    if(mv){
      pdx = idx;
      index = 0;
      while((pdx = next_word(args, pdx)) > 0){
        value = args + pdx;
        pos = strtol(value, &end, 10);
        if((*end != ' ') && (*end != '\0')){
          print_control(cl, cd, 1, "vector: unable to extract a position for dimension %d\n", index);
          return 0;
        }

        if(mxk_place_point_ax(mv, index, pos)){
          print_control(cl, cd, 1, "vector: unable to set position %d for dimension %d\n", pos, index);
          return 0;
        }

        index++;
      }

      if(index < mv->v_dim){
        print_control(cl, cd, 1, "vector: insufficient dimensions, expected %d, got %d\n", index, mv->v_dim);
        return 0;
      }

      if(mxk_make_point_ax(mv, buffer)){
        print_control(cl, cd, 1, "vector: unable register point for <%s>\n", buffer);
        return 0;
      }

      count++;
    } else {
      print_control(cl, cd, 1, "vector: unable to access vector number %d\n", id);
    }
  }

  if(count > 0){
    print_control(cl, cd, 1, "vector: registered %d %s\n", count, (count > 0) ? "points" : "point");
  } else {
    print_control(cl, cd, 1, "vector: failed to register point for id %d\n", id);
  }

  return 0;
}

int mxk_return_zero_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "vector", MXK_MISC_OPTIONAL, 1, mxk_zero_return_ax);

  return 0;
}

int mxk_non_zero_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "vector", MXK_MISC_OPTIONAL, 0, mxk_zero_return_ax);

  return 0;
}

int mxk_pass_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "vector", MXK_MISC_OPTIONAL, 1, mxk_pass_ax);

  return 0;
}

int mxk_ignore_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "vector", MXK_MISC_OPTIONAL, 0, mxk_pass_ax);

  return 0;
}

int mxk_list_ax_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, j, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_vector *mv;
  struct mxk_axis *ma;
  struct mxk_point *mp;
  char *axis, *key, *value;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "vector");
  if(type <= 0){
    print_control(cl, cd, 1, "vector: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    mv = user_loop(ls, tx);

    if(mv){

      print_control(cl, cd, 0, "vector[%d]: dimensions: %d, points: %d, return-to-zero: %s, pass: %s\n", tag_loop(ls, tx), mv->v_dim, mv->v_size, mv->v_zero ? "yes" : "no", mv->v_pass ? "yes" : "no");
      for(i = 0; i < mv->v_dim; i++){
        ma = &(mv->v_axes[i]);
        axis = code_to_name(mxk_abs_codes, ma->a_code);
        if(ma->a_min < ma->a_max){
          print_control(cl, cd, 0, "axis[%d]: %s [%d<=%d<=%d]\n", i, axis ? axis : "<undefined>", ma->a_min, ma->a_current, ma->a_max);
        } else {
          print_control(cl, cd, 0, "axis[%d]: %s [uninitialised]\n", i, axis ? axis : "<undefined>");
        }
      }
      for(i = 0; i < mv->v_size; i++){
        mp = &(mv->v_points[i]);
        key = code_to_name(mxk_key_codes, mp->p_key);
        print_control(cl, cd, 0, "point[%d]: %s <%d", i, key ? key : "<none>", mp->p_axis[0]);
        for(j = 1; j < mv->v_dim; j++){
          print_control(cl, cd, 0, ",%d", mp->p_axis[j]);
        }
        print_control(cl, cd, 0, ">\n");
      }
      count++;
    } else {
      print_control(cl, cd, 1, "vector[%d]: unavailable\n", id);
    }
  }
  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

/***********************************************************************/

static int mxk_insert_ax(void *vector, int type, int id)
{
  struct mxk_vector *mv;
  struct mxk_client *mc;

  mv = vector;

  mc = mxk_insert_client(mv->v_vector, mv->v_count, type, id);
  if(mc == NULL){
    return -1;
  }

  mv->v_vector = mc;
  mv->v_count++;

  return 0;
}

int mxk_startup_ax(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "vector", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_ax);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_ax(struct mxk_global *mg)
{
}
