/*
 * (c) Marc Welz 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#ifndef CBU_C_
#define CBU_C_

#ifdef DEBUG
#include <stdio.h>
#endif
#include <stdlib.h>

#include <cbu.h>

#ifndef CBU_ANCHOR
#error require CBU_ANCHOR, the name of the cbu_anchor structure inside CBU_LOAD
#endif

#ifndef CBU_LOAD
#error require CBU_LOAD, the name of the datum structure 
#endif

void destroy_cbu(struct cbu_state *cs)
{
  struct CBU_LOAD *prev, *next, *cl;

  if(cs == NULL){
    return;
  }

  while(cs->s_head){
    cl = cs->s_head;

    prev = cl->CBU_ANCHOR.a_prev;
    next = cl->CBU_ANCHOR.a_next;

    prev->CBU_ANCHOR.a_next = next;
    next->CBU_ANCHOR.a_prev = prev;

    cl->CBU_ANCHOR.a_prev = NULL;
    cl->CBU_ANCHOR.a_next = NULL;

    free(cl);
    
    if(prev == cl){
      cs->s_head = NULL;
    } else {
      cs->s_head = prev;
    }
  }

  cs->s_tail = NULL;
  cs->s_size = 0;
  cs->s_used = 0;

  free(cs);
}

struct cbu_state *create_cbu(int size)
{
  struct cbu_state *cs;
  struct CBU_LOAD *cl, *prev, *next;
  int i;

  cs = malloc(sizeof(struct cbu_state));
  if(cs == NULL){
    return NULL;
  }

  cs->s_size = 0;
  cs->s_used = 0;

  cs->s_head = NULL;
  cs->s_tail = NULL;

  for(i = 0; i < size; i++){
    cl = malloc(sizeof(struct CBU_LOAD));
    if(cl == NULL){
      destroy_cbu(cs);
      return NULL;
    }

    cl->CBU_ANCHOR.a_valid = 0;

    if(i){
      next = cs->s_head;
      prev = cs->s_head->CBU_ANCHOR.a_prev;

      cl->CBU_ANCHOR.a_next = next;
      cl->CBU_ANCHOR.a_prev = prev;

      next->CBU_ANCHOR.a_prev = cl;
      prev->CBU_ANCHOR.a_next = cl;
    } else {
      cl->CBU_ANCHOR.a_next = cl;
      cl->CBU_ANCHOR.a_prev = cl;

      cs->s_head = cl;
      cs->s_tail = cl;
    }
  }

  cs->s_size = size;

  return cs;
}

struct CBU_LOAD *push_cbu(struct cbu_state *cs, int clobber)
{
  struct CBU_LOAD *cl, *head, *tail;

  head = cs->s_head;

  cl = head->CBU_ANCHOR.a_next;

  if(cl == cs->s_tail){
    if(clobber){
      tail = cl->CBU_ANCHOR.a_next;
      tail->CBU_ANCHOR.a_valid = 0;
      cs->s_tail = tail;
    } else {
      return NULL;
    }
  } else {
    cs->s_used++;
  }

  cl->CBU_ANCHOR.a_valid = 1;
  cs->s_head = cl;

  return cl;
}

struct CBU_LOAD *queue_cbu(struct cbu_state *cs, int clobber)
{
  struct CBU_LOAD *cl, *tail;

  cl = cs->s_tail;

  tail = cl->CBU_ANCHOR.a_prev;

  if(tail == cs->s_head){
    if(clobber){
      cs->s_head = tail->CBU_ANCHOR.a_prev;
      tail->CBU_ANCHOR.a_valid = 0;
    } else {
      return NULL;
    }
  } else {
    cs->s_used++;
  }

  cl->CBU_ANCHOR.a_valid = 1;
  cs->s_tail = tail;

  return cl;
}

struct CBU_LOAD *top_cbu(struct cbu_state *cs)
{
  if(cs->s_head == cs->s_tail){
#ifdef DEBUG
    fprintf(stderr, "top: empty\n");
#endif
    return NULL;
  }

#ifdef DEBUG
  if(cs->s_head->CBU_ANCHOR.a_valid == 0){
    fprintf(stderr, "top: logic failure: top not valid\n");
  }
#endif

  return cs->s_head;
}

struct CBU_LOAD *pop_cbu(struct cbu_state *cs)
{
  struct CBU_LOAD *cl;

  if(cs->s_head == cs->s_tail){
#ifdef DEBUG
    fprintf(stderr, "pop: empty\n");
#endif
    return NULL;
  }

#ifdef DEBUG
  if(cs->s_head->CBU_ANCHOR.a_valid == 0){
    fprintf(stderr, "pop: logic failure: top not valid\n");
  }
#endif

  cl = cs->s_head;

  cs->s_head = cl->CBU_ANCHOR.a_prev;

  cl->CBU_ANCHOR.a_valid = 0;
  cs->s_used--;
#ifdef DEBUG
  if(cs->s_used < 0){
    fprintf(stderr, "pop: logic failure: used=%d\n", cs->s_used);
  }
#endif

  return cl;
}

int check_cbu(struct cbu_state *cs, int crash)
{
  struct CBU_LOAD *cl;
  int size, used, valid;

  if(cs == NULL){
#ifdef DEBUG
    fprintf(stderr, "check: empty structure\n");
#endif
    if(crash) abort();
    return -1;
  }

  size = 0;
  used = 0;
  valid = 1;

  cl = cs->s_head;

  do{
#ifdef DEBUG
    fprintf(stderr, "check[%d]=%p:", size, cl);
#endif
    if(cl->CBU_ANCHOR.a_valid){
      used++;
#ifdef DEBUG
      fprintf(stderr, " used");
#endif
      if(valid == 0){
#ifdef DEBUG
        fprintf(stderr, " - item should not be marked as used\n");
#endif
        if(crash) abort();
        return -1;
      }
    } else {
#ifdef DEBUG
      fprintf(stderr, " empty");
#endif
      valid = 0;
    }
    size++;
#ifdef DEBUG
    if(cl == cs->s_head){
      fprintf(stderr, " [head]");
    }
    if(cl == cs->s_tail){
      fprintf(stderr, " [tail]");
    }
    fprintf(stderr, "\n");
#endif


    cl = cl->CBU_ANCHOR.a_prev;
  } while(cl != cs->s_head);

  if(cs->s_size != size){
#ifdef DEBUG
    fprintf(stderr, "check: size mismatch %d != %d\n", size, cs->s_size);
#endif
    if(crash) abort();
    return -1;
  }

  if(cs->s_used != used){
#ifdef DEBUG
    fprintf(stderr, "check: used mismatch %d != %d\n", used, cs->s_used);
#endif
    if(crash) abort();
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr, "check: %d/%d in use\n", cs->s_used, cs->s_size);
#endif

  return 0;
}

#endif
