/*
 * (c) Marc Welz 2005, 2006, 2007, 2008:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <loop.h>
#include <word-work.h>

#include "mxk.h"

/* chord: match multiple keys, generate one */

#define MXK_QUEUE_SIZE 16

#define MXK_CHORD_ENCODE_DEPTH(p)     (((p) << 16) & 0xff0000)
#define MXK_CHORD_DECODE_DEPTH(p)     (((p) >> 16) & 0xff)

#define MXK_CHORD_ENCODE_PRIORITY(p)  (((p) << 8) & 0xff00)
#define MXK_CHORD_DECODE_PRIORITY(p)  (((p) >> 8) & 0xff)

#define MXK_CHORD_ENCODE_PERMUTE(p)   ((p) & 0xff)
#define MXK_CHORD_DECODE_PERMUTE(p)   ((p) & 0xff)

struct mxk_chord{
  struct mxk_queue *h_queue;
  struct mxk_degap *h_degap;
  struct mxk_key *h_key;

  int h_sync;
  int h_bypass;

  struct mxk_client *h_vector;
  int h_count;
};

static struct mxk_chord *mxk_create_chord();
static void mxk_destroy_chord(struct mxk_chord *mh);

static int mxk_bypass_chord(struct list_loop *ls, struct task_loop *tx, int bypass, char *args);
static int mxk_next_permutation_chord(int *vector, int size);
static int mxk_add_chord(struct list_loop *ls, struct task_loop *tx, int field, char *args);

static int mxk_match_chord(struct mxk_chord *mh);
static int mxk_run_chord(struct list_loop *ls, struct task_loop *tl, int mask);

/**************************************************************************/

static struct mxk_chord *mxk_create_chord()
{
  struct mxk_chord *mh;

  mh = malloc(sizeof(struct mxk_chord));
  if(mh == NULL){
    return NULL;
  }

  mh->h_queue = NULL;
  mh->h_degap = NULL;
  mh->h_key = NULL;

  mh->h_vector = NULL;
  mh->h_count = 0;

  mh->h_sync = 0;
  mh->h_bypass = 0;

  mh->h_queue = mxk_start_queue(MXK_QUEUE_SIZE);
  if(mh->h_queue == NULL){
    mxk_destroy_chord(mh);
    return NULL;
  }

  mh->h_degap = mxk_start_degap();
  if(mh->h_degap == NULL){
    mxk_destroy_chord(mh);
    return NULL;
  }

  mh->h_key = mxk_start_key();
  if(mh->h_key == NULL){
    mxk_destroy_chord(mh);
    return NULL;
  }

  return mh;
}

static void mxk_destroy_chord(struct mxk_chord *mh)
{
  if(mh){
    if(mh->h_queue){
      mxk_stop_queue(mh->h_queue);
      mh->h_queue = NULL;
    }
    if(mh->h_degap){
      mxk_stop_degap(mh->h_degap);
      mh->h_degap = NULL;
    }
    if(mh->h_key){
      mxk_stop_key(mh->h_key);
      mh->h_key = NULL;
    }

    mxk_shutdown_client(mh->h_vector, mh->h_count);
    mh->h_vector = NULL;
    mh->h_count = 0;

    free(mh);
  }
}

/**************************************************************************/

static struct code_name mxk_queue_map[] = {
  { "press",  MXK_QUEUE_OLD_PRESS  },
  { "release",  MXK_QUEUE_NEW_RELEASE},
  { "old",  MXK_QUEUE_OLD_RELEASE},
  { "end",  MXK_QUEUE_TOTAL_SIZE },
  { NULL,  -1 },
};

struct mxk_chord_stack
{
  struct mxk_key *c_key;
  int c_current;
};

#define MXK_CHORD_CSIZE 4

int mxk_list_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_chord *mh;
  struct mxk_queue *mq;
  struct mxk_key *mk;
  int cd, type, id, idx, count, i, j, cp, csize, m;
  struct mxk_chord_stack *cs, *ct;
  char x, y;
  char *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "chord");
  if(type <= 0){
    print_control(cl, cd, 1, "chord: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      ptr = args + idx;
      id = strtoul(ptr, NULL, 10);
    }
  }

  count = 0;
  csize = MXK_CHORD_CSIZE;
  cs = malloc(sizeof(struct mxk_chord_stack) * csize);

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    mh = user_loop(ls, tx);
    if(mh){
      print_control(cl, cd, 0, "chord[%d]: unmatched=%s\n", tag_loop(ls, tx), mh->h_bypass ? "bypass" : "ignore");
      mq = mh->h_queue;
      print_control(cl, cd, 0, "chord queue:");
      i = j = 0;
      while(i < MXK_QUEUE_MAP){
        if(j < mq->q_index[i]){
          x = '<';
          ptr = mxk_reverse_degap(mh->h_degap, mq->q_queue[j].i_code);
          y = '>';
          /* WARNING: probably incorrect, need to be mod queue size */
          j++;
        } else {
          x = '[';
          ptr = code_to_name(mxk_queue_map, i);
          y = ']';
          i++;
        }
        if(ptr){
          print_control(cl, cd, 0, " %c%s%c", x, ptr, y);
        }
      }
      print_control(cl, cd, 0, "\n");

      cp = 0;
      cs[cp].c_key = mh->h_key;
      cs[cp].c_current = 0;
      m = 1;
      while(cp >= 0){
#if DEBUG > 1
        fprintf(stderr, "chord: cs[%d]<%p>: size=%d, current=%d, m=%d, x=%d\n", cp, cs[cp].c_key, cs[cp].c_key->k_size, cs[cp].c_current, m, cs[cp].c_key->k_size ? 0 : cs[cp].c_key->u.k_value);
#endif
        switch(m){
          case -1 : 
           if((cs[cp].c_current + 1) < cs[cp].c_key->k_size){ /* marking and changing direction */
             cs[cp].c_current = cs[cp].c_current + 1;
             m = 1;
           } else { /* ascending */
             cp--;
           }
          break;
          case  1 : 
           if(cs[cp].c_key->k_size > 0){ /* descending */
             ct = cs;
             if((cp + 1) >= csize){
               ct = realloc(cs, sizeof(struct mxk_chord_stack) * (csize + 1));
               if(ct){
                 csize++;
               }
             }
             if(ct){
               cs = ct;
               cs[cp + 1].c_current = 0;
               cs[cp + 1].c_key = &(cs[cp].c_key->u.k_next[cs[cp].c_current]);
               cp++;
             } else {
               cp = (-1);
             }
           } else { /* leaf */
             mk = cs[cp].c_key;
             if(mk->u.k_value >= 0){
               j = mk->u.k_value;
               ptr = code_to_name(mxk_key_codes, j);
               if(ptr){
                 print_control(cl, cd, 0, "chord: %s <-", ptr);
               } else {
                 print_control(cl, cd, 0, "chord: %d <-", j);
               }
               
               for(i = cp - 1; i >= 0; i--){
                 j = cs[i].c_current;
                 ptr = mxk_reverse_degap(mh->h_degap, j);
                 if(ptr){
                   print_control(cl, cd, 0, " %s", ptr);
                 } else {
                   print_control(cl, cd, 0, " %d", j);
                 }
               }

               ptr = code_to_name(mxk_queue_map, mk->k_depth);
               if(ptr){
                 print_control(cl, cd, 0, " [%s]", ptr);
               }
               print_control(cl, cd, 0, " (%u)", mk->k_priority);

               print_control(cl, cd, 0, "\n");
             }

             m = (-1);
           }
        }
      }

      count++;
    } else {
      print_control(cl, cd, 0, "chord[%d]: unavailable\n", id);
    }
  }

  if(cs){
    free(cs);
  }

  print_control(cl, cd, 1, "chord: %d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

int mxk_start_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_chord *mh;
  int cd, type, id, idx;
  char *value;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "chord");

#ifdef DEBUG
  if(type <= 0){
    fprintf(stderr, "chord: logic problem: type not defined\n");
    abort();
  }
#endif

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
#ifdef DEBUG
      fprintf(stderr, "chord: key is %d,%d\n", type, id);
#endif
    }
  }

  mh = mxk_create_chord(MXK_QUEUE_OLD_RELEASE);
  if(mh == NULL){
    print_control(cl, cd, 1, "unable to allocate chord state\n");
    return 0;
  }

  tx = add_tag_loop(ls, -1, mh, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_chord, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "unable to start chord task (%d,%d)\n", type, id);
    mxk_destroy_chord(mh);
    return 0;
  }

  text_loop(ls, tx, "chord:%d", id);
  print_control(cl, cd, 1, "created chord task (%d,%d)\n", type, id);

  return 0;
}

/************************************************************************/

int mxk_bypass_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_OPTIONAL, 1, mxk_bypass_chord);

  return 0;
}

int mxk_ignore_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_OPTIONAL, 0, mxk_bypass_chord);

  return 0;
}

static int mxk_bypass_chord(struct list_loop *ls, struct task_loop *tx, int bypass, char *args)
{
  struct mxk_chord *mh;

  mh = user_loop(ls, tx);

  mh->h_bypass = bypass;

  return 0;
}

/*************************************************************************/

int mxk_old_order_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int field;

  field = MXK_CHORD_ENCODE_DEPTH(MXK_QUEUE_OLD_RELEASE) | MXK_CHORD_ENCODE_PRIORITY(1) | MXK_CHORD_ENCODE_PERMUTE(0);

  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_REQUIRED, field, mxk_add_chord);

  return 0;
}

int mxk_old_set_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int field;

  field = MXK_CHORD_ENCODE_DEPTH(MXK_QUEUE_OLD_RELEASE) | MXK_CHORD_ENCODE_PRIORITY(0) | MXK_CHORD_ENCODE_PERMUTE(1);

  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_REQUIRED, field, mxk_add_chord);

  return 0;
}

int mxk_new_order_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int field;

  field = MXK_CHORD_ENCODE_DEPTH(MXK_QUEUE_NEW_RELEASE) | MXK_CHORD_ENCODE_PRIORITY(3) | MXK_CHORD_ENCODE_PERMUTE(0);

  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_REQUIRED, field, mxk_add_chord);

  return 0;
}

int mxk_new_set_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int field;

  field = MXK_CHORD_ENCODE_DEPTH(MXK_QUEUE_NEW_RELEASE) | MXK_CHORD_ENCODE_PRIORITY(2) | MXK_CHORD_ENCODE_PERMUTE(1);

  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_REQUIRED, field, mxk_add_chord);

  return 0;
}

int mxk_gap_order_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int field;

  field = MXK_CHORD_ENCODE_DEPTH(MXK_QUEUE_OLD_PRESS) | MXK_CHORD_ENCODE_PRIORITY(5) | MXK_CHORD_ENCODE_PERMUTE(0);

  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_REQUIRED, field, mxk_add_chord);

  return 0;
}

int mxk_gap_set_chord_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int field;

  field = MXK_CHORD_ENCODE_DEPTH(MXK_QUEUE_OLD_PRESS) | MXK_CHORD_ENCODE_PRIORITY(4) | MXK_CHORD_ENCODE_PERMUTE(1);

  mxk_parse_apply(ls, tl, args, "chord", MXK_MISC_REQUIRED, field, mxk_add_chord);

  return 0;
}

static int mxk_next_permutation_chord(int *vector, int size)
{
  int i;
  int *ptr;
#ifdef DEBUG
  int j;
#endif

  /* could be replaced with "int ptr[size];" */
  ptr = alloca(sizeof(int) * size);
  if(ptr == NULL){
    return -1;
  }

  for(i = 0; i < size; i++){
    ptr[i] = 0;
  }

  i = 0;
  ptr[vector[i]] = 1;
  vector[i] = vector[i] + 1;

  for(;;){
    if(vector[i] >= size){
#if DEBUG > 1
      fprintf(stderr, "permute: vector[%d] = %d, exceeds > %d\n", i, vector[i], size);
#endif
      if((i + 1)>= size){
#if DEBUG > 1
        fprintf(stderr, "permute: problem space exhausted\n");
#endif
        return 0;
      } else {
        vector[i] = 0;
        i++;
        ptr[vector[i]] = 1;
        vector[i] = vector[i] + 1;
      }
    } else if(ptr[vector[i]] == 0){
#if DEBUG > 1
      fprintf(stderr, "vector[%d] = %d, already in use\n", i, vector[i]);
#endif
      vector[i] = vector[i] + 1;
    } else {
#if DEBUG > 1
      fprintf(stderr, "vector[%d] = %d, available, reducing to %d\n", i, vector[i], i - 1);
#endif
      ptr[vector[i]] = 0;
      i--;
      if(i < 0){
#ifdef DEBUG
        for(j = 0; j < size; j++){
          if(ptr[j]){
            fprintf(stderr, "logic problem: %d not allocated\n", j);
            abort();
          }
        }
#endif
        return 1;
      }
    }
  }
}

static int mxk_add_chord(struct list_loop *ls, struct task_loop *tx, int field, char *args)
{
  struct mxk_chord *mh;
  char *number, *end;
  int *ptr, *vector, *reorder, *sequence;
  int value, elements, i, result, idx, priority, permute, depth;
#if 0
  int count;
#endif

#ifdef DEBUG
  fprintf(stderr, "chord: about to parse <%s>\n", args);
#endif

  if(args == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  permute = MXK_CHORD_DECODE_PERMUTE(field);
  depth   = MXK_CHORD_DECODE_DEPTH(field);

  number = args + idx;
  priority = strtoul(number, &end, 10);
  if(end[0] == ' '){
    idx = next_word(args, idx);
    if(idx < 0){
      return -1;
    }
  } else {
    priority = MXK_CHORD_DECODE_PRIORITY(field);
  }

  ptr = mxk_parse_vector_key(args + idx, 0);
  if(ptr == NULL){
    return -1;
  }

  value = ptr[0];
  vector = &(ptr[1]);

  if((value == (-1)) || (vector[0] == (-1))){
    free(ptr);
    return -1;
  }

#if 0
  mxk_reverse_chord(vector);
#endif

  mh = user_loop(ls, tx);
  if(mh == NULL){
    free(ptr);
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr, "chord: adding to degap\n");
#endif

  /* insert into reduction table */
  for(i = 0; vector[i] >= 0; i++){
    if(mxk_insert_degap(mh->h_degap, vector[i]) < 0){
      free(ptr);
      return -1;
    }
  }
  elements = i;

#if 0
  mxk_dump_degap(mh->h_degap, stderr);
  fprintf(stderr, "chord: using degap\n");
#endif
  for(i = 0; i < elements; i++){
    vector[i] = mxk_lookup_degap(mh->h_degap, vector[i]);
  }
#ifdef DEBUG
  fprintf(stderr, "chord: constructed degap (%d elements): ", elements);
  for(i = 0; i < elements; i++){
    fprintf(stderr, " %d", vector[i]);
  }
  fprintf(stderr, "\n");
#endif

  sequence = malloc(sizeof(int) * (elements));
  if(sequence == NULL){
    free(ptr);
    return -1;
  }
  for(i = 0; i < elements; i++){
    sequence[i] = elements - (i + 1);
  }

  reorder = malloc(sizeof(int) * (elements + 1));
  if(reorder == NULL){
    free(ptr);
    free(sequence);
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr, "chord: building key\n");
#endif

#if 0
  count = 0;
#endif
  do{
    for(i = 0; i < elements; i++){
      reorder[sequence[i]] = vector[i];
    }
#ifdef DEBUG
    fprintf(stderr, "chord: inserting reordered: ");
    for(i = 0; i < elements; i++){
      fprintf(stderr, " %d", reorder[i]);
    }
    fprintf(stderr, "\n");
#endif
    reorder[elements] = (-1);
    result = mxk_insert_key(mh->h_key, reorder, value, priority, depth);
#if 0
    if(result > 0){
      count++;
    }
#endif
  } while((result >= 0) && (permute && (mxk_next_permutation_chord(sequence, elements) > 0)));

  free(ptr);
  free(reorder);
  free(sequence);

  if(result < 0){
    return -1;
  }

  return 0;
}

/************************************************************************/

static int mxk_match_chord(struct mxk_chord *mh)
{
  struct mxk_queue *mq;
  struct mxk_key *mk, *xk;
  int i, limit, code, result, priority, length;

  length = 0;
  priority = (-1);
  result = (-1);
  mq = mh->h_queue;
  mk = mh->h_key;
  limit = mq->q_index[MXK_QUEUE_OLD_RELEASE];

#ifdef DEBUG
  fprintf(stderr, "match: limit is %d\n", limit);
#endif

  i = 0;
  for(;;){

    if(i <= mq->q_index[mk->k_depth]){
      if(mk->k_priority >= priority){
        priority = mk->k_priority;
        if(mk->k_size <= 0){
          result = mk->u.k_value;
        } else {
          xk = &(mk->u.k_next[0]);
#ifdef DEBUG
          if(xk->k_size > 0){
            fprintf(stderr, "match: major logic failure: @[%d]: element 0 not empty\n", i);
            abort();
          }
#endif
          result = xk->u.k_value;
        }
        if(result >= 0){
          length = i;
        }
#ifdef DEBUG
        fprintf(stderr, "match: updating best %s match @[%d] to %d(p=%d, l=%d)\n", mk->k_size ? "interior" : "leaf", i, result, priority, length);
#endif
#ifdef DEBUG
      } else {
        fprintf(stderr, "match: not considering@%d: priority %d too low\n", i, mk->k_priority);
#endif
      }
#ifdef DEBUG
    } else {
      fprintf(stderr, "match: not considering@%d: depth %d ends before %d\n", i, mk->k_depth, mq->q_index[mk->k_depth]);
#endif
    }

    if(mk->k_size <= 0){
#ifdef DEBUG
      fprintf(stderr, "match: encountered terminal@%d\n", i);
#endif
      mxk_consume_queue(mq, length);
      return result;
    }

    code = mq->q_queue[i].i_code;
    if(mk->k_size <= code){
#ifdef DEBUG
      fprintf(stderr, "match: encountered short interior@%d: %d <= %d\n", i, mk->k_size, code);
#endif
      mxk_consume_queue(mq, length);
      return result;
    }

    if(i < limit){
      mk = &(mk->u.k_next[code]);
#ifdef DEBUG
      fprintf(stderr, "match: further descent@%d: via index %d\n", i, code);
#endif
      i++;
    } else {
#ifdef DEBUG
      fprintf(stderr, "match: reached limit@%d\n", limit);
#endif
      mxk_consume_queue(mq, length);
      return result;
    }

  }
}

static int mxk_run_chord(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_chord *mh;
  struct mxk_global *mg;
  struct link_loop *nk;
  struct input_event *vt;
  int code, result;

  mh = user_loop(ls, tl);
  mg = global_loop(ls);

  if(mask & LO_STOP_MASK){
    mxk_destroy_chord(mh);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vt = nk->k_data;
        switch(vt->type){
          case EV_SYN :
            if(mh->h_sync == 0){
              mxk_send_client(ls, tl, mh->h_vector, mh->h_count, EV_SYN, 0, 0);
              mh->h_sync = 1;
            }
            break;
          case EV_KEY :
            code = mxk_lookup_degap(mh->h_degap, vt->code);
            if(code >= 0){
              mxk_update_queue(ls, mh->h_queue, code, vt->value);

              code = mxk_match_chord(mh);
#ifdef DEBUG
              fprintf(stderr, "chord: match result=%d\n", code);
#endif
              if(code >= 0){
                result  = mxk_send_client(ls, tl, mh->h_vector, mh->h_count, EV_KEY, code, 1);
                result += mxk_send_client(ls, tl, mh->h_vector, mh->h_count, EV_SYN, 0, 0);
                result += mxk_send_client(ls, tl, mh->h_vector, mh->h_count, EV_KEY, code, 0);
                mh->h_sync = 0;

#ifdef DEBUG
                if(result < 0){
                  fprintf(stderr, "chord: problem sending to clients\n");
                }
#endif
              }
            } else {
              if(mh->h_bypass){
                mxk_send_client(ls, tl, mh->h_vector, mh->h_count, vt->type, vt->code, vt->value);
                mh->h_sync = 0;
              } else {
#ifdef DEBUG
                fprintf(stderr, "chord: suppressing key 0x%x\n", vt->code);
#endif
              }
            }
            break;
          default :
#ifdef DEBUG
            fprintf(stderr, "chord: odd, %d not a key, discarding\n", vt->type);
#endif
            /* could also have a bypass mode for this */
            break;
        }
      } else /* if ... */ {
#ifdef DEBUG
        fprintf(stderr, "chord: received odd link type %d\n", nk->k_type);
#endif
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/********************************************************************/

int mxk_insert_chord(void *source, int type, int id)
{
  struct mxk_chord *mh;
  struct mxk_client *mc;

  mh = source;

  mc = mxk_insert_client(mh->h_vector, mh->h_count, type, id);
  if(mc == NULL){
    return -1;
  }

  mh->h_vector = mc;
  mh->h_count++;

  return 0;
}

int mxk_startup_chord(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "chord", MXK_EVT_TARGET | MXK_EVT_SOURCE, -1, mxk_insert_chord);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_chord(struct mxk_global *mg)
{
}
