/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <linux/input.h>

#include <code-name.h>

struct code_name mxk_event_types[] = {
  { "sync",      EV_SYN },
  { "key",       EV_KEY },
  { "relative",  EV_REL },
  { "absolute",  EV_ABS },
  { "miscellany",EV_MSC },
#ifdef EV_SW
  { "sw",        EV_SW  },
#endif
  { "led",       EV_LED },
  { "sound",     EV_SND },
  { "repeat",    EV_REP },
  { "feedback",  EV_FF },
  { "power",     EV_PWR },
  { "status",    EV_FF_STATUS },
  { NULL,        -1 }
};

struct code_name mxk_led_codes[] = {
  { "num",     LED_NUML },
  { "caps",    LED_CAPSL },
  { "scroll",  LED_SCROLLL },
  { "compose", LED_COMPOSE },
  { "kana",    LED_KANA },
  { "sleep",   LED_SLEEP },
  { "suspend", LED_SUSPEND },
  { "mute",    LED_MUTE },
  { "misc",    LED_MISC },
#ifdef LED_MAIL
  { "mail",    LED_MAIL },
#endif
#ifdef LED_CHARGING
  { "charging",    LED_CHARGING },
#endif
  { NULL, -1 }
};

struct code_name mxk_key_values[] = {
  { "release", 0 },
  { "press",   1 },
  { "repeat",  2 },
  { NULL, -1 }
};

struct code_name mxk_key_codes[] = {
  { "one", KEY_1 },
  { "two", KEY_2 },
  { "three", KEY_3 },
  { "four", KEY_4 },
  { "five", KEY_5 },
  { "six", KEY_6 },
  { "seven", KEY_7 },
  { "eight", KEY_8 },
  { "nine", KEY_9 },
  { "zero", KEY_0 },

  { "q", KEY_Q },
  { "w", KEY_W },
  { "e", KEY_E },
  { "r", KEY_R },
  { "t", KEY_T },
  { "y", KEY_Y },
  { "u", KEY_U },
  { "i", KEY_I },
  { "o", KEY_O },
  { "p", KEY_P },
  { "a", KEY_A },
  { "s", KEY_S },
  { "d", KEY_D },
  { "f", KEY_F },
  { "g", KEY_G },
  { "h", KEY_H },
  { "j", KEY_J },
  { "k", KEY_K },
  { "l", KEY_L },
  { "z", KEY_Z },
  { "x", KEY_X },
  { "c", KEY_C },
  { "v", KEY_V },
  { "b", KEY_B },
  { "n", KEY_N },
  { "m", KEY_M },

  { "escape", KEY_ESC },
  { "backspace", KEY_BACKSPACE },
  { "tab", KEY_TAB },
  { "comma", KEY_COMMA },
  { "dot", KEY_DOT },
  { "semicolon", KEY_SEMICOLON },
  { "apostrophe", KEY_APOSTROPHE },
  { "grave", KEY_GRAVE },
  { "minus", KEY_MINUS },
  { "equal", KEY_EQUAL },
  { "slash", KEY_SLASH },
  { "space", KEY_SPACE },
  { "enter", KEY_ENTER },
  { "backslash", KEY_BACKSLASH },

  { "leftbrace", KEY_LEFTBRACE },
  { "rightbrace", KEY_RIGHTBRACE },

  { "leftshift", KEY_LEFTSHIFT },
  { "rightshift", KEY_RIGHTSHIFT },
  { "leftalt", KEY_LEFTALT },
  { "rightalt", KEY_RIGHTALT },
  { "leftcontrol", KEY_LEFTCTRL },
  { "rightcontrol", KEY_RIGHTCTRL },

  { "f1", KEY_F1 },
  { "f2", KEY_F2 },
  { "f3", KEY_F3 },
  { "f4", KEY_F4 },
  { "f5", KEY_F5 },
  { "f6", KEY_F6 },
  { "f7", KEY_F7 },
  { "f8", KEY_F8 },
  { "f9", KEY_F9 },
  { "f10", KEY_F10 },
  { "f11", KEY_F11 },
  { "f12", KEY_F12 },

  { "capslock", KEY_CAPSLOCK },
  { "numlock", KEY_NUMLOCK },
  { "scrolllock", KEY_SCROLLLOCK },

  { "sysrq",  KEY_SYSRQ },
  { "pause",  KEY_PAUSE },

  { "insert",  KEY_INSERT },
  { "delete",  KEY_DELETE },
  { "home",  KEY_HOME },
  { "end",  KEY_END },
  { "pageup",  KEY_PAGEUP },
  { "pagedown",  KEY_PAGEDOWN },

  { "left",  KEY_LEFT },
  { "right",  KEY_RIGHT },
  { "up",  KEY_UP },
  { "down",  KEY_DOWN },

  { "kp0",  KEY_KP0 },
  { "kp1",  KEY_KP1 },
  { "kp2",  KEY_KP2 },
  { "kp3",  KEY_KP3 },
  { "kp4",  KEY_KP4 },
  { "kp5",  KEY_KP5 },
  { "kp6",  KEY_KP6 },
  { "kp7",  KEY_KP7 },
  { "kp8",  KEY_KP8 },
  { "kp9",  KEY_KP9 },

  { "kpdot",  KEY_KPDOT },
  { "kpplus",  KEY_KPPLUS },
  { "kpminus",  KEY_KPMINUS },
  { "kpenter",  KEY_KPENTER },
  { "kpslash",  KEY_KPSLASH },
  { "kpasterisk",  KEY_KPASTERISK },

  { "leftmeta", KEY_LEFTMETA },
  { "rightmeta", KEY_RIGHTMETA },
  { "compose", KEY_COMPOSE },

  { "stop", KEY_STOP },
  { "again", KEY_AGAIN },
  { "props", KEY_PROPS },
  { "undo", KEY_UNDO },
  { "front", KEY_FRONT },
  { "copy", KEY_COPY },
  { "open", KEY_OPEN },
  { "paste", KEY_PASTE },
  { "find", KEY_FIND },
  { "cut", KEY_CUT },
  { "help", KEY_HELP },
  { "menu", KEY_MENU },
  { "calc", KEY_CALC },
  { "setup", KEY_SETUP },
  { "sleep", KEY_SLEEP },
  { "wakeup", KEY_WAKEUP },
  { "file", KEY_FILE },
  { "sendfile", KEY_SENDFILE },
  { "deletefile", KEY_DELETEFILE },
  { "xfer", KEY_XFER },
  { "prog1", KEY_PROG1 },
  { "prog2", KEY_PROG2 },
  { "www", KEY_WWW },
  { "msdos", KEY_MSDOS },
  { "coffee", KEY_COFFEE },
  { "direction", KEY_DIRECTION },
  { "cyclewindows", KEY_CYCLEWINDOWS },
  { "mail", KEY_MAIL },
  { "bookmarks", KEY_BOOKMARKS },
  { "computer", KEY_COMPUTER },
  { "back", KEY_BACK },
  { "forward", KEY_FORWARD },
  { "closecd", KEY_CLOSECD },
  { "ejectcd", KEY_EJECTCD },
  { "ejectclosecd", KEY_EJECTCLOSECD },
  { "nextsong", KEY_NEXTSONG },
  { "playpause", KEY_PLAYPAUSE },
  { "previoussong", KEY_PREVIOUSSONG },
  { "stopcd", KEY_STOPCD },
  { "record", KEY_RECORD },
  { "rewind", KEY_REWIND },
  { "phone", KEY_PHONE },
  { "iso", KEY_ISO },
  { "config", KEY_CONFIG },
  { "homepage", KEY_HOMEPAGE },
  { "refresh", KEY_REFRESH },
  { "exit", KEY_EXIT },
  { "move", KEY_MOVE },
  { "edit", KEY_EDIT },
  { "scrollup", KEY_SCROLLUP },
  { "scrolldown", KEY_SCROLLDOWN },

  { "kpleftparen", KEY_KPLEFTPAREN },
  { "kprightparen", KEY_KPRIGHTPAREN },
#ifdef KEY_KPCOMMA
  { "kpcomma", KEY_KPCOMMA },
#endif

#if 0
  { "intl1", KEY_INTL1 },
  { "intl2", KEY_INTL2 },
  { "intl3", KEY_INTL3 },
  { "intl4", KEY_INTL4 },
  { "intl5", KEY_INTL5 },
  { "intl6", KEY_INTL6 },
  { "intl7", KEY_INTL7 },
  { "intl8", KEY_INTL8 },
  { "intl9", KEY_INTL9 },
  { "lang1", KEY_LANG1 },
  { "lang2", KEY_LANG2 },
  { "lang3", KEY_LANG3 },
  { "lang4", KEY_LANG4 },
  { "lang5", KEY_LANG5 },
  { "lang6", KEY_LANG6 },
  { "lang7", KEY_LANG7 },
  { "lang8", KEY_LANG8 },
  { "lang9", KEY_LANG9 },
#endif

  { "playcd", KEY_PLAYCD },
  { "pausecd", KEY_PAUSECD },
  { "prog3", KEY_PROG3 },
  { "prog4", KEY_PROG4 },
  { "suspend", KEY_SUSPEND },
  { "close", KEY_CLOSE },
  { "play", KEY_PLAY },
  { "fastforward", KEY_FASTFORWARD },
  { "bassboost", KEY_BASSBOOST },
  { "print", KEY_PRINT },
  { "hp", KEY_HP },
  { "camera", KEY_CAMERA },
  { "sound", KEY_SOUND },
  { "question", KEY_QUESTION },
  { "email", KEY_EMAIL },
  { "chat", KEY_CHAT },
  { "search", KEY_SEARCH },
  { "connect", KEY_CONNECT },
  { "finance", KEY_FINANCE },
  { "sport", KEY_SPORT },
  { "shop", KEY_SHOP },
  { "alterase", KEY_ALTERASE },
  { "cancel", KEY_CANCEL },
  { "brightnessdown", KEY_BRIGHTNESSDOWN },
  { "brightnessup", KEY_BRIGHTNESSUP },
  { "media", KEY_MEDIA },

#ifdef KEY_SWITCHVIDEOMODE
  { "switchvideomode", KEY_SWITCHVIDEOMODE },
#endif
#ifdef KEY_KBDILLUMTOGGLE
  { "ki-toggle", KEY_KBDILLUMTOGGLE },
#endif
#ifdef KEY_KBDILLUMDOWN
  { "ki-down", KEY_KBDILLUMDOWN },
#endif
#ifdef KEY_KBDILLUMUP
  { "ki-up", KEY_KBDILLUMUP },
#endif
#ifdef KEY_SEND
  { "send", KEY_SEND },
#endif
#ifdef KEY_REPLY
  { "reply", KEY_REPLY },
#endif
#ifdef KEY_FORWARDMAIL
  { "forwardmail", KEY_FORWARDMAIL },
#endif
#ifdef KEY_SAVE
  { "save", KEY_SAVE },
#endif
#ifdef KEY_DOCUMENTS
  { "documents", KEY_DOCUMENTS },
#endif
#ifdef KEY_BATTERY
  { "battery", KEY_BATTERY },
#endif
#ifdef KEY_BLUETOOTH
  { "bluetooth", KEY_BLUETOOTH },
#endif
#ifdef KEY_WLAN
  { "wlan", KEY_WLAN },
#endif

  { "unknown", KEY_UNKNOWN },

  { "btn-misc", BTN_MISC },
  { "btn-0", BTN_0 },
  { "btn-1", BTN_1 },
  { "btn-2", BTN_2 },
  { "btn-3", BTN_3 },
  { "btn-4", BTN_4 },
  { "btn-5", BTN_5 },
  { "btn-6", BTN_6 },
  { "btn-7", BTN_7 },
  { "btn-8", BTN_8 },
  { "btn-9", BTN_9 },

  /* mouse buttons */
  { "leftbutton", BTN_LEFT },
  { "rightbutton", BTN_RIGHT },
  { "middlebutton", BTN_MIDDLE },
  { "sidebutton",  BTN_SIDE },
  { "extrabutton",  BTN_EXTRA },
  { "forwardbutton",  BTN_FORWARD },
  { "backbutton",  BTN_BACK },
  { "taskbutton",  BTN_TASK },

  /* joystick buttons */
  { "trigger", BTN_TRIGGER },
  { "thumb", BTN_THUMB },
  { "thumb2", BTN_THUMB2 },
  { "top", BTN_TOP },
  { "top2", BTN_TOP2 },
  { "pinkie", BTN_PINKIE },
  { "base", BTN_BASE },
  { "base2", BTN_BASE2 },
  { "base3", BTN_BASE3 },
  { "base4", BTN_BASE4 },
  { "base5", BTN_BASE5 },
  { "base6", BTN_BASE6 },
  { "dead", BTN_DEAD },

  { "gamepad", BTN_GAMEPAD },
  { "btn-a", BTN_A },
  { "btn-b", BTN_B },
  { "btn-c", BTN_C },
  { "btn-x", BTN_X },
  { "btn-y", BTN_Y },
  { "btn-z", BTN_Z },
  { "tl", BTN_TL },
  { "tr", BTN_TR },
  { "tl2", BTN_TL2 },
  { "tr2", BTN_TR2 },
  { "btn-select", BTN_SELECT },
  { "start", BTN_START },
  { "mode", BTN_MODE },
  { "thumbl", BTN_THUMBL },
  { "thumbr", BTN_THUMBR },

  { "digi", BTN_DIGI },
  { "touch", BTN_TOUCH },
  { "stylus", BTN_STYLUS },
  { "stylus2", BTN_STYLUS2 },
  { "tool-pen", BTN_TOOL_PEN },
  { "tool-rubber", BTN_TOOL_RUBBER },
  { "tool-brush", BTN_TOOL_BRUSH },
  { "tool-pencil", BTN_TOOL_PENCIL },
  { "tool-airbrush", BTN_TOOL_AIRBRUSH },
  { "tool-finger", BTN_TOOL_FINGER },
  { "tool-mouse", BTN_TOOL_MOUSE },
  { "tool-lens", BTN_TOOL_LENS },
  { "tool-doubletap", BTN_TOOL_DOUBLETAP },
  { "tool-tripletap", BTN_TOOL_TRIPLETAP },

  { "wheel", BTN_WHEEL },
  { "gear-down", BTN_GEAR_DOWN },
  { "gear-up", BTN_GEAR_UP },

  { "ok", KEY_OK },
  { "select", KEY_SELECT },
  { "goto", KEY_GOTO },
  { "clear", KEY_CLEAR },
  { "power2", KEY_POWER2 },
  { "option", KEY_OPTION },
  { "info", KEY_INFO },
  { "time", KEY_TIME },
  { "vendor", KEY_VENDOR },
  { "archive", KEY_ARCHIVE },
  { "program", KEY_PROGRAM },
  { "channel", KEY_CHANNEL },
  { "favorites", KEY_FAVORITES },
  { "epg", KEY_EPG },
  { "pvr", KEY_PVR },
  { "mhp", KEY_MHP },
  { "language", KEY_LANGUAGE },
  { "title", KEY_TITLE },
  { "subtitle", KEY_SUBTITLE },
  { "angle", KEY_ANGLE },
  { "zoom", KEY_ZOOM },
  { "mode", KEY_MODE },
  { "keyboard", KEY_KEYBOARD },
  { "screen", KEY_SCREEN },
  { "pc", KEY_PC },
  { "tv", KEY_TV },
  { "tv2", KEY_TV2 },
  { "vcr", KEY_VCR },
  { "vcr2", KEY_VCR2 },
  { "sat", KEY_SAT },
  { "sat2", KEY_SAT2 },
  { "cd", KEY_CD },
  { "tape", KEY_TAPE },
  { "radio", KEY_RADIO },
  { "tuner", KEY_TUNER },
  { "player", KEY_PLAYER },
  { "text", KEY_TEXT },
  { "dvd", KEY_DVD },
  { "aux", KEY_AUX },
  { "mp3", KEY_MP3 },
  { "audio", KEY_AUDIO },
  { "video", KEY_VIDEO },
  { "directory", KEY_DIRECTORY },
  { "list", KEY_LIST },
  { "memo", KEY_MEMO },
  { "calendar", KEY_CALENDAR },
  { "red", KEY_RED },
  { "green", KEY_GREEN },
  { "yellow", KEY_YELLOW },
  { "blue", KEY_BLUE },
  { "channelup", KEY_CHANNELUP },
  { "channeldown", KEY_CHANNELDOWN },
  { "first", KEY_FIRST },
  { "last", KEY_LAST },
  { "ab", KEY_AB },
  { "next", KEY_NEXT },
  { "restart", KEY_RESTART },
  { "slow", KEY_SLOW },
  { "shuffle", KEY_SHUFFLE },
  { "break", KEY_BREAK },
  { "previous", KEY_PREVIOUS },
  { "digits", KEY_DIGITS },
  { "teen", KEY_TEEN },
  { "twen", KEY_TWEN },

#ifdef KEY_VIDEOPHONE
  { "videophone", KEY_VIDEOPHONE     },
#endif
#ifdef KEY_GAMES
  { "games", KEY_GAMES          },
#endif
#ifdef KEY_ZOOMIN
  { "zoomin", KEY_ZOOMIN         },
#endif
#ifdef KEY_ZOOMOUT
  { "zoomout", KEY_ZOOMOUT        },
#endif
#ifdef KEY_ZOOMRESET
  { "zoomreset", KEY_ZOOMRESET      },
#endif
#ifdef KEY_WORDPROCESSOR
  { "wordprocessor", KEY_WORDPROCESSOR  },
#endif
#ifdef KEY_EDITOR
  { "editor", KEY_EDITOR         },
#endif
#ifdef KEY_SPREADSHEET
  { "spreadsheet", KEY_SPREADSHEET    },
#endif
#ifdef KEY_GRAPHICSEDITOR
  { "graphicseditor", KEY_GRAPHICSEDITOR },
#endif
#ifdef KEY_PRESENTATION
  { "presentation", KEY_PRESENTATION   },
#endif
#ifdef KEY_DATABASE
  { "database", KEY_DATABASE       },
#endif
#ifdef KEY_NEWS
  { "news", KEY_NEWS           },
#endif
#ifdef KEY_VOICEMAIL
  { "voicemail", KEY_VOICEMAIL      },
#endif
#ifdef KEY_ADDRESSBOOK
  { "addressbook", KEY_ADDRESSBOOK    },
#endif
#ifdef KEY_MESSENGER
  { "messenger", KEY_MESSENGER      },
#endif

  { "del-eol", KEY_DEL_EOL },
  { "del-eos", KEY_DEL_EOS },
  { "ins-line", KEY_INS_LINE },
  { "del-line", KEY_DEL_LINE },

#if 0
  /* made up keys, unique+original+shortish town names in za */
  { "pretoria", KEY_MAX + 1 },
  { "orania", KEY_MAX + 2 },
  { "bokvlei", KEY_MAX + 3 },
  { "calvinia", KEY_MAX + 4 },
  { "garies", KEY_MAX + 5 },
  { "goukamma", KEY_MAX + 6 },
  { "kariega", KEY_MAX + 7 },
  { "knysna", KEY_MAX + 8 },
  { "kakamas", KEY_MAX + 9 },
  { "keimoes", KEY_MAX + 10 },
  { "kuruman", KEY_MAX + 11 },
  { "kommetjie", KEY_MAX + 12 },
  { "konka", KEY_MAX + 13 },
  { "nababeep",  KEY_MAX + 14 },
  { "namies", KEY_MAX + 15 },
  { "okiep", KEY_MAX + 16 },
  { "prieska", KEY_MAX + 17 },
  { "sishen", KEY_MAX + 18 },
  { "eshowe", KEY_MAX + 19 },
  { "ifafa", KEY_MAX + 20 },
  { "inanda", KEY_MAX + 21 },
  { "isipingo", KEY_MAX + 22 },
  { "ixopo", KEY_MAX + 23 },
  { "jozini", KEY_MAX + 24 },
  { "madadeni", KEY_MAX + 25 },
  { "mkuze", KEY_MAX + 26 },
  { "nongoma", KEY_MAX + 27 },
  { "pongola", KEY_MAX + 28 },
  { "sezela", KEY_MAX + 29 },
  { "tongaat", KEY_MAX + 30 },
  { "tugela", KEY_MAX + 31 },
  { "ulundi", KEY_MAX + 32 },
  { "umlazi", KEY_MAX + 33 },
  { "umzinto", KEY_MAX + 34 },
  { "umzumbe", KEY_MAX + 35 },
  { "uvongo", KEY_MAX + 36 },
  { "zinkwazi", KEY_MAX + 37 },
#endif

  { "virt1", KEY_MAX + 1 },
  { "virt2", KEY_MAX + 2 },
  { "virt3", KEY_MAX + 3 },
  { "virt4", KEY_MAX + 4 },
  { "virt5", KEY_MAX + 5 },
  { "virt6", KEY_MAX + 6 },
  { "virt7", KEY_MAX + 7 },
  { "virt8", KEY_MAX + 8 },
  { "virt9", KEY_MAX + 9 },
  { "virt10", KEY_MAX + 10 },
  { "virt11", KEY_MAX + 11 },
  { "virt12", KEY_MAX + 12 },
  { "virt13", KEY_MAX + 13 },
  { "virt14", KEY_MAX + 14 },
  { "virt15", KEY_MAX + 15 },
  { "virt16", KEY_MAX + 16 },
  { "virt17", KEY_MAX + 17 },
  { "virt18", KEY_MAX + 18 },
  { "virt19", KEY_MAX + 19 },
  { "virt20", KEY_MAX + 20 },
  { "virt21", KEY_MAX + 21 },
  { "virt22", KEY_MAX + 22 },
  { "virt23", KEY_MAX + 23 },
  { "virt24", KEY_MAX + 24 },
  { "virt25", KEY_MAX + 25 },
  { "virt26", KEY_MAX + 26 },
  { "virt27", KEY_MAX + 27 },
  { "virt28", KEY_MAX + 28 },
  { "virt29", KEY_MAX + 29 },
  { "virt30", KEY_MAX + 30 },
  { "virt31", KEY_MAX + 31 },
  { "virt32", KEY_MAX + 32 },
  { "virt33", KEY_MAX + 33 },
  { "virt34", KEY_MAX + 34 },
  { "virt35", KEY_MAX + 35 },
  { "virt36", KEY_MAX + 36 },
  { "virt37", KEY_MAX + 37 },
  { "virt38", KEY_MAX + 38 },
  { "virt39", KEY_MAX + 39 },
  { "virt40", KEY_MAX + 40 },
  { "virt41", KEY_MAX + 41 },
  { "virt42", KEY_MAX + 42 },
  { "virt43", KEY_MAX + 43 },
  { "virt44", KEY_MAX + 44 },
  { "virt45", KEY_MAX + 45 },
  { "virt46", KEY_MAX + 46 },
  { "virt47", KEY_MAX + 47 },
  { "virt48", KEY_MAX + 48 },
  { "virt49", KEY_MAX + 49 },
  { "virt50", KEY_MAX + 50 },
  { "virt51", KEY_MAX + 51 },
  { "virt52", KEY_MAX + 52 },
  { "virt53", KEY_MAX + 53 },
  { "virt54", KEY_MAX + 54 },
  { "virt55", KEY_MAX + 55 },
  { "virt56", KEY_MAX + 56 },
  { "virt57", KEY_MAX + 57 },
  { "virt58", KEY_MAX + 58 },
  { "virt59", KEY_MAX + 59 },
  { "virt60", KEY_MAX + 60 },
  { "virt61", KEY_MAX + 61 },
  { "virt62", KEY_MAX + 62 },
  { "virt63", KEY_MAX + 63 },
  { "virt64", KEY_MAX + 64 },
  { "virt65", KEY_MAX + 65 },
  { "virt66", KEY_MAX + 66 },
  { "virt67", KEY_MAX + 67 },
  { "virt68", KEY_MAX + 68 },
  { "virt69", KEY_MAX + 69 },
  { "virt70", KEY_MAX + 70 },
  { "virt71", KEY_MAX + 71 },
  { "virt72", KEY_MAX + 72 },
  { "virt73", KEY_MAX + 73 },
  { "virt74", KEY_MAX + 74 },
  { "virt75", KEY_MAX + 75 },
  { "virt76", KEY_MAX + 76 },
  { "virt77", KEY_MAX + 77 },
  { "virt78", KEY_MAX + 78 },
  { "virt79", KEY_MAX + 79 },
  { "virt80", KEY_MAX + 80 },
  { "virt81", KEY_MAX + 81 },
  { "virt82", KEY_MAX + 82 },
  { "virt83", KEY_MAX + 83 },
  { "virt84", KEY_MAX + 84 },
  { "virt85", KEY_MAX + 85 },
  { "virt86", KEY_MAX + 86 },
  { "virt87", KEY_MAX + 87 },
  { "virt88", KEY_MAX + 88 },
  { "virt89", KEY_MAX + 89 },
  { "virt90", KEY_MAX + 90 },
  { "virt91", KEY_MAX + 91 },
  { "virt92", KEY_MAX + 92 },
  { "virt93", KEY_MAX + 93 },
  { "virt94", KEY_MAX + 94 },
  { "virt95", KEY_MAX + 95 },
  { "virt96", KEY_MAX + 96 },
  { "virt97", KEY_MAX + 97 },
  { "virt98", KEY_MAX + 98 },
  { "virt99", KEY_MAX + 99 },
  { "virt100", KEY_MAX + 100 },
  { "virt101", KEY_MAX + 101 },
  { "virt102", KEY_MAX + 102 },
  { "virt103", KEY_MAX + 103 },
  { "virt104", KEY_MAX + 104 },
  { "virt105", KEY_MAX + 105 },
  { "virt106", KEY_MAX + 106 },
  { "virt107", KEY_MAX + 107 },
  { "virt108", KEY_MAX + 108 },
  { "virt109", KEY_MAX + 109 },
  { "virt110", KEY_MAX + 110 },
  { "virt111", KEY_MAX + 111 },
  { "virt112", KEY_MAX + 112 },
  { "virt113", KEY_MAX + 113 },
  { "virt114", KEY_MAX + 114 },
  { "virt115", KEY_MAX + 115 },
  { "virt116", KEY_MAX + 116 },
  { "virt117", KEY_MAX + 117 },
  { "virt118", KEY_MAX + 118 },
  { "virt119", KEY_MAX + 119 },
  { "virt120", KEY_MAX + 120 },
  { "virt121", KEY_MAX + 121 },
  { "virt122", KEY_MAX + 122 },
  { "virt123", KEY_MAX + 123 },
  { "virt124", KEY_MAX + 124 },
  { "virt125", KEY_MAX + 125 },
  { "virt126", KEY_MAX + 126 },
  { "virt127", KEY_MAX + 127 },
  { "virt128", KEY_MAX + 128 },
  { "virt129", KEY_MAX + 129 },
  { "virt130", KEY_MAX + 130 },
  { "virt131", KEY_MAX + 131 },
  { "virt132", KEY_MAX + 132 },
  { "virt133", KEY_MAX + 133 },
  { "virt134", KEY_MAX + 134 },
  { "virt135", KEY_MAX + 135 },
  { "virt136", KEY_MAX + 136 },
  { "virt137", KEY_MAX + 137 },
  { "virt138", KEY_MAX + 138 },
  { "virt139", KEY_MAX + 139 },
  { "virt140", KEY_MAX + 140 },
  { "virt141", KEY_MAX + 141 },
  { "virt142", KEY_MAX + 142 },
  { "virt143", KEY_MAX + 143 },
  { "virt144", KEY_MAX + 144 },
  { "virt145", KEY_MAX + 145 },
  { "virt146", KEY_MAX + 146 },
  { "virt147", KEY_MAX + 147 },
  { "virt148", KEY_MAX + 148 },
  { "virt149", KEY_MAX + 149 },
  { "virt150", KEY_MAX + 150 },
  { "virt151", KEY_MAX + 151 },
  { "virt152", KEY_MAX + 152 },
  { "virt153", KEY_MAX + 153 },
  { "virt154", KEY_MAX + 154 },
  { "virt155", KEY_MAX + 155 },
  { "virt156", KEY_MAX + 156 },
  { "virt157", KEY_MAX + 157 },
  { "virt158", KEY_MAX + 158 },
  { "virt159", KEY_MAX + 159 },
  { "virt160", KEY_MAX + 160 },
  { "virt161", KEY_MAX + 161 },
  { "virt162", KEY_MAX + 162 },
  { "virt163", KEY_MAX + 163 },
  { "virt164", KEY_MAX + 164 },
  { "virt165", KEY_MAX + 165 },
  { "virt166", KEY_MAX + 166 },
  { "virt167", KEY_MAX + 167 },
  { "virt168", KEY_MAX + 168 },
  { "virt169", KEY_MAX + 169 },
  { "virt170", KEY_MAX + 170 },
  { "virt171", KEY_MAX + 171 },
  { "virt172", KEY_MAX + 172 },
  { "virt173", KEY_MAX + 173 },
  { "virt174", KEY_MAX + 174 },
  { "virt175", KEY_MAX + 175 },
  { "virt176", KEY_MAX + 176 },
  { "virt177", KEY_MAX + 177 },
  { "virt178", KEY_MAX + 178 },
  { "virt179", KEY_MAX + 179 },
  { "virt180", KEY_MAX + 180 },
  { "virt181", KEY_MAX + 181 },
  { "virt182", KEY_MAX + 182 },
  { "virt183", KEY_MAX + 183 },
  { "virt184", KEY_MAX + 184 },
  { "virt185", KEY_MAX + 185 },
  { "virt186", KEY_MAX + 186 },
  { "virt187", KEY_MAX + 187 },
  { "virt188", KEY_MAX + 188 },
  { "virt189", KEY_MAX + 189 },
  { "virt190", KEY_MAX + 190 },
  { "virt191", KEY_MAX + 191 },
  { "virt192", KEY_MAX + 192 },
  { "virt193", KEY_MAX + 193 },
  { "virt194", KEY_MAX + 194 },
  { "virt195", KEY_MAX + 195 },
  { "virt196", KEY_MAX + 196 },
  { "virt197", KEY_MAX + 197 },
  { "virt198", KEY_MAX + 198 },
  { "virt199", KEY_MAX + 199 },
  { "virt200", KEY_MAX + 200 },
  { "virt201", KEY_MAX + 201 },
  { "virt202", KEY_MAX + 202 },
  { "virt203", KEY_MAX + 203 },
  { "virt204", KEY_MAX + 204 },
  { "virt205", KEY_MAX + 205 },
  { "virt206", KEY_MAX + 206 },
  { "virt207", KEY_MAX + 207 },
  { "virt208", KEY_MAX + 208 },
  { "virt209", KEY_MAX + 209 },
  { "virt210", KEY_MAX + 210 },
  { "virt211", KEY_MAX + 211 },
  { "virt212", KEY_MAX + 212 },
  { "virt213", KEY_MAX + 213 },
  { "virt214", KEY_MAX + 214 },
  { "virt215", KEY_MAX + 215 },
  { "virt216", KEY_MAX + 216 },
  { "virt217", KEY_MAX + 217 },
  { "virt218", KEY_MAX + 218 },
  { "virt219", KEY_MAX + 219 },
  { "virt220", KEY_MAX + 220 },
  { "virt221", KEY_MAX + 221 },
  { "virt222", KEY_MAX + 222 },
  { "virt223", KEY_MAX + 223 },
  { "virt224", KEY_MAX + 224 },
  { "virt225", KEY_MAX + 225 },
  { "virt226", KEY_MAX + 226 },
  { "virt227", KEY_MAX + 227 },
  { "virt228", KEY_MAX + 228 },
  { "virt229", KEY_MAX + 229 },
  { "virt230", KEY_MAX + 230 },
  { "virt231", KEY_MAX + 231 },
  { "virt232", KEY_MAX + 232 },
  { "virt233", KEY_MAX + 233 },
  { "virt234", KEY_MAX + 234 },
  { "virt235", KEY_MAX + 235 },
  { "virt236", KEY_MAX + 236 },
  { "virt237", KEY_MAX + 237 },
  { "virt238", KEY_MAX + 238 },
  { "virt239", KEY_MAX + 239 },
  { "virt240", KEY_MAX + 240 },
  { "virt241", KEY_MAX + 241 },
  { "virt242", KEY_MAX + 242 },
  { "virt243", KEY_MAX + 243 },
  { "virt244", KEY_MAX + 244 },
  { "virt245", KEY_MAX + 245 },
  { "virt246", KEY_MAX + 246 },
  { "virt247", KEY_MAX + 247 },
  { "virt248", KEY_MAX + 248 },
  { "virt249", KEY_MAX + 249 },
  { "virt250", KEY_MAX + 250 },
  { "virt251", KEY_MAX + 251 },
  { "virt252", KEY_MAX + 252 },
  { "virt253", KEY_MAX + 253 },
  { "virt254", KEY_MAX + 254 },
  { "virt255", KEY_MAX + 255 },
  { "virt256", KEY_MAX + 256 },

  { NULL, -1 }
};

struct code_name mxk_rel_codes[] = {

  { "x", REL_X },
  { "y", REL_Y },
  { "z", REL_Z },
#ifdef REL_RX
  { "rx", REL_RX },
#endif
#ifdef REL_RY
  { "ry", REL_RY },
#endif
#ifdef REL_RZ
  { "rz", REL_RZ },
#endif
  { "hwheel", REL_HWHEEL },
  { "dial", REL_DIAL },
  { "wheel", REL_WHEEL },
  { "misc", REL_MISC },

  { NULL, -1 }
};

struct code_name mxk_abs_codes[] = {

  { "x", ABS_X },
  { "y", ABS_Y },
  { "z", ABS_Z },

  { "rx", ABS_RX },
  { "ry", ABS_RY },
  { "rz", ABS_RZ },

  { "throttle",   ABS_THROTTLE },
  { "rudder",     ABS_RUDDER },
  { "wheel",      ABS_WHEEL },
  { "gas",        ABS_GAS },
  { "brake",      ABS_BRAKE },

  { "hat0x",      ABS_HAT0X },
  { "hat0y",      ABS_HAT0Y },
  { "hat1x",      ABS_HAT1X },
  { "hat1y",      ABS_HAT1Y },
  { "hat2x",      ABS_HAT2X },
  { "hat2y",      ABS_HAT2Y },
  { "hat3x",      ABS_HAT3X },
  { "hat3y",      ABS_HAT3Y },

  { "pressure",   ABS_PRESSURE },
  { "distance",   ABS_DISTANCE },
  { "tiltx",      ABS_TILT_X },
  { "tilty",      ABS_TILT_Y },
  { "toolwidth",  ABS_TOOL_WIDTH },
  { "volume",     ABS_VOLUME },
  { "misc",       ABS_MISC },

  { NULL, -1 }
};
