/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>

#include <code-name.h>

#include "mxk.h"

struct mxk_degap *mxk_start_degap()
{
  struct mxk_degap *md;

  md = malloc(sizeof(struct mxk_degap));
  if(md == NULL){
    return NULL;
  }

  md->d_vector = 0;
  md->d_size = 0;
  md->d_max = 1;

  return md;
}

void mxk_stop_degap(struct mxk_degap *md)
{
  if(md){
    if(md->d_vector){
      free(md->d_vector);
      md->d_vector = NULL;
      md->d_size = 0;
    }
    free(md);
  }
}

#ifdef DEBUG
int mxk_dump_degap(struct mxk_degap *md, FILE *fp)
{
  int i;

  fprintf(fp, "degap: %d/%d: ", md->d_size, md->d_max);

  for(i = 0; i < md->d_size; i++){
    if(md->d_vector[i] >= 0){
      fprintf(fp, " %d", md->d_vector[i]);
    } else {
      fprintf(fp, " .");
    }
  }

  fprintf(fp, "\n");

  return 0;
}
#endif

char *mxk_reverse_degap(struct mxk_degap *md, int compact)
{
  int i;
  char *ptr;

  for(i = 0; i < md->d_size; i++){
    if(md->d_vector[i] == compact){
      ptr = code_to_name(mxk_key_codes, i);
      return ptr;
    }
  }

  return NULL;
}

int mxk_insert_degap(struct mxk_degap *md, int value)
{
  int *vector;
  int i;

#ifdef DEBUG
  fprintf(stderr, "degap: about to insert %d\n", value);
#endif

  if(value < 0){
    return -1;
  }

  if(md->d_max >= 0xffff){
#ifdef DEBUG
    fprintf(stderr, "degap: error: too many entries %d\n", md->d_max);
#endif
    return -1;
  }

#ifdef DEBUG
  if(value > 0xffff){
    fprintf(stderr, "degap: warning: inserting really large value %d\n", value);
  }
#endif

  if(md->d_size <= value){
    vector = realloc(md->d_vector, sizeof(int) * (value + 1));
    if(vector == NULL){
      return -1;
    }
    for(i = md->d_size; i <= value; i++){
      vector[i] = (-1);
    }
    md->d_vector = vector;
    md->d_size = value + 1;
  }

  if(md->d_vector[value] < 0){
    md->d_vector[value] = md->d_max++;
  }

#ifdef DEBUG
  fprintf(stderr, "degap: mapping %d to %d\n", value, md->d_vector[value]);
#endif

  return md->d_vector[value];
}

int mxk_lookup_degap(struct mxk_degap *md, int value)
{
  if((value < 0) || (md->d_size <= value)){
    return -1;
  }
  return md->d_vector[value];
}
