/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <loop.h>
#include <word-work.h>

#include "mxk.h"

int mxk_linkup_find(struct list_loop *ls, char *sname, int sid, char *tname, int tid)
{
  struct task_loop *tx;
  struct mxk_find *mf;
  struct mxk_global *mg;
  int ttype, stype;
  int (*insert)(void *server, int type, int id);
  void *ptr;

  mg = global_loop(ls);

  mf = mxk_test_find(mg, tname);
  if(mf == NULL){
    return -1;
  }
  ttype = mf->f_type;

  mf = mxk_test_find(mg, sname);
  if(mf == NULL){
    return -1;
  }
  stype = mf->f_type;
  insert = mf->f_insert;

#ifdef DEBUG
  fprintf(stderr, "linkup: source=%d:%d, target=%d:%d\n", stype, sid, ttype, tid);
#endif

  for(tx = find_task_tag_loop(ls, NULL, NULL, stype, sid); tx; tx = find_task_tag_loop(ls, NULL, tx, stype, sid)){
    ptr = user_loop(ls, tx);
    if(ptr){
      if((*insert)(ptr, ttype, tid)){
        return -1;
      }
    }
  }

  return 0;
}

int mxk_enroll_find(struct list_loop *ls, char *name, int flags, int count, int (*insert)(void *server, int type, int id))
{
  struct mxk_find *mf;
  struct mxk_global *mg;
  char *copy;
  int type;

  mg = global_loop(ls);
  if(mg == NULL){
    return -1;
  }

  mf = mxk_test_find(mg, name);
  if(mf){
    return mf->f_type;
  }

  copy = strdup(name);
  if(copy == NULL){
    return -1;
  }

  type = claim_task_name_loop(ls, count, 1, name);
  if(type <= 0){
    free(copy);
    return -1;
  }

  mf = realloc(mg->g_table, sizeof(struct mxk_find) * (mg->g_tsize + 1));
  if(mf == NULL){
    free(copy);
    return -1;
  }

  mg->g_table = mf;
  mf = &(mg->g_table[mg->g_tsize]);
  mg->g_tsize++;

#ifdef DEBUG
  fprintf(stderr, "enroll: size now %d\n", mg->g_tsize);
  if((flags & MXK_EVT_SOURCE) && (insert == NULL)){
    fprintf(stderr, "warning: won't be able to use <%s> as source, no insertion function given\n", name);
  }
#endif

  mf->f_name = copy;
  mf->f_type = type;
  mf->f_flags = flags;
  mf->f_insert = insert;

  return type;
}

void mxk_end_find(struct mxk_global *mg)
{
  int i;
  struct mxk_find *mf;

  if(mg && mg->g_table){
    for(i = 0; i < mg->g_tsize; i++){
      mf = &(mg->g_table[i]);
      if(mf->f_name){
        free(mf->f_name);
        mf->f_name = NULL;
      }
    }
    free(mg->g_table);
    mg->g_table = NULL;
  }

}

struct mxk_find *mxk_test_find(struct mxk_global *mg, char *name)
{
  struct mxk_find *mf;
  int i;

#ifdef DEBUG
  if((mg == NULL) || (name == NULL)){
    fprintf(stderr, "find: logic problems: NULL parameters\n");
    abort();
  }
#endif

  for(i = 0; i < mg->g_tsize; i++){
    mf = &(mg->g_table[i]);
    if(!strcmp(mf->f_name, name)){
      return mf;
    }
  }

  return NULL;
}

#if 0
/* UNUSED, replaced by libloop supplied function */
int mxk_type_find(struct mxk_global *mg, char *name)
{
  struct mxk_find *mf;

  mf = mxk_test_find(mg, name);
  if(mf == NULL){
#ifdef DEBUG
    fprintf(stderr, "type: unable to find <%s>\n", name);
    abort();
#else
    return 0;
#endif
  }

  return mf->f_type;
}
#endif

int mxk_show_find_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int fd;
  int i;
  struct mxk_global *mg;
  struct mxk_find *mf;
  struct control_loop *cl;

  fd = fd_loop(ls, tl);
  cl = user_loop(ls, tl);
  mg = global_loop(ls);

  for(i = 0; i < mg->g_tsize; i++){
    mf = &(mg->g_table[i]);
#ifdef DEBUG
    fprintf(stderr, "show: mf[%d], mf=%p, name=%p type=%d\n", i, mf, mf->f_name, mf->f_type);
#endif
    print_control(cl, fd, 1, "%s(%d):\tsource=%s target=%s\n",
        (mf->f_name ? mf->f_name : "<no name>"),
         mf->f_type,
        (mf->f_flags & MXK_EVT_SOURCE) ? "yes," : "no, ",
        (mf->f_flags & MXK_EVT_TARGET) ? "yes" : "no");
  }

  print_control(cl, fd, 1, "%d %s\n", mg->g_tsize, (mg->g_tsize == 1) ? "type" : "types");

  return 0;
}

int mxk_linkup_find_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int idx, fd;
  char source[MXK_NAME_SIZE], target[MXK_NAME_SIZE];
  int sid, tid;
  char *ptr;
  struct control_loop *cl;

  fd = fd_loop(ls, tl);
  cl = user_loop(ls, tl);

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, fd, 1, "no parameters, need two parameters\n");
    return 0;
  }

  if(load_word(source, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, fd, 1, "first parameter seems too long\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    print_control(cl, fd, 1, "one parameter, need two parameters\n");
    return 0;
  }

  if(load_word(target, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, fd, 1, "second parameter seems too long\n");
    return 0;
  }

  ptr = strchr(source, ':');
  if(ptr){
    ptr[0] = '\0';
    ptr++;
    sid = atoi(ptr);
  } else {
    sid = 0;
  }

  ptr = strchr(target, ':');
  if(ptr){
    ptr[0] = '\0';
    ptr++;
    tid = atoi(ptr);
  } else {
    tid = 0;
  }

  if(mxk_linkup_find(ls, source, sid, target, tid)){
    print_control(cl, fd, 1, "unable to link %s:%d to %s:%d\n", source, sid, target, tid);
    return 0;
  }

  return 0;
}

