/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>

#include <word-work.h>

#include "mxk.h"

static void mxk_stop_recursive_key(struct mxk_key *mk);

int *mxk_parse_vector_key(char *args, int offset)
{
  char string[MXK_NAME_SIZE];
  int *vector, *tmp;
  int idx, size, i;

  vector = NULL;

  idx = start_word(args, 0);
  if(idx < 0){
#ifdef DEBUG
    fprintf(stderr, "parse key: unable to retrieve first argument\n");
#endif
    return NULL;
  }

  for(i = 0; i < offset; i++){
    idx = next_word(args, idx);
    if(idx < 0){
#ifdef DEBUG
    fprintf(stderr, "parse key: unable to skip argument %d\n", i);
#endif
      return NULL;
    }
  }

  size = 0;
  vector = malloc(sizeof(int)); /* for terminator */
  if(vector == NULL){
#ifdef DEBUG
    fprintf(stderr, "parse key: unable to allocate initial word\n");
#endif
    return NULL;
  }

  while(idx >= 0){
    if(load_word(string, MXK_NAME_SIZE - 1, args, idx) >= 0){
      tmp = realloc(vector, sizeof(int) * (size + 2)); /* for final entry */
      if(tmp){
        vector = tmp;
        vector[size] = name_to_code(mxk_key_codes, string);
        if(vector[size] >= 0){
          size++;
          idx = next_word(args, idx);
          continue; /* WARNING: unusual control flow */
        }
      }

    }
#ifdef DEBUG
    fprintf(stderr, "parse key: unable to process word at %d in <%s>\n", idx, args);
#endif
    if(vector){
      free(vector);
    }
    return NULL;
  }

  vector[size] = (-1);

  return vector;
}

struct mxk_key *mxk_start_key()
{
  struct mxk_key *mk;

  mk = malloc(sizeof(struct mxk_key));
  if(mk == NULL){
    return NULL;
  }

  mk->k_size = 0;
  mk->u.k_value = (-1);
  mk->k_priority = 0;
  mk->k_depth = 0;

  return mk;
}

static void mxk_stop_recursive_key(struct mxk_key *mk)
{
  int i;

  if(mk == NULL){
    return;
  }

  if(mk->k_size){
    for(i = 0; i < mk->k_size; i++){
      mxk_stop_recursive_key(&(mk->u.k_next[i]));
    }
    free(mk->u.k_next);
    mk->u.k_next = NULL;
    mk->k_size = 0;
  }
}

void mxk_stop_key(struct mxk_key *mk)
{
  if(mk){
    mxk_stop_recursive_key(mk);
    free(mk);
  }
}

int mxk_insert_key(struct mxk_key *mk, int *vector, int value, int priority, int depth)
{
  struct mxk_key *tk, *xk, *yk;
  int i, j, start;

  if(vector[0] < 0){
    return -1;
  }

  switch(depth){
    case MXK_QUEUE_OLD_PRESS   :
    case MXK_QUEUE_NEW_RELEASE :
    case MXK_QUEUE_OLD_RELEASE :
      break;
    default :
#ifdef DEBUG
      fprintf(stderr, "key: depth out of range %d\n", depth);
#endif
      return -1;
  }

#ifdef DEBUG
  if(mk == NULL){
    fprintf(stderr, "key: logic problem: no root\n");
    abort();
  }
#endif
  tk = mk;

#ifdef DEBUG
  fprintf(stderr, "key:");
#endif

  for(i = 0; vector[i] >= 0; i++){
#ifdef DEBUG
    if(vector[i] == 0){
      fprintf(stderr, "key insert: logic problem: inserted key has to be larger than zero\n");
    }
#endif

    if(tk->k_size <= vector[i]){

#ifdef DEBUG
      fprintf(stderr, " (%d->%d)", tk->k_size, vector[i] + 1);
#endif

      if(tk->k_size){
        yk = realloc(tk->u.k_next, sizeof(struct mxk_key) * (vector[i] + 1));
        if(yk == NULL){
          return -1;
        }
        tk->u.k_next = yk;
        start = tk->k_size;
      } else {
        yk = malloc(sizeof(struct mxk_key) * (vector[i] + 1));
        if(yk == NULL){
          return -1;
        }
        xk = &(yk[0]);

        xk->k_size = 0;
        xk->u.k_value = tk->u.k_value;
        xk->k_priority = tk->k_priority;
        xk->k_depth = tk->k_depth;

        tk->u.k_next = yk;
        start = 1; /* don't clobber first special */
      }

      /* WARNING: shouldn't we start with start rather than k_size ? */
      for(j = start; j <= vector[i]; j++){
        xk = &(tk->u.k_next[j]);
        xk->k_size = 0;
        xk->k_priority = 0;
        xk->k_depth = 0;
        xk->u.k_value = (-1);
      }

      tk->k_size = vector[i] + 1;
    }

    tk = &(tk->u.k_next[vector[i]]);
#ifdef DEBUG
    fprintf(stderr, " <%p>%d", tk, vector[i]);
#endif
  }

  if(tk->k_priority > priority){
#ifdef DEBUG
    fprintf(stderr, " = refusing to clobber existing element of priority %d\n", tk->k_priority);
#endif
    return 1;
  }

  tk->k_priority = priority;
  tk->k_depth = depth;

  if(tk->k_size){
    xk = &(tk->u.k_next[0]);

#ifdef DEBUG
    if(xk->k_size){
      fprintf(stderr, "insert: logic problem: index 0 has nonzero size\n");
      abort();
    }
#endif

    xk->u.k_value = value;
    xk->k_priority = priority;
    xk->k_depth = depth;

  } else {
#ifdef DEBUG
    fprintf(stderr, " = %d (terminal)\n", value);
#endif
    tk->u.k_value = value;
  }

  return 0;
}

#ifdef STANDALONE

#define BUFFER  1024

#include "codes.c"

int main()
{
  char buffer[BUFFER];
  int *vector;
  unsigned int offset, i;

  offset = 0;
  printf("%d> ", offset);
  fflush(stdout);

  while(fgets(buffer, BUFFER - 1, stdin)){

    vector = mxk_parse_vector_key(buffer, offset);

    if(vector){
      printf("vector:");
      for(i = 0; vector[i] >= 0; i++){
        printf(" %d", vector[i]);
      }
      printf("\n");
    } else {
      printf("vector empty\n");
    }

    offset++;
    offset = offset % 4;

    printf("%d> ", offset);
    fflush(stdout);

  }

  return 0;
}
#endif
