/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <netdb.h>
#include <sched.h>
#include <ctype.h>
#include <signal.h>

#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>

#include "loop.h"
#include "priv.h"

static void finish_loop(struct list_loop *ls, struct task_loop *tl);

#if 0
static volatile int notice = 0;

static void handler(int number)
{
  if(notice == 0){
    notice = number;
  }
}

/* WARNING: no signal handler installed yet */

sag.sa_handler = handler;
sag.sa_flags = 0;
sigfillset(&(sag.sa_mask));
#endif

/******************** main loop and friends ******************************************/

int main_loop(struct list_loop *ls)
{
  return main_loop_mode(ls, LO_FAVOUR_BALANCE);
}

int main_loop_mode(struct list_loop *ls, int mode)
{
  int fcpu, fio, icpu, iio, bounded, skips, mask, result;
  fd_set fsr, fsw;
  struct timeval wi, wt;
  struct task_loop *tl, *tp;
  struct job_loop *jz;

  jz = &(ls->l_job[0]);
  gettimeofday(&(ls->l_now), NULL);
  skips = 0;

  icpu = (mode == LO_FAVOUR_IO)  ? 0 : 1;
  iio = (mode == LO_FAVOUR_CPU)  ? 0 : 1;

  for(ls->l_up = 1; ls->l_up;){

    /* decide how to wait for tasks */
    FD_ZERO(&fsr);
    FD_ZERO(&fsw);
    bounded = 0;
    fcpu = 0;
    fio = 0;
    tl = ls->l_head;
    ls->l_mfd = (-1);
    while(tl){
      if(tl->t_mask & LO_WORK_MASK){
        fcpu += icpu;
        bounded = 1;
        wt = ls->l_now;
      }
      if(tl->t_mask & LO_READ_MASK){
#ifdef DEBUG
        if(tl->t_fd < 0){
          fprintf(stderr, "warning: loading invalid read file descriptor\n");
        }
#endif
        fio += iio;
        if(tl->t_fd > ls->l_mfd){
          ls->l_mfd = tl->t_fd;
        }
        FD_SET(tl->t_fd, &fsr);
      }
      if(tl->t_mask & LO_WRITE_MASK){
#ifdef DEBUG
        if(tl->t_fd < 0){
          fprintf(stderr, "warning: loading invalid write file descriptor\n");
        }
#endif
        fio += iio;
        if(tl->t_fd > ls->l_mfd){
          ls->l_mfd = tl->t_fd;
        }
        FD_SET(tl->t_fd, &fsw);
      }
      if(tl->t_mask & LO_MORE_MASK){
        if((tl->t_max <= 0) || (tl->t_have[LO_OUTQ] < tl->t_max[LO_OUTQ])){
          fcpu += icpu;
          bounded = 1;
          wt = ls->l_now;
        }
      }
      if(tl->t_mask & LO_WAIT_MASK){
        if(tl->t_new){
          fcpu += icpu;
          bounded = 1;
          wt = ls->l_now;
        }
      }
      if(tl->t_stop != LO_STOP_NOT){ /* about to be destroyed */
#ifdef DEBUG
        fprintf(stderr, "core: can schedule destruction\n");
#endif
        fcpu += icpu;
        bounded = 1;
        wt = ls->l_now;
      }
      if(tl->t_mask & LO_TIME_MASK){
        /* if first timer or smaller than previous timers */
        if((bounded == 0) || (cmp_time_loop(&(tl->t_timeout), &wt) < 0)){
          wt = tl->t_timeout;
          bounded = 1;
        } 
      }
      tl = tl->t_next;
    }

    if((jz->j_used > 0) && (bounded || (ls->l_mfd >= 0))){

      if(skips < fcpu){ /* if FAVOUR cpu, fio always zero, so unless fcpu zero, this always taken */
        FD_ZERO(&fsr);
        FD_ZERO(&fsw);
        skips += fio;
        result = 0;
      } else { /* if FAVOUR io, fcpu always zero, so this always taken */
        skips = 0;
        if(bounded){
          sub_time_loop(&wi, &wt, &(ls->l_now));
#ifdef DEBUG
          fprintf(stderr, "core: wait interval %ld.%06ld for select\n", wi.tv_sec, wi.tv_usec);
#endif
          result = select(ls->l_mfd + 1, &fsr, &fsw, NULL, &wi);
        } else {
          /* unbounded wait */
          result = select(ls->l_mfd + 1, &fsr, &fsw, NULL, NULL);
        }
      }
    } else {
      result = 0;
      ls->l_up = 0;
    }

    gettimeofday(&(ls->l_now), NULL);

    switch(result){
      case -1 :
        switch(errno){
          case EAGAIN :
          case EINTR  :
            break;
          default :
            ls->l_up = 0;
            break;
        }
        break;
    }

    /* decide on tasks to run */
    tl = ls->l_head;
    tp = NULL;
    while(tl){
      mask = tl->t_mask & LO_WORK_MASK;
      if(tl->t_mask & LO_READ_MASK){
        if(FD_ISSET(tl->t_fd, &fsr)){
          mask |= LO_READ_MASK;
        }
      }
      if(tl->t_mask & LO_WRITE_MASK){
        if(FD_ISSET(tl->t_fd, &fsw)){
          mask |= LO_WRITE_MASK;
        }
      }
      if(tl->t_mask & LO_MORE_MASK){
        if((tl->t_max <= 0) || (tl->t_have[LO_OUTQ] < tl->t_max[LO_OUTQ])){
#ifdef DEBUG
          fprintf(stderr, "core: can queue more links: have=%d, max=%d\n", tl->t_have[LO_OUTQ], tl->t_max[LO_OUTQ]);
#endif
          mask |= LO_MORE_MASK;
        }
      }
      if(tl->t_mask & LO_WAIT_MASK){
        if(tl->t_new){
          mask |= LO_WAIT_MASK;
        }
      }
      if(tl->t_mask & LO_TIME_MASK){
#ifdef DEBUG
        fprintf(stderr, "core: now %lu.%06lu, timeout=%lu.%06lu, mask=%02x\n", ls->l_now.tv_sec, ls->l_now.tv_usec, tl->t_timeout.tv_sec, tl->t_timeout.tv_usec, mask);
#endif
        if(cmp_time_loop(&(tl->t_timeout), &(ls->l_now)) <= 0){
          mask |= LO_TIME_MASK;
          /* restart for next time */
          add_time_loop(&(tl->t_timeout), &(ls->l_now), &(tl->t_interval));
        }
      }
      if(tl->t_stop == LO_STOP_PREPARE){ 
        tl->t_stop = LO_STOP_FINISH;
        if(tl->t_mask & LO_STOP_MASK){
          mask |= LO_STOP_MASK;
        }
      }

#ifdef DEBUG
      fprintf(stderr, "run %p: fd=%d, new=%p%c, mask=0x%02x, req=0x%02x stop=%d %s\n", tl, tl->t_fd, tl->t_new, (tl->t_new == tl->t_links[LO_INQ]) ? '=' : '+', mask, tl->t_mask, tl->t_stop, tl->t_text);
#endif

      if(mask){
#ifdef DEBUG
        if(tl->t_call == NULL){
          fprintf(stderr, "core: logic failure - null call\n");
          abort();
        }
#endif
        (*(tl->t_call))(ls, tl, mask);
      }

      if(tl->t_stop == LO_STOP_FINISH){
#ifdef DEBUG
        fprintf(stderr, "core: deleting entry\n");
#endif
        finish_loop(ls, tl);
        if(tp){
          tp->t_next = tl->t_next;
          tl->t_next = ls->l_spare;
          ls->l_spare = tl;
          tl = tp->t_next;
        } else {
          ls->l_head = tl->t_next;
          tl->t_next = ls->l_spare;
          ls->l_spare = tl;
          tl = ls->l_head;
        }
        ls->l_spare_size++;
#ifdef DEBUG
        if(jz->j_used < 0){
          fprintf(stderr,  "core: unreasonable deletion: %d\n", jz->j_used);
        }
#endif
      } else { /* advance normally */
        tp = tl;
        tl = tl->t_next;
      }
    }

#ifdef DEBUG
    fprintf(stderr, "--- ran tasks, next iteration ---\n");
#endif

  }

  return -1;
}

#define LO_INIT_TASKS  4
#define LO_INIT_LINKS 16

struct list_loop *startup_loop(void *data, int tasks, int links)
{
  struct list_loop *ls;
  struct task_loop *tl;
  struct link_loop *nk;
  struct pool_loop *pl;
  struct job_loop *jl;
  int ti, ki;

  signal(SIGPIPE, SIG_IGN);

  ls = malloc(sizeof(struct list_loop));
  if(ls == NULL){
    return NULL;
  }

  ls->l_data = data;
  ls->l_mfd = (-1);

  ls->l_spare = NULL;
  ls->l_head = NULL;

  ls->l_pool = NULL;
  ls->l_pool_size = 0;

  ls->l_job = NULL;
  ls->l_job_size = 0;

  gettimeofday(&(ls->l_now), NULL);

  ls->l_start = ls->l_now.tv_sec;
  ls->l_was = 0;

  for(ti = 0; ti < LO_INIT_TASKS; ti++){
    tl = malloc(sizeof(struct task_loop));
    if(tl){
      tl->t_call = NULL;
      tl->t_text[0] = '\0';
      tl->t_next = ls->l_spare;
      ls->l_spare = tl;
    }
  }
  ls->l_spare_size = ti;

  jl = malloc(sizeof(struct job_loop));
  if(jl){
    jl->j_type = 0;
    jl->j_used = 0;
    jl->j_max = (tasks && (LO_INIT_TASKS > tasks)) ? LO_INIT_TASKS : tasks;
    jl->j_name = NULL;

    ls->l_job = jl;
    ls->l_job_size = 1;
  }


  pl = malloc(sizeof(struct pool_loop));
  if(pl){
    pl->p_free = NULL;
    pl->p_empty = NULL;
    pl->p_type = 0;

    pl->p_have = 0;
    pl->p_used = 0;
    pl->p_max = (links && (LO_INIT_LINKS > links)) ? LO_INIT_LINKS : links;
    pl->p_pool = LO_INIT_LINKS;

    pl->p_name = NULL;

    for(ki = 0; ki < LO_INIT_LINKS; ki++){
      nk = malloc(sizeof(struct link_loop));
      if(nk){
        nk->k_type = 0;
        nk->k_data = NULL;
        nk->k_task[LO_INQ] = nk->k_task[LO_OUTQ] = NULL;
        nk->k_back[LO_INQ] = nk->k_back[LO_OUTQ] = NULL;
        nk->k_next[LO_INQ] = nk->k_next[LO_OUTQ] = NULL;

        pl->p_free = enqueue_loop(pl->p_free, nk, LO_FREEQ);
      }
    }
    pl->p_have = ki;

    ls->l_pool = pl;
    ls->l_pool_size = 1;
  }

  if((ls->l_spare == NULL) || (ls->l_pool == NULL) || (ls->l_job == NULL)){
    shutdown_loop(ls);
    return NULL;
  }

  return ls;
}

int shutdown_loop(struct list_loop *ls)
{
  struct task_loop *tl;
  struct link_loop *nk;
  struct pool_loop *pl;
  struct job_loop *jl;
  int i;

  if(ls == NULL){
    return -1;
  }

  tl = ls->l_head;
  while(tl){
    stop_loop(ls, tl);

    tl->t_stop = LO_STOP_FINISH;

    unwind_loop(ls, tl);
    finish_loop(ls, tl);

    ls->l_head = tl->t_next;
    tl->t_next = ls->l_spare;
    ls->l_spare = tl;
    tl = ls->l_head;

    ls->l_spare_size++;
  }

  while(ls->l_spare){
    tl = ls->l_spare;
    ls->l_spare = tl->t_next;
    tl->t_next = NULL;
    ls->l_spare_size--;
    free(tl);
  }

  if(ls->l_pool){
    for(i = 0; i < ls->l_pool_size; i++){
      pl = &(ls->l_pool[i]);
      while(pl->p_free){
        nk = pl->p_free;
        pl->p_free = dequeue_loop(pl->p_free, nk, LO_FREEQ);
        if(nk->k_data){
          if(pl->p_empty){
            (*(pl->p_empty))(ls, nk->k_data);
          }
          nk->k_data = NULL;
          nk->k_type = 0;
        }
        free(nk);
      }
#if DEBUG
      if(pl->p_used > 0){
        fprintf(stderr, "shutdown: warning: %d links of type %d still in use\n", pl->p_used, i);
      }
#endif
      pl->p_have = 0;
      pl->p_max = 0;
      pl->p_pool = 0;
      if(pl->p_name){
        free(pl->p_name);
        pl->p_name = NULL;
      }
    }
    free(ls->l_pool);
    ls->l_pool = NULL;
    ls->l_pool_size = 0;
  }

  if(ls->l_job){
    for(i = 0; i < ls->l_job_size; i++){
      jl = &(ls->l_job[i]);
      if(jl->j_name){
        free(jl->j_name);
        jl->j_name = NULL;
      }
    }
    free(ls->l_job);
    ls->l_job = NULL;
    ls->l_job_size = 0;
  }

#ifdef DEBUG
  if(ls->l_spare_size){
    fprintf(stderr, "core: logic failure, remaining spare=%d\n", ls->l_spare_size);
  }
#endif

  ls->l_spare_size = 0;

  free(ls);

  return 0;
}

/******************** loop enrollment ************************************************/

struct task_loop *add_loop(struct list_loop *ls, int fd, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type)
{
  return add_tag_loop(ls, fd, data, mask, call, tv, type, 0);
}

struct task_loop *add_tag_loop(struct list_loop *ls, int fd, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type, int tag)
{
  struct task_loop *tx, *tl;
  struct job_loop *jl, *jz;
  
  tl = ls->l_spare;

#ifdef DEBUG
  if(call == NULL){
    fprintf(stderr, "core: usage problem: no function call\n");
    abort();
  }

  if(tl == NULL){
    fprintf(stderr, "core: logic problem: no spare left\n");
    abort();
  }

  if((type < 0) || (type > ls->l_job_size)){
    fprintf(stderr, "core: logic problem: task type too large\n");
    abort();
  }
#endif

  jl = NULL;
  if(type > 0){
    jl = &(ls->l_job[type]);
    if((jl->j_max >= 0) && (jl->j_max <= jl->j_used)){
#ifdef DEBUG
      fprintf(stderr, "core: over local task quota %d\n", jl->j_max);
#endif
      return NULL;
    }
  }

  /* zero entry is special, it counts all entries */
  jz = &(ls->l_job[0]);
  if((jz->j_max >= 0) && (jz->j_max <= jz->j_used)){
#ifdef DEBUG
    fprintf(stderr, "core: over global task quota %d\n", jz->j_max);
#endif
    return NULL;
  }

  tl->t_call = call;
  tl->t_mask = mask;
  tl->t_stop = LO_STOP_NOT;
  tl->t_fd = fd;
  tl->t_data = data;
  tl->t_type = type;
  tl->t_tag = tag; /* TODO: maybe if tag == -1, autogenerated uniq tag */

  tl->t_links[LO_INQ] = tl->t_links[LO_OUTQ] = NULL;
  tl->t_new = NULL;
  tl->t_have[LO_INQ] = tl->t_have[LO_OUTQ] = 0;
  tl->t_max[LO_INQ] = tl->t_max[LO_OUTQ] = 0;

  if(tv){
    tl->t_interval.tv_sec = tv->tv_sec;
    tl->t_interval.tv_usec = tv->tv_usec;

    add_time_loop(&(tl->t_timeout), &(ls->l_now), &(tl->t_interval));
  }

  tx = tl->t_next;
  if(tx == NULL){
    tx = malloc(sizeof(struct task_loop));
    if(tx){
      tx->t_call = NULL;
      tx->t_text[0] = '\0';
      tx->t_next = NULL;
      tl->t_next = tx;
      ls->l_spare_size++;
    }
  }

  if(tx == NULL){
    return NULL; /* bomb out, caller has to close fd and sort out data */
  }

  if(fd > ls->l_mfd){
    ls->l_mfd = fd;
  }

#ifdef DEBUG
  if(jl){
    fprintf(stderr, "core: added type %d, number %d(%d)\n", type, jl->j_used, jz->j_used);
  } else {
    fprintf(stderr, "core: added anonymous, number %d\n", jz->j_used);
  }
#endif

  ls->l_spare_size--;

  ls->l_spare = tx;
  tl->t_next = ls->l_head;
  ls->l_head = tl;

  if(jl){
    jl->j_used = jl->j_used + 1;
  }
  jz->j_used = jz->j_used + 1;

  return tl;
}

/**********************************************************************/

static void finish_loop(struct list_loop *ls, struct task_loop *tl)
{
  struct job_loop *jl, *jz;
#ifdef DEBUG
  if(tl->t_stop < LO_STOP_FINISH){
    fprintf(stderr, "finish: logic problem: finishing diehard task\n");
    abort();
  }
  if(tl->t_links[LO_OUTQ]){
    fprintf(stderr, "finish: logic problem: finishing with outgoing tasks\n");
    abort();
  }
  if(tl->t_type >= ls->l_job_size){
    fprintf(stderr, "finish: logic problem: finishing with unknown task type\n");
    abort();
  }
#endif

  while(tl->t_links[LO_INQ]){
    discard_link_loop(ls, tl->t_links[LO_INQ]);
  }

  tl->t_call = NULL;
  tl->t_text[0] = '\0';

  if(tl->t_type > 0){
    jl = &(ls->l_job[tl->t_type]);
    jl->j_used = jl->j_used - 1;
  }
  jz = &(ls->l_job[0]);
  jz->j_used = jz->j_used - 1;

  tl->t_type = 0;

  if(tl->t_fd >= 0){
    close(tl->t_fd);
    tl->t_fd = (-1);
  }
#if 0 /* clean up, maybe too clever */
  if(tl->t_data){
    free(tl->t_data);
    tl->t_data = NULL;
  }
#endif
}

int unwind_loop(struct list_loop *ls, struct task_loop *tl)
{
  /* WARNING: can be called inside shutdown and mainloop */
#ifdef DEBUG
  fprintf(stderr, "unwind: pop task %p\n", tl);
  if(tl->t_stop < LO_STOP_FINISH){
    fprintf(stderr, "unwind: logic problem: unwinding task not yet finished\n");
    abort();
  }
#endif
  if((tl->t_mask & LO_STOP_MASK) && (tl->t_call != NULL)){
    return (*(tl->t_call))(ls, tl, LO_STOP_MASK);
  } else {
    return 0;
  }
}

int stop_loop(struct list_loop *ls, struct task_loop *tl)
{

#ifdef DEBUG
  fprintf(stderr, "stop: stopping %p\n", tl);
  if(tl == NULL){
    return -1;
  }
#endif

  if(tl->t_stop != LO_STOP_NOT){
    return 0;
  }

  tl->t_stop = LO_STOP_PREPARE;
  ls->l_was++;

  while(tl->t_links[LO_OUTQ]){
    detach_link_loop(ls, tl->t_links[LO_OUTQ]);
  }

  return 0;
}

/***********************************************************************/

