/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "loop.h"
#include "priv.h"

/* TODO: add link quotas per type */

/* create a link between two tasks */

struct link_loop *send_link_loop(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type)
{
  struct link_loop *nk;
  struct pool_loop *pl, *pz;
  int i, t, m;

  if((tl == NULL) || (tx == NULL) || (tl == tx) || (tx->t_stop != LO_STOP_NOT)){
    return NULL;
  }

  if((tl->t_max[LO_OUTQ] > 0) && (tl->t_have[LO_OUTQ] >= tl->t_max[LO_OUTQ])){
#ifdef DEBUG
    fprintf(stderr, "send_link: sender over quota %d\n", tl->t_max[LO_OUTQ]);
#endif
    return NULL;
  }

  if((tx->t_max[LO_INQ] > 0) && (tx->t_have[LO_INQ] >= tx->t_max[LO_INQ])){
#ifdef DEBUG
    fprintf(stderr, "send_link: receiver over quota %d\n", tx->t_max[LO_INQ]);
#endif
    return NULL;
  }

#ifdef DEBUG
  if((type < 0) || (type >= ls->l_pool_size)){
    fprintf(stderr, "major logic failure: requested link type out of range\n");
    abort();
  }
#endif

  pl = NULL;
  if(type){
    pl = &(ls->l_pool[type]);
    if((pl->p_max >= 0) && (pl->p_max <= pl->p_used)){
#ifdef DEBUG
      fprintf(stderr, "link: over local quota %d\n", pl->p_max);
#endif
      return NULL;
    }
  }

  /* WARNING: pz is special; counts *all* entries, not just [0] */
  pz = &(ls->l_pool[0]);
  if((pz->p_max >= 0) && (pz->p_max <= pz->p_used)){
#ifdef DEBUG
    fprintf(stderr, "link: over global quota %d\n", pz->p_max);
#endif
    return NULL;
  }

  nk = NULL;
  /* first choice: grab cached link of same type */
  if(pl && (pl->p_free)){ 
    nk = pl->p_free;
    pl->p_free = dequeue_loop(pl->p_free, nk, LO_FREEQ);
    pl->p_have = pl->p_have - 1;
    pz->p_have = pz->p_have - 1;
  } 

  /* second choice: grab empty link */
  if((nk == NULL) && (pz->p_free)){ 
    nk = pz->p_free;
    pz->p_free = dequeue_loop(pz->p_free, nk, LO_FREEQ);
    nk->k_data = NULL;
    pz->p_have = pz->p_have - 1;
#ifdef DEBUG
    fprintf(stderr, "link: grabbed empty (%d left)\n", pz->p_have);
#endif
  } 

  /* third choice: allocate if no reserve or used less than pool */
  if((nk == NULL) && ((pz->p_have <= 0) || ((pz->p_pool < 0) || (pz->p_used < pz->p_pool)))){
    nk = malloc(sizeof(struct link_loop));
    if(nk){
      nk->k_type = 0;
      nk->k_data = NULL;
      nk->k_task[LO_INQ] = nk->k_task[LO_OUTQ] = NULL;
      nk->k_back[LO_INQ] = nk->k_back[LO_OUTQ] = NULL;
      nk->k_next[LO_INQ] = nk->k_next[LO_OUTQ] = NULL;
    }
  }

  /* fourth choice: grab stuff from anybody */
  if((nk == NULL) && (pz->p_have > 0)){
    t = 0;
    m = 0;
    for(i = 1; i < ls->l_pool_size; i++){
      pl = &(ls->l_pool[i]);
      if(pl->p_have > m){
        t = i;
        m = pl->p_have;
      }
    }
    if(m > 0){ /* got something */
      pl = &(ls->l_pool[t]);
      nk = pl->p_free;
      pl->p_free = dequeue_loop(pl->p_free, nk, LO_FREEQ);
      pl->p_have = pl->p_have - 1;
      pz->p_have = pz->p_have - 1;
      if(pl->p_empty){
        (*(pl->p_empty))(ls, nk->k_data);
      }
      nk->k_data = NULL;
      nk->k_type = 0;
    }
  }

  pl = NULL;
  if(type){
    pl = &(ls->l_pool[type]);
  }

  if(nk == NULL){
#ifdef DEBUG
    if(pl){
      fprintf(stderr, "link: unable to find anything local %d: used: %d(%d), quota %d(%d)\n", type, pl->p_used, pz->p_used, pl->p_max, pz->p_max);
    } else {
      fprintf(stderr, "link: unable to find anything global: used: %d, quota %d\n", pz->p_used, pz->p_max);
    }
#endif
    return NULL;
  }

  if(pl){
    pl->p_used = pl->p_used + 1;
  }
  pz->p_used = pz->p_used + 1;

#ifdef DEBUG
  if(pl){
    fprintf(stderr, "link: allocated local link[%d]: used: %d(%d), have more %d(%d)\n", type, pl->p_used, pz->p_used, pl->p_have, pz->p_have);
  } else {
    fprintf(stderr, "link: allocated global link: used: %d, have more %d\n", pz->p_used, pz->p_have);
  }
  fprintf(stderr, "link: enqueing: in=%p, out=%p\n", tx, tl);
#endif

  if(tl->t_stop == LO_STOP_NOT){
    tl->t_links[LO_OUTQ] = enqueue_loop(tl->t_links[LO_OUTQ], nk, LO_OUTQ);
    tl->t_have[LO_OUTQ] = tl->t_have[LO_OUTQ] + 1;
    nk->k_task[LO_OUTQ] = tl;
  }

  tx->t_links[LO_INQ] = enqueue_loop(tx->t_links[LO_INQ], nk, LO_INQ);
  tx->t_have[LO_INQ] = tx->t_have[LO_INQ] + 1;
  nk->k_task[LO_INQ] = tx;

#ifdef DEBUG
  dumpqueue_loop(stderr, tl->t_links[LO_OUTQ], LO_OUTQ, "link: out");
  dumpqueue_loop(stderr, tx->t_links[LO_INQ], LO_INQ, "link: in");
#endif

  if(tx->t_new == NULL){
    tx->t_new = nk;
  }

  return nk;
}

/* break the given link between two tasks */

int discard_link_loop(struct list_loop *ls, struct link_loop *nk)
{
  struct task_loop *tl;
  struct link_loop *nx;
  struct pool_loop *pl, *pz;
  int i;

#ifdef DEBUG
  if(nk->k_type >= ls->l_pool_size){
    fprintf(stderr, "break logic error: destroying unknown type\n");
    abort();
  }
#endif

#ifdef DEBUG
  fprintf(stderr, "break: removing %p\n", nk);
#endif

  for(i = LO_INQ; i <= LO_OUTQ; i++){
    if(nk->k_task[i]){
      tl = nk->k_task[i];
#ifdef DEBUG
      dumpqueue_loop(stderr, tl->t_links[i], i, "break: before ");
      if(!(isqueue_loop(tl->t_links[i], nk, i))){
        fprintf(stderr, "logic failure: split link list (%p[%d] not in %p)\n", nk, i, tl->t_links[i]);
        fflush(stderr);
        abort();
      }
#endif

      if(i == LO_INQ){
        if(tl->t_new == nk){ /* new needs to be updated if it is removed */
          nx = nk->k_next[i];
#ifdef DEBUG
          fprintf(stderr, "break: new from %p ", tl->t_new);
#endif
          if((nx == nk) || (tl->t_links[i] == nx)){
            tl->t_new = NULL;
          } else {
            tl->t_new = nx;
          }
#ifdef DEBUG
          fprintf(stderr, "to %p\n", tl->t_new);
#endif
        }
      }
      tl->t_links[i] = dequeue_loop(tl->t_links[i], nk, i);
      tl->t_have[i] = tl->t_have[i] - 1;
      
#ifdef DEBUG
      dumpqueue_loop(stderr, tl->t_links[i], i, "break: after ");
#endif

      nk->k_task[i] = NULL;
    }

  }

  pl = &(ls->l_pool[nk->k_type]);
  pz = &(ls->l_pool[0]);

  pz->p_used--;
  if(nk->k_type > 0){
    pl->p_used--;
  }
  if((nk->k_type > 0) && ((pl->p_pool == 0) || (pl->p_have < pl->p_pool))){
    pz->p_have++;
    pl->p_have++;
    pl->p_free = enqueue_loop(pl->p_free, nk, LO_FREEQ);
  } else {
    if(nk->k_data){
      if(pl->p_empty){
        (*(pl->p_empty))(ls, nk->k_data);
      }
      nk->k_data = NULL;
      nk->k_type = 0;
    }
    if((pz->p_pool == 0) || (pz->p_have < pz->p_pool)){
      pz->p_have++;
      pz->p_free = enqueue_loop(pz->p_free, nk, LO_FREEQ);
    } else {
      free(nk);
    }
  }

  return 0;
}

/* view new incomming links */

struct link_loop *receive_link_loop(struct list_loop *ls, struct task_loop *tl)
{
#ifdef DEBUG
  fprintf(stderr, "receive: task=%p, return %p (head->%p)\n", tl, tl->t_new, tl->t_links[LO_INQ]);
#endif
  return tl->t_new;
}

/* turn new links into ones which won't trigger WAIT */

int ack_link_loop(struct list_loop *ls, struct task_loop *tl, struct link_loop *nk)
{
  struct link_loop *nx;

#ifdef DEBUG
  if(!(isqueue_loop(tl->t_links[LO_INQ], nk, LO_INQ))){
    fprintf(stderr, "logic failure: acking nonmember link\n");
    fflush(stderr);
    abort();
  }
#endif

  nx = nk->k_next[LO_INQ];
  if(nx == tl->t_links[LO_INQ]){
    tl->t_new = NULL;
  } else {
    tl->t_new = nx;
  }

#ifdef DEBUG
  fprintf(stderr, "ack: acked %p, new now %p\n", nk, tl->t_new);
#endif

  return 0;
}

/* return non-new links */

struct link_loop *work_link_loop(struct list_loop *ls, struct task_loop *tl)
{
  if(tl->t_links[LO_INQ] == tl->t_new){
    return NULL;
  }

  return tl->t_links[LO_INQ];
}

/* mark non-new links as new again */

void reset_link_loop(struct list_loop *ls, struct task_loop *tl)
{
#ifdef DEBUG
  fprintf(stderr, "reset: requeuing links (was %p, now %p)\n", tl->t_new, tl->t_links[LO_INQ]);
#endif
  tl->t_new = tl->t_links[LO_INQ];
}

/* disown / forget about outgoing links, used on shutdown */

int detach_link_loop(struct list_loop *ls, struct link_loop *nk)
{
  struct task_loop *tl, *tx;

  tx = nk->k_task[LO_INQ];
  tl = nk->k_task[LO_OUTQ];

#ifdef DEBUG
  if(tx == NULL){
    fprintf(stderr, "logic problem: can't forget, otherwise no reference\n");
    abort();
  }
  if(!(isqueue_loop(tx->t_links[LO_INQ], nk, LO_INQ) && isqueue_loop(tl->t_links[LO_OUTQ], nk, LO_OUTQ))){
    fprintf(stderr, "logic failure: forgetting broken links\n");
    fflush(stderr);
    abort();
  }
#endif

  tl->t_links[LO_OUTQ] = dequeue_loop(tl->t_links[LO_OUTQ], nk, LO_OUTQ);
  tl->t_have[LO_OUTQ] = tl->t_have[LO_OUTQ] - 1;

  nk->k_task[LO_OUTQ] = NULL;

  return 0;
}

int respond_link_loop(struct list_loop *ls, struct link_loop *nk)
{
  if(reply_link_loop(ls, nk)){
    return discard_link_loop(ls, nk);
  }
  return 0;
}

/* flip link around, if other side still there */

int reply_link_loop(struct list_loop *ls, struct link_loop *nk)
{
  struct task_loop *tl, *tx;
  struct link_loop *nx;

  tl = nk->k_task[LO_INQ];
  tx = nk->k_task[LO_OUTQ];
  
  if(tx == NULL){ /* other side has gone away */
#ifdef DEBUG
    fprintf(stderr, "reply: problem - other side has gone away\n");
#endif
    return -1;
  }

  if((tl->t_max[LO_OUTQ] > 0) && (tl->t_have[LO_OUTQ] >= tl->t_max[LO_OUTQ])){
#ifdef DEBUG
    fprintf(stderr, "reply: over output quota %d\n", tl->t_max[LO_OUTQ]);
#endif
    return -1;
  }

  if((tx->t_max[LO_INQ] > 0) && (tx->t_have[LO_INQ] >= tx->t_max[LO_INQ])){
#ifdef DEBUG
    fprintf(stderr, "reply: over input quota %d\n", tx->t_max[LO_INQ]);
#endif
    return -1;
  }

#ifdef DEBUG
  dumpqueue_loop(stderr, tl->t_links[LO_INQ], LO_INQ,   "before flip tl in");
  dumpqueue_loop(stderr, tl->t_links[LO_OUTQ], LO_OUTQ, "before flip tl out");
  dumpqueue_loop(stderr, tx->t_links[LO_INQ], LO_INQ,   "before flip tx in");
  dumpqueue_loop(stderr, tx->t_links[LO_OUTQ], LO_OUTQ, "before flip tx out");
#endif

#ifdef DEBUG
  if(!(isqueue_loop(tl->t_links[LO_INQ], nk, LO_INQ) && isqueue_loop(tx->t_links[LO_OUTQ], nk, LO_OUTQ))){
    fprintf(stderr, "logic failure: flipping broken links\n");
    fflush(stderr);
    abort();
  }
#endif

  /* update new pointers on both sides */
  nx = nk->k_next[LO_INQ];
  if((nx == nk) || (tl->t_links[LO_INQ] == nx)){
    tl->t_new = NULL;
  } else {
    tl->t_new = nx;
  }
#ifdef DEBUG
  fprintf(stderr, "flip: updated new is %p (nx=%p, nk=%p)\n", tl->t_new, nx, nk);
#endif

  if(tx->t_new == NULL){
    tx->t_new = nk;
  }

  tl->t_links[LO_INQ] = dequeue_loop(tl->t_links[LO_INQ], nk, LO_INQ);
  tx->t_links[LO_OUTQ] = dequeue_loop(tx->t_links[LO_OUTQ], nk, LO_OUTQ);

  tl->t_have[LO_INQ] = tl->t_have[LO_INQ] - 1;
  tx->t_have[LO_OUTQ] = tx->t_have[LO_OUTQ] - 1;

  tx->t_links[LO_INQ] = enqueue_loop(tx->t_links[LO_INQ], nk, LO_INQ);
  tl->t_links[LO_OUTQ] = enqueue_loop(tl->t_links[LO_OUTQ], nk, LO_OUTQ);

  tx->t_have[LO_INQ] = tx->t_have[LO_INQ] + 1;
  tl->t_have[LO_OUTQ] = tl->t_have[LO_OUTQ] + 1;

#ifdef DEBUG
  dumpqueue_loop(stderr, tl->t_links[LO_INQ], LO_INQ,   "after flip tl in");
  dumpqueue_loop(stderr, tl->t_links[LO_OUTQ], LO_OUTQ, "after flip tl out");
  dumpqueue_loop(stderr, tx->t_links[LO_INQ], LO_INQ,   "after flip tx in");
  dumpqueue_loop(stderr, tx->t_links[LO_OUTQ], LO_OUTQ, "after flip tx out");
#endif

  nk->k_task[LO_OUTQ] = tl;
  nk->k_task[LO_INQ] = tx;

  return 0;
}

int refer_link_loop(struct list_loop *ls, struct link_loop *nk, struct task_loop *tx)
{
  struct task_loop *tl;
  struct link_loop *nx;

  tl = nk->k_task[LO_INQ];

  if((tx == nk->k_task[LO_OUTQ]) || (tx == NULL) || (tx->t_stop)){
#ifdef DEBUG
    fprintf(stderr, "refer: problem - other side has gone away, or is same as this side\n");
#endif
    return -1;
  }

  if((tx->t_max[LO_INQ] > 0) && (tx->t_have[LO_INQ] >= tx->t_max[LO_INQ])){
#ifdef DEBUG
    fprintf(stderr, "refer: refer target over quota %d\n", tx->t_max[LO_INQ]);
#endif
    return -1;
  }

#ifdef DEBUG
  if(tl == NULL){
    fprintf(stderr, "refer: major logic problem: data sink not there\n");
    abort();
  }
#endif

#ifdef DEBUG
  dumpqueue_loop(stderr, tl->t_links[LO_INQ], LO_INQ, "refer: before tl");
  dumpqueue_loop(stderr, tx->t_links[LO_INQ], LO_INQ, "refer: before tx");
#endif

#ifdef DEBUG
  if(!isqueue_loop(tl->t_links[LO_INQ], nk, LO_INQ)){
    fprintf(stderr, "logic failure: referring broken links\n");
    fflush(stderr);
    abort();
  }
#endif

#ifdef DEBUG /* can be skipped, will just do wasted work */
  if(tx == tl){
    fprintf(stderr, "refer: referring to self, so doing nothing\n");
    return 0;
  }
#endif

  if(tl->t_new == nk){ /* new needs to be updated if it is removed */
    nx = nk->k_next[LO_INQ];
#ifdef DEBUG
    fprintf(stderr, "refer: updating new pointer of old from %p ", tl->t_new);
#endif
    if((nx == nk) || (tl->t_links[LO_INQ] == nx)){
      tl->t_new = NULL;
    } else {
      tl->t_new = nx;
    }
#ifdef DEBUG
    fprintf(stderr, "to %p\n", tl->t_new);
#endif
  }

  tl->t_links[LO_INQ] = dequeue_loop(tl->t_links[LO_INQ], nk, LO_INQ);
  tl->t_have[LO_INQ] = tl->t_have[LO_INQ] - 1;
  nk->k_task[LO_INQ] = NULL;
      
#ifdef DEBUG
  dumpqueue_loop(stderr, tl->t_links[LO_INQ], LO_INQ, "refer: after tl");
#endif

  if(tx->t_new == NULL){
    tx->t_new = nk;
  }

  tx->t_links[LO_INQ] = enqueue_loop(tx->t_links[LO_INQ], nk, LO_INQ);
  tx->t_have[LO_INQ] = tx->t_have[LO_INQ] + 1;
  nk->k_task[LO_INQ] = tx;

#ifdef DEBUG
  dumpqueue_loop(stderr, tx->t_links[LO_INQ], LO_INQ,   "refer: after tx");
#endif

  return 0;
}

