/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

#include <sys/resource.h>
#include <sys/types.h>

int misc_unblock_loop(int fd)
{
  int flags;

  flags = fcntl(fd, F_GETFL, 0);

  if (flags < 0){
    return -1;
  }

  return fcntl(fd, F_SETFL, O_NONBLOCK | flags);
}

int misc_xnprintf_loop(char *buffer, size_t size, char *format, ...)
{
  int result;
  va_list ap;

  if((buffer == NULL) || (size <= 0)){
    return 0;
  }

  va_start(ap, format);

  result = vsnprintf(buffer, size, format, ap);
  if((result < 0) || (result >= size)){
    buffer[size - 1] = '\0';
    result = size - 1;
  }

  va_end(ap);

  return result;
}

void misc_close_loop(int fd, ...)
{
  int match;
  int td;
  va_list ap;

  match = 0;
  va_start(ap, fd);

  td = va_arg(ap, int);
  while(td != (-1)){
    if(td == fd){
      match = 1;
    }
    td = va_arg(ap, int);
  }

  if(match == 0){
    close(fd);
  }

  va_end(ap);
}

int misc_drop_loop(char *id, char *dir, int resource)
{
  struct rlimit r;
  struct passwd *pw;
  int result;
  uid_t uid;
  gid_t gid;

  uid = 0;
  gid = 0;
  result = 0;

  if(id){
    uid = atoi(id);
    if(uid <= 0){
      pw = getpwnam(id);
      if(pw == NULL){
	fprintf(stderr, "drop: unable to look up user %s\n", id);
        result++;
      } else {
	uid = pw->pw_uid;
	gid = pw->pw_gid;
      }
    } else {
      gid = uid;
    }
  }

  if(dir != NULL){	/* do chroot */
    if(chroot(dir)){
      fprintf(stderr, "drop: unable to change root to %s: %s\n", dir, strerror(errno));
      result++;
    }
  }
  chdir("/");

  if(id){			/* now change id */
    if(setgroups(0, NULL)){
      fprintf(stderr, "drop: unable to delete supplementary groups: %s\n", strerror(errno));
      result++;
    }

    if(setgid(gid) || setuid(uid)){
      fprintf(stderr, "drop: unable to change id to %d/%d: %s\n", uid, gid, strerror(errno));
      result++;
    }
  }

  if(resource){
    r.rlim_cur = 0;
    r.rlim_max = 0;

    if(setrlimit(RLIMIT_NPROC, &r)){
      fprintf(stderr, "drop: unable to reduce process limit: %s\n", strerror(errno));
      result++;
    }
  }


  return result;
}
