/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "loop.h"
#include "priv.h"

struct link_loop *enqueue_loop(struct link_loop *head, struct link_loop *add, int index)
{
  struct link_loop *back;
#ifdef DEBUG
  if((index < 0) || (index > 1)){
    fprintf(stderr, "logic failure: busted index\n");
    fflush(stderr);
    abort();
  }
  if(add == NULL){
    fprintf(stderr, "queue: adding nothing\n");
    abort();
  }
#endif
  if(head){
    back = head->k_back[index];

    add->k_next[index] = head;
    add->k_back[index] = back;

    back->k_next[index] = add;
    head->k_back[index] = add;

    return head;
  } else {
    add->k_back[index] = add;
    add->k_next[index] = add;

    return add;
  }
}

struct link_loop *dequeue_loop(struct link_loop *head, struct link_loop *del, int index)
{
  struct link_loop *back, *next;

  if(head == NULL){
    return NULL;
  }

  back = del->k_back[index];
  next = del->k_next[index];

  del->k_back[index] = NULL;
  del->k_next[index] = NULL;

  if(del == next){
    return NULL;
  } 

  next->k_back[index] = back;
  back->k_next[index] = next;
  
  if(del == head){
    return next;
  }

  return head;
}

int isqueue_loop(struct link_loop *head, struct link_loop *test, int index)
{
  struct link_loop *next;

  if(head == NULL){
    return 0;
  }

  next = head;
  do{
    if(next == test){
      return 1;
    }
    next = next->k_next[index];
  } while(next != head);

  return 0;
}

void dumpqueue_loop(FILE *fp, struct link_loop *head, int index, char *prefix)
{
  struct link_loop *next, *last;

  fprintf(fp, "%s[%d]: head->", prefix ? prefix : "queue", index);

  if(head == NULL){
    fprintf(fp, " <empty>\n");
  } else {
    next = head;
    do{
      fprintf(fp, " <%p> ", next); fflush(fp);
      last = next;
      next = next->k_next[index];
      if(next == NULL){
        fprintf(fp, " <ERROR: last: %p<%p>%p>", last->k_back[index], last, last->k_next[index]);
        next = head;
      } else if(next->k_back[index] != last){
        fprintf(fp, " <DAMAGED: back %p>", next->k_back[index]);
      }
    } while(next != head);
    fprintf(fp, "\n");
  }
}

