/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <loop.h>
#include <code-name.h>
#include <word-work.h>

#include "mxk.h"

void mxk_milli_to_tv(struct timeval *tv, unsigned int ms)
{
  tv->tv_sec = ms / 1000;
  tv->tv_usec = (ms % 1000) * 1000;
}

char *mxk_skip_space(char *ptr)
{
  int i;

  for(i = 0; (ptr[i] == ' ' ) || (ptr[i] == '\t'); i++);

  return (ptr + i);
}

#if 0 /* too dangerous to ignore syns */
struct input_event *mxk_get_key(struct list_loop *ls, struct link_loop *nk, int hit)
{
  struct mxk_global *mg;
  struct input_event *vp;

  mg = global_loop(ls);
  if(mg == NULL){
    return NULL;
  }

  if(nk->k_type != mg->g_event){
    return NULL;
  }

  vp = nk->k_data;
  if(vp == NULL){
    return NULL;
  }

  if(vp->type != EV_KEY){
    return NULL;
  }

#if 0 /* resolved that releases move along pipeline, repeats genrally not */
  if(hit && (vp->value != 1)){
    return NULL;
  }
#endif

  return vp;
}
#endif

int mxk_parse_apply(struct list_loop *ls, struct task_loop *tl, char *args, char *name, int empty, int special, int (*fn)(struct list_loop *ls, struct task_loop *tx, int special, char *args))
{
  struct control_loop *cl;
  struct task_loop *tx;
  int cd, result, count, id, idx, type, errors;
  char *value, *end, *next;
  struct mxk_global *mg;

  mg = global_loop(ls);
  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, name);
  if(type <= 0){
    print_control(cl, cd, 1, "%s: type not defined\n", name);
    return -1;
  }

  id = 0;
  next = NULL;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, &end, 10);
      switch(end[0]){
        case '\n' :
        case '\r' :
        case '\0' :
        case ' '  :
          idx = next_word(args, idx);
          if(idx > 0){
            next = args + idx;
          }
          break;
        default :
          print_control(cl, cd, 1, "%s: first parameter has to be an integer\n", name);
          return -1;
      }
    } else {
      if(empty == MXK_MISC_REQUIRED){
        print_control(cl, cd, 1, "%s: require nonempty parameters\n", name);
        return -1;
      }
    }
  } else {
    if(empty == MXK_MISC_REQUIRED){
      print_control(cl, cd, 1, "%s: require parameters\n", name);
      return -1;
    }
  }

#ifdef DEBUG
  fprintf(stderr, "apply: working %s (%d, %d)\n", name, type, id);
#endif

  errors = 0;
  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    result = (*fn)(ls, tx, special, next);
    if(result < 0){
      errors++;
    }
    count++;
  }

  switch(count){
    case 0 :
      print_control(cl, cd, 1, "%s: no match for (%d,%d)\n", name, type, id);
      return 1;
    case 1:
      if((mg->g_verbose > 1) || errors){
        print_control(cl, cd, 1, "%s: request %s for (%d,%d)\n", name, errors ? "failed" : "completed", type, id);
      }
      return errors ? (-1) : 0;
    default :
      if((mg->g_verbose > 1) || errors){
        if(count <= errors){
          print_control(cl, cd, 1, "%s: request failed in %d instances for (%d,%d)\n", name, count, type, id);
        } else {
        print_control(cl, cd, 1, "%s: %d of %d requests completed for (%d,%d)\n", name, count - errors, count, type, id);
        }
      }
      return errors ? 1 : 0;
  }
}

int mxk_parse_evt(char *args, struct input_event *evt)
{
  int idx, type, code, value;
  char buffer[MXK_NAME_SIZE];
  char *vptr, *cptr;
  int result;

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    return -1;
  }

  cptr = strchr(buffer, ':');
  if(cptr){
    *cptr = '\0';
    cptr++;
    type = name_to_code(mxk_event_types, buffer);
    if(type < 0){
      type = atoi(buffer);
    }
    if((type < 0) || (type >= EV_MAX)){
#ifdef DEBUG
      fprintf(stderr, "parse: unknown event type <%s>\n", buffer);
#endif
      return -1;
    }
  } else {
    type = EV_KEY;
    cptr = buffer;
  }

  vptr = strchr(cptr, '/');
  if(vptr){
    *vptr = '\0';
    vptr++;
    result = 0;
  } else {
    /* WARNING: currently unused return code > 0 */
    result = 1;
  }

  code = 0;
  value = 0;
  switch(type){
    case EV_SYN :
#ifdef DEBUG
      fprintf(stderr, "parse: assuming sync\n");
#endif
    break;
    case  EV_KEY :
      value = 1;
      if(vptr){
        value = name_to_code(mxk_key_values, vptr);
        if(value < 0){
#ifdef DEBUG
          fprintf(stderr, "parse: unknown key value\n");
#endif
          return -1;
        }
      }
      code = name_to_code(mxk_key_codes, cptr);
      if(code < 0){
        code = atoi(buffer);
      }
      if(code < 0){
#ifdef DEBUG
        fprintf(stderr, "parse: unknown key code\n");
#endif
        return -1;
      }
    break;
    case EV_ABS :
    case EV_REL :
      value = 1;
      if(vptr){
        value = atoi(vptr);
        if((type == EV_REL) && (value == 0)){
#ifdef DEBUG
          fprintf(stderr, "parse: zero motion considered an error\n");
#endif
          return -1;
        }
      }
      switch(type){
        case EV_REL :
          code = name_to_code(mxk_rel_codes, cptr);
          break;
        case EV_ABS :
          code = name_to_code(mxk_abs_codes, cptr);
          break;
      }
      if(code < 0){
#ifdef DEBUG
        fprintf(stderr, "parse: unknown/unsupported move\n");
#endif
        return -1;
      }
    break;
    default :
      value = 0;
      if(vptr){
        value = atoi(vptr);
      }
      code = atoi(cptr);
      return -1;
  }

  evt->type = type;
  evt->code = code;
  evt->value = value;

  return result;
}

int mxk_string_evt(char *buffer, int len, int type, int code, int value)
{
  char *stype, *scode, *svalue;
  int result;

  stype = code_to_name(mxk_event_types, type);
  scode = NULL;
  svalue = NULL;

  if(stype){
    switch(type){
      case EV_SYN :
        if(code == SYN_REPORT){
          result = snprintf(buffer, len, "%s:", stype);
          return ((result < 0) || (result >= len)) ? (-1) : result;
        }
        break;
      case EV_KEY :
        scode = code_to_name(mxk_key_codes, code);
        svalue = code_to_name(mxk_key_values, value);
        if(scode && svalue){
          result = snprintf(buffer, len, "%s/%s", scode, svalue);
          return ((result < 0) || (result >= len)) ? (-1) : result;
        }
        break;
      case EV_REL :
        scode = code_to_name(mxk_rel_codes, code);
        break;
      case EV_ABS :
        scode = code_to_name(mxk_abs_codes, code);
        break;
    }

    if(scode){
      if(svalue){
        result = snprintf(buffer, len, "%s:%s/%s", stype, scode, svalue);
      } else {
        result = snprintf(buffer, len, "%s:%s/%d", stype, scode, value);
      }
    } else {
      result = snprintf(buffer, len, "%s:%d/%d", stype, code, value);
    }
  } else {
    result = snprintf(buffer, len, "%d:%d/%d", type, code, value);
  }

  return ((result < 0) || (result >= len)) ? (-1) : result;
}

int mxk_print_evt(struct control_loop *cl, int cd, int type, int code, int value)
{
  char buffer[MXK_NAME_SIZE];
  int result;

  result = mxk_string_evt(buffer, MXK_NAME_SIZE, type, code, value);
  if(result < 0){
    return -1;
  }

  return print_control(cl, cd, 0, "%s", buffer);
}

int mxk_has_bit(unsigned long *vector, unsigned int max, unsigned int test)
{
  unsigned int d, m;

  if(max <= test){
    return 0;
  }

  m = test % (8 * sizeof(unsigned long));
  d = test / (8 * sizeof(unsigned long));

  return (vector[d] & (1UL << m)) ? 1 : 0;
}

int mxk_copying_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int cd;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  print_control(cl, cd, 0, "(c) 1999,2004-2009 Marc Welz\n");
  print_control(cl, cd, 1, "Distribution under the terms of the GNU GPL v3 or newer\n");

  return 0;
}

int mxk_version_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int cd;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  print_control(cl, cd, 1, "mxk-%s\n", VERSION);

  return 0;
}

int mxk_echo_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int cd;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  if(args != NULL){
    print_control(cl, cd, 1, "%s\n", args);
  } else {
    print_control(cl, cd, 1, "\n");
  }

  return 0;
}
