/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>

#include <loop.h>
#include <word-work.h>

#include "mxk.h"

struct mxk_global *start_state(int verbose)
{
  struct mxk_global *mg;

  mg = malloc(sizeof(struct mxk_global));
  if(mg == NULL){
    return NULL;
  }

  mg->g_length = 0;
  mg->g_nonce = 0;

  mg->g_event = (-1);

  mg->g_patterns = NULL;
  mg->g_psize = 0;

  mg->g_table = NULL;
  mg->g_tsize = 0;

  mg->g_verbose = verbose;

  return mg;
}

void stop_state(struct mxk_global *mg)
{
  if(mg == NULL){
    return;
  }

  mxk_shutdown_watch(mg);
  mxk_shutdown_source(mg);
  mxk_shutdown_light(mg);
  mxk_end_find(mg);

  free(mg);
}

void *mxk_setup_control(struct list_loop *ls, void *data, struct sockaddr *addr, int fd)
{
  struct control_loop *cl;
  int result = 0;

  cl = default_setup_control(ls, data, addr, fd);
  if(cl == NULL){
    return NULL;
  }

  result += add_command_help_control(cl, "echo",             &mxk_echo_command, "echo parameters");
  result += add_command_help_control(cl, "copying",          &mxk_copying_command, "display copyright notice");
  result += add_command_help_control(cl, "version",          &mxk_version_command, "display version");

  result += add_command_help_control(cl, "name-source",      &mxk_name_source_command, "generate a source which matches the supplied substring");
  result += add_command_help_control(cl, "file-source",      &mxk_file_source_command, "generate a source from the supplied device file");
  result += add_command_help_control(cl, "caps-source",      &mxk_caps_source_command, "generate a source from a device with the supplied capabilities");
  result += add_command_help_control(cl, "list-sources",     &mxk_list_source_command, "list available sources");
  result += add_command_help_control(cl, "used-sources",     &mxk_used_source_command, "show used sources");
  result += add_command_help_control(cl, "bypass-source",    &mxk_bypass_source_command, "discard events from a source");
  result += add_command_help_control(cl, "activate-source",  &mxk_activate_source_command, "use a source nonexclusively");
  result += add_command_help_control(cl, "lock-source",      &mxk_lock_source_command, "use a source exclusively (grab)");
  result += add_command_help_control(cl, "repeat-source",    &mxk_forward_repeat_source_command, "relay repeat events generated by a source");
  result += add_command_help_control(cl, "norepeat-source",  &mxk_discard_repeat_source_command, "suppress repeat events generated by a source");
  /* should also have forward-repeat and discard-repeat for symetry */
  result += add_command_help_control(cl, "fixup-source",     &mxk_replace_repeat_source_command, "rewrite repeat events generated by a source");

  result += add_command_help_control(cl, "check-source",     &mxk_check_flaky_source_command, "check for source errors");
  result += add_command_help_control(cl, "believe-source",   &mxk_believe_flaky_source_command, "disable error checks");

  result += add_command_help_control(cl, "start-target",     &mxk_target_command, "create a virtual keyboard as target");

  result += add_command_help_control(cl, "list-patterns",    &mxk_list_pattern_command, "list led patterns");
  result += add_command_help_control(cl, "load-pattern",     &mxk_load_pattern_command, "create a led pattern");
  result += add_command_help_control(cl, "start-light",      &mxk_start_light_command, "run a named led pattern");
  result += add_command_help_control(cl, "stop-light",       &mxk_stop_light_command, "terminate a running led pattern");

  result += add_command_help_control(cl, "echo-watch",       &mxk_echo_watch_command, "generate a message to be received by watches");
  result += add_command_help_control(cl, "start-watch",      &mxk_start_watch_command, "change current connection into a watch");

  result += add_command_help_control(cl, "connect-node",     &mxk_linkup_find_command, "relay events from one node to another");
  result += add_command_help_control(cl, "list-nodes",       &mxk_show_find_command, "list available node types");

  result += add_command_help_control(cl, "start-chord",      &mxk_start_chord_command, "create a node matching multiple events");
  result += add_command_help_control(cl, "list-chord",       &mxk_list_chord_command, "create a node matching multiple events");
  result += add_command_help_control(cl, "old-order-chord",  &mxk_old_order_chord_command, "set pattern matched in order and includes already released keys");
  result += add_command_help_control(cl, "new-order-chord",  &mxk_new_order_chord_command, "set pattern matched in order and limited to pressed keys");
  result += add_command_help_control(cl, "gap-order-chord",  &mxk_gap_order_chord_command, "set pattern matched in order limited to unmatched pressed keys");
  result += add_command_help_control(cl, "old-set-chord",    &mxk_old_set_chord_command, "set pattern matched in any order and includes already released keys");
  result += add_command_help_control(cl, "new-set-chord",    &mxk_new_set_chord_command, "set pattern matched in any order but limited to pressed keys");
  result += add_command_help_control(cl, "gap-set-chord",    &mxk_gap_set_chord_command, "set pattern matched in any order limited to unmatched pressed keys");
  result += add_command_help_control(cl, "bypass-chord",     &mxk_bypass_chord_command, "nonmatching key sequences are forwarded");
  result += add_command_help_control(cl, "ignore-chord",     &mxk_ignore_chord_command, "nonmatching key sequences are discarded");

  result += add_command_help_control(cl, "list-events",      &mxk_list_events_command, "list known input events");
  result += add_command_help_control(cl, "list-codes",       &mxk_list_codes_command, "list available keycodes");
  result += add_command_help_control(cl, "list-relative",    &mxk_list_relative_command, "list relative movement axes");
  result += add_command_help_control(cl, "list-absolute",    &mxk_list_absolute_command, "list absolute movement axes");

#if 0
  result += add_command_help_control(cl, "up-bit",     &mxk_start_up_bit_command, "create a bit which emits an event on up edge");
  result += add_command_help_control(cl, "down-bit",   &mxk_start_down_bit_command, "create a bit which emits an event on down edge");
  result += add_command_help_control(cl, "sticky-bit", &mxk_start_sticky_bit_command, "create a bit which emits an event on up level");
  result += add_command_help_control(cl, "pair-bit",   &mxk_start_pair_bit_command, "create a bit which emits an event on either edge");

  result += add_command_control(cl, "light-bit",   &mxk_light_bit_command);
  result += add_command_control(cl, "default-bit", &mxk_default_bit_command);
  result += add_command_control(cl, "key-bit",     &mxk_key_bit_command);
#endif

  result += add_command_help_control(cl, "start-array",      &mxk_start_array_command, "create a node which maps a single event to multiple ones");
  result += add_command_help_control(cl, "list-arrays",      &mxk_list_array_command, "list content of array nodes");
  result += add_command_help_control(cl, "light-array",      &mxk_light_array_command, "set light pattern when array enters record mode");
  result += add_command_help_control(cl, "press-array",      &mxk_press_array_command, "register a keypress and associate it with an array of events");
  result += add_command_help_control(cl, "release-array",    &mxk_release_array_command, "register a keyrelease and associate it with an array of events");
  result += add_command_help_control(cl, "quick-array",      &mxk_quick_array_command, "compact form of press-array with expansions and tests");
  result += add_command_help_control(cl, "any-array",        &mxk_any_array_command, "allow any key to be mapped in array");
  result += add_command_help_control(cl, "preset-array",     &mxk_preset_array_command, "allow only existing mappings to be redefined in array");
  result += add_command_help_control(cl, "bypass-array",     &mxk_bypass_array_command, "relay unhandled events in array");
  result += add_command_help_control(cl, "ignore-array",     &mxk_ignore_array_command, "ignore unhandled events in array");

  result += add_command_help_control(cl, "start-branch",     &mxk_start_branch_command, "create a node which functions as conditional");
  result += add_command_help_control(cl, "list-branches",    &mxk_list_branch_command, "display branch node details");
  result += add_command_help_control(cl, "light-branch",     &mxk_light_branch_command, "run named pattern when branch is in up/true state");
  result += add_command_help_control(cl, "dark-branch",      &mxk_dark_branch_command, "run named pattern when branch is in down/false state");
  result += add_command_help_control(cl, "up-branch",        &mxk_up_branch_command, "set branch state to up/true state");
  result += add_command_help_control(cl, "down-branch",      &mxk_down_branch_command, "set branch state to down/false state");
  result += add_command_help_control(cl, "key-branch",       &mxk_key_branch_command, "set effect of event on branch");
  result += add_command_help_control(cl, "default-branch",   &mxk_default_branch_command, "set effect of default event on branch");

  result += add_command_help_control(cl, "start-stat",       &mxk_start_stat_command, "create a summary data collection module");
  result += add_command_help_control(cl, "list-stats",       &mxk_list_stat_command, "display accumulated statistics");

  result += add_command_help_control(cl, "start-mode",       &mxk_start_mode_command, "create a mode switch");
  result += add_command_help_control(cl, "list-modes",       &mxk_list_mode_command, "display mode switch information");
  result += add_command_help_control(cl, "setup-mode",       &mxk_setup_mode_command, "configure a mode state");
  result += add_command_help_control(cl, "enter-mode",       &mxk_state_mode_command, "set the current mode state");
  result += add_command_help_control(cl, "key-mode",         &mxk_key_mode_command, "associate a key with a mode transition");

  result += add_command_help_control(cl, "start-repeater",   &mxk_start_repeat_command, "start a repeater");
  result += add_command_help_control(cl, "list-repeaters",   &mxk_list_repeat_command, "display repeaters");
  result += add_command_help_control(cl, "adjust-repeater",  &mxk_adjust_repeat_command, "set keys to change repeater dynamically");

  result += add_command_help_control(cl, "start-single",     &mxk_start_single_command, "start a single axis to single key mapper");
  result += add_command_help_control(cl, "list-singles",     &mxk_list_single_command, "list single axis to single key mappers");
  result += add_command_help_control(cl, "sustain-single",   &mxk_sustained_single_command, "keep output key pressed if no return below threshold");
  result += add_command_help_control(cl, "quick-single",     &mxk_quick_single_command, "do not keep output key pressed");
  result += add_command_help_control(cl, "time-single",      &mxk_timed_single_command, "do not keep output key pressed");
  result += add_command_help_control(cl, "limit-single",     &mxk_threshold_single_command, "set the movement threshold at which keys are generated");

  result += add_command_help_control(cl, "start-vector",     &mxk_start_ax_command, "start many absolute axis to key mapper");
  result += add_command_help_control(cl, "list-vectors",     &mxk_list_ax_command, "list absolute axis to key mappers");
  result += add_command_help_control(cl, "point-vector",     &mxk_point_ax_command, "map a point in absolute axis space to a key");
  result += add_command_help_control(cl, "rezero-vector",    &mxk_return_zero_ax_command, "require return to null axis position before generating next key");
  result += add_command_help_control(cl, "drift-vector",     &mxk_non_zero_ax_command, "generate next key without requiring a return to null axis position");

  result += add_command_help_control(cl, "start-absrel",     &mxk_start_ar_command, "start absolute to relative mapper");
  result += add_command_help_control(cl, "list-absrels",     &mxk_list_ar_command, "list absolute to relative motion mappers");
  result += add_command_help_control(cl, "brake-absrel",     &mxk_brake_ar_command, "set brake key for absolute to relative mapper");
  result += add_command_help_control(cl, "split-absrel",     &mxk_split_ar_command, "set bottom and top fraction for scaling factors in absolute to relative mapper");
  result += add_command_help_control(cl, "pass-absrel",      &mxk_pass_ar_command, "forward unrelated events for absolute to relative mapper");
  result += add_command_help_control(cl, "ignore-absrel",    &mxk_ignore_ar_command, "discard unrelated events in absolute to relative mapper");

  result += add_command_help_control(cl, "start-external",   &mxk_start_external_command, "starts external interface");
  result += add_command_help_control(cl, "list-externals",   &mxk_list_external_command, "list external interfaces");

  result += add_command_help_control(cl, "start-sorter",     &mxk_start_sorter_command, "starts a sorter instance");
  result += add_command_help_control(cl, "list-sorters",     &mxk_list_sorter_command, "list sorters");
  result += add_command_help_control(cl, "match-sorter",     &mxk_match_sorter_command, "enter an event into sorter");
  result += add_command_help_control(cl, "bin-sorter",       &mxk_bin_sorter_command, "change current bin of sorter");
  result += add_command_help_control(cl, "reserve-sorter",   &mxk_reserve_sorter_command, "initialise event type with default bin");

  result += add_command_help_control(cl, "start-smit",       &mxk_start_smit_command, "initialise a save/emit instance");
  result += add_command_help_control(cl, "list-smits",       &mxk_list_smit_command, "list save/emit instances");
  result += add_command_help_control(cl, "key-smit",         &mxk_key_smit_command, "add a key to save/emit instance");
  result += add_command_help_control(cl, "default-smit",     &mxk_default_smit_command, "set the default for a save/emit instance");

  result += add_command_help_control(cl, "start-brasc",      &mxk_start_brasc_command, "create a braille ascii instance");
  result += add_command_help_control(cl, "list-brascs",      &mxk_list_brasc_command, "list braille ascii instances");

  result += add_command_help_control(cl, "start-timer",      &mxk_start_timster_command, "create a timer instance");
  result += add_command_help_control(cl, "list-timers",      &mxk_list_timster_command, "list timer instances");
  result += add_command_help_control(cl, "trigger-timer",    &mxk_trigger_timster_command, "set key starting timer");
  result += add_command_help_control(cl, "reset-timer",      &mxk_reset_timster_command, "set key resetting timer");
  result += add_command_help_control(cl, "clear-timer",      &mxk_clear_timster_command, "set key stoppping timer");
  result += add_command_help_control(cl, "arm-timer",        &mxk_arm_timster_command, "activate timer");
  result += add_command_help_control(cl, "off-timer",        &mxk_off_timster_command, "disable timer");
  result += add_command_help_control(cl, "auto-timer",       &mxk_auto_timster_command, "disable timer");
  result += add_command_help_control(cl, "single-timer",     &mxk_single_timster_command, "disable timer");

  result += add_command_help_control(cl, "start-guess",      &mxk_start_guess_command, "create a simple guessing instance");
  result += add_command_help_control(cl, "list-guesses",     &mxk_list_guess_command, "list simple guessing instances");

  result += add_command_help_control(cl, "start-genb",      &mxk_start_genb_command, "create a generic brailler");
  result += add_command_help_control(cl, "list-genbs",      &mxk_list_genb_command, "list generic brailler instances");

#if 0
  result += add_command_help_control(cl, "start-farm",      &mxk_start_farm_command, "create a simple prediction farmer");
  result += add_command_help_control(cl, "list-farms",      &mxk_list_farm_command, "list simple prediction farmer");
  result += add_command_help_control(cl, "propose-farm",    &mxk_propose_farm_command, "generate an entry (testing)");
#endif

  result += add_command_help_control(cl, "start-reduce",    &mxk_start_reduce_command, "create a key event filtering node");
  result += add_command_help_control(cl, "list-reducers",   &mxk_list_reduce_command, "list reducing nodes");
  result += add_command_help_control(cl, "key-reduce",      &mxk_key_reduce_command, "add key events to reducing filter");
  result += add_command_help_control(cl, "discard-reduce",  &mxk_discard_reduce_command, "do not forward enumerated events");
  result += add_command_help_control(cl, "relay-reduce",    &mxk_relay_reduce_command, "only forward enumerated events");
  result += add_command_help_control(cl, "self-reduce",     &mxk_self_reduce_command, "forward the event itself");
  result += add_command_help_control(cl, "replace-reduce",  &mxk_replace_reduce_command, "replace event by the specified key when forwarding");
  result += add_command_help_control(cl, "merge-reduce",    &mxk_merge_reduce_command, "merge concurrent events when replacing");
  result += add_command_help_control(cl, "distinct-reduce", &mxk_distinct_reduce_command, "forward concurrent events separately when replacing");

  result += add_command_help_control(cl, "start-bug",       &mxk_start_bug_command, "create an interval sensitive keys");
  result += add_command_help_control(cl, "list-bugs",       &mxk_list_bug_command, "list interval sensitive keys");
  result += add_command_help_control(cl, "bypass-bug",      &mxk_bypass_bug_command, "relay nonmatching key events");
  result += add_command_help_control(cl, "ignore-bug",      &mxk_ignore_bug_command, "discard nomatching key events");

  if(result){
    shutdown_control(cl);
    return NULL;
  }

  return cl;
}

void usage(char *app)
{
  printf("usage: %s [-a] [-d] [-c] [-h] [-q] [-v] [-z] [-u [unix-socket]] [-i user-name] [-p script-file] [-r directory] [-C [log-file]] [-D [log-file]] [-F]\n", app);
  printf("-a                 attach to already running instance\n");
  printf("-c                 copyright information\n");
  printf("-d                 run as daemon (after initial console exits)\n");
  printf("-h                 this help\n");
  printf("-q                 quiet\n");
  printf("-v                 increase verbosity (can be given multiple times)\n");
  printf("-z                 disable initial console\n");
  printf("-u unix-socket     provide control interface on socket\n");
  printf("-p script-file     run script at startup\n");
  printf("-C log-file        write log file at crash\n");
  printf("-D log-file        write debug log file\n");
  printf("-F                 make log file a fifo\n");
}

void copyright(char *app)
{
  printf("mxk (Marc's eXtended Keyboard) v%s: (c) 1999,2004-2009 Marc Welz\n", VERSION);
  printf("Propagation Conditions: FSF GNU GPL v3 or newer\n");
}

/****************************************************************/

int main(int argc, char **argv)
{
  struct list_loop *ls;
  struct task_loop *task;
  struct control_loop *cl;
  struct mxk_global *mg;

  int fd;
  char *tty, *app, *data;
  int i, j;
  char c;
  int mon_flags;

  char *log_file = NULL;
  char *init_file = NULL;

  char *user_loop = NULL;
  char *dir_root = NULL;
  int enable_fifo = 0;
  int console_flags = LO_RUN_CONSOLE;
  int enable_crash = 0;
  int enable_debug = 0;
  int attach = 0;

  char *control_path = MXK_CONTROL_SOCKET;

  int task_limit = 32;
  int link_limit = 64;
  int verbose = 1;

  /* arg parse */

  app = argv[0];
  i = j = 1;

  while (i < argc) {
    if (argv[i][0] == '-') {
      c = argv[i][j];
      switch (c) {
        case 'a' :
          attach = 1;
          j++;
          break;
        case 'c' :
          copyright(app);
          error_exit(EX_OK, NULL);
          break;
        case 'd' :
          console_flags |= LO_DETACH_CONSOLE;
          j++;
          break;
        case 'z' :
          console_flags &= (~LO_RUN_CONSOLE);
          j++;
          break;
        case 'F' :
          enable_fifo = 1;
          j++;
          break;
        case 'h' :
          usage(app);
          error_exit(EX_OK, NULL);
          break;
        case 'v' :
          verbose++;
          j++;
          break;
        case 'q' :
          verbose = 0;
          j++;
          break;
        case 'i' :
        case 'p' :
        case 'r' :
        case 'u' :
        case 'D' :
        case 'C' :
          j++;
          data = NULL;
          if (argv[i][j] == '\0') {
            if(((i + 1) < argc) && (argv[i + 1][0] != '-')){
              j = 0;
              i++;
              data = argv[i] + j;
            } else {
              data = NULL;
            }
          } else {
            data = argv[i] + j;
          }
          switch(c){
            case 'i' :
              user_loop = data;
              break;
            case 'p' :
              init_file = data;
              console_flags |= LO_FILE_CONSOLE;
              break;
            case 'r' :
              dir_root = data;
              break;
            case 'u' :
              control_path = data;
              break;
            case 'D' :
              enable_debug = 1;
              log_file = data;
              /* fall ? */
            case 'C' :
              enable_crash = 1;
              log_file = data;
              break;
          }
          i++;
          j = 1;
          break;
        case 't' :
        case 'l' :
          j++;
          if (argv[i][j] == '\0') {
            j = 0;
            i++;
          }
          if (i >= argc) {
            error_exit(EX_USAGE, "%s: argument needs a parameter\n", argv[0]);
          }
          switch(c){
            case 't' :
              task_limit = atoi(argv[i] + j);
              break;
            case 'l' :
              link_limit = atoi(argv[i] + j);
              break;
          }
          i++;
          j = 1;
          break;
        case '-' :
          j++;
          break;
        case '\0':
          j = 1;
          i++;
          break;
        default:
          error_exit(EX_USAGE, "%s: unknown option -%c\n", app, argv[i][j]);
          break;
      }
    } else {
      error_exit(EX_USAGE, "%s: unknown parameter %s\n", app, argv[i]);
      i++;
    }
  }

  if(attach){
    commit_attach_console(control_path, LO_RUN_CONSOLE);
    error_exit(EX_OSERR, "%s: unable to attach to server on %s: %s\n", app, control_path, errno ? strerror(errno) : "Lost error code");
  }

  if(enable_crash || enable_debug){
    if(log_file == NULL){
      /* no detach if output requested to stderr */
      console_flags |= (~LO_DETACH_CONSOLE);
    }
  }

  mon_flags = LO_MON_START | LO_MON_DEFER; /* start output, but write to stderr until detach or absolve */
  if(enable_crash){
    mon_flags |= LO_MON_ABORT;
    if(!enable_debug){
      mon_flags |= LO_MON_TRUNCATE;
    }
  }
  if(enable_debug){
    mon_flags |= LO_MON_EXIT;
  }
  if(enable_fifo){
    mon_flags |= LO_MON_FIFO;
  }

  if(error_monitor(log_file, mon_flags)){
    error_exit(EX_OSERR, "%s: unable to start error monitor\n", app);
  }

  if(verbose){
    fprintf(stderr, "%s: starting\n", app);
  }

  tty = ttyname(STDIN_FILENO);

  /* aim: keep (possible) parent from exiting if logging to */
  fd = start_profile_console(init_file, console_flags);
  if(fd < 0){
    error_exit(EX_OSERR, "%s: unable to perform setup\n", app);
  }

  mg = start_state(verbose);
  if(mg == NULL){
    error_exit(EX_OSERR, "%s: unable to allocate global state\n", app);
  }

  ls = startup_loop(mg, task_limit, link_limit);
  if(ls == NULL){
    error_exit(EX_SOFTWARE, "%s: unable to setup loop\n", app);
  }

  if(mg->g_verbose){
    fprintf(stderr, "%s: initialising node logic\n", app);
  }

  if(mxk_startup_watch(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise watch logic\n", app);
  }

  if(mxk_startup_light(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise light logic\n", app);
  }

  if(mxk_startup_chord(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise chord logic\n", app);
  }

  if(mxk_startup_source(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise source logic\n", app);
  }

  if(mxk_startup_target(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise target logic\n", app);
  }

#if 0
  if(mxk_startup_bit(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise bit state logic\n", app);
  }
#endif

  if(mxk_startup_branch(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise bit state logic\n", app);
  }

  if(mxk_startup_array(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise array logic\n", app);
  }

  if(mxk_startup_stat(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise statistics collection logic\n", app);
  }

  if(mxk_startup_mode(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise mode switch logic\n", app);
  }

  if(mxk_startup_repeat(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise repeater logic\n", app);
  }

  if(mxk_startup_single(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise movement to single logic\n", app);
  }

  if(mxk_startup_ax(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise multi axis absolute movement logic\n", app);
  }

  if(mxk_startup_ar(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise multi axis absolute to relative logic\n", app);
  }

  if(mxk_startup_external(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise logic to external interfaces\n", app);
  }

  if(mxk_startup_sorter(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise filter logic\n", app);
  }

  if(mxk_startup_smit(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise save/emit logic\n", app);
  }

  if(mxk_startup_brasc(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise braille/ascii logic\n", app);
  }

  if(mxk_startup_timster(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise timer logic\n", app);
  }

  if(mxk_startup_guess(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise simple guessing logic\n", app);
  }

  if(mxk_startup_genb(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise generic brailler logic\n", app);
  }

#if 0
  if(mxk_startup_farm(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise simple genetic farm logic\n", app);
  }
#endif

  if(mxk_startup_reduce(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise reduction logic\n", app);
  }

  if(mxk_startup_bug(ls)){
    error_exit(EX_SOFTWARE, "%s: unable to initialise delay sensitive key logic\n", app);
  }

  mg->g_event = claim_link_loop(ls, 64, 4, &default_empty_loop);
  if(mg->g_event <= 0){
    error_exit(EX_SOFTWARE, "%s: unable to allocate an event link type\n", app);
  }

  if(console_flags){
    if(mg->g_verbose){
      fprintf(stderr, "%s: starting console\n", app);
    }

    cl = mxk_setup_control(ls, NULL, NULL, fd);
    if(cl == NULL){
      error_exit(EX_SOFTWARE, "%s: unable to add console task\n", app);
    }

    task = add_console_loop(ls, fd, cl, LO_READ_MASK | LO_STOP_MASK, run_control, tty);
    if(task == NULL){
      shutdown_control(cl);
      close(fd);
      error_exit(EX_SOFTWARE, "%s: unable to add console task\n", app);
    }
  } else {
    close(fd);
  }

  if(control_path){
    if(mg->g_verbose){
      fprintf(stderr, "%s: starting unix monitor on %s\n", app, control_path);
    }

    task = add_unix_loop(ls, control_path, NULL, mxk_setup_control, LO_READ_MASK | LO_STOP_MASK, run_control, 0);
    if(task == NULL){
      fprintf(stderr, "%s: unable to add unix control task for %s\n", app, control_path);
      return EX_SOFTWARE;
    }
  }

#if 0
  task = add_unix_loop(ls, gray_path, NULL, start_client, LO_READ_MASK | LO_STOP_MASK, gray_run, 0);
  if(task == NULL){
    fprintf(stderr, "%s: unable to add policy server for %s\n", app, gray_path);
    return EX_SOFTWARE;
  }
#endif

#ifndef DEBUG
  if(misc_drop_loop(user_loop, dir_root, 1)){
    fprintf(stderr, "%s: unable to reduce privileges\n", app);
    return EX_SOFTWARE;
  }
#endif

  if(mg->g_verbose){
    fprintf(stderr, "%s: ready to serve requests\n", app);
  }

  if(enable_crash || enable_debug){
    if(!enable_debug){ /* for crash only emit things when they go wrong */
      error_disable();
    }
    if(log_file){ /* if there is a file we can detach */
      error_detach();
    }
  } else {
    error_absolve(); /* no logging requested, so stop child and thus detach */
  }

  /* WARNING: after this point only DEBUG to stderr */

  main_loop_mode(ls, LO_FAVOUR_CPU);

  shutdown_loop(ls);

  stop_state(mg);
  error_absolve();

  return EX_OK;
}
