#ifndef MXK_H_
#define MXK_H_

/* client */

struct mxk_client{
  int c_type;
  int c_id;
};

#include <sys/time.h>

struct mxk_key
{
  union mxk_fork{
    struct mxk_key *k_next;
    int k_value;
  } u;
  unsigned short k_size;
  unsigned char k_priority;
  unsigned char k_depth;
};

/* degap */

struct mxk_degap
{
  int *d_vector;
  int d_size;
  int d_max;
};

/* queue */

struct mxk_qitem{
  int i_code;
  int i_id;
};

#define MXK_QUEUE_MAP   4

#define MXK_QUEUE_OLD_PRESS   0
#define MXK_QUEUE_NEW_RELEASE 1
#define MXK_QUEUE_OLD_RELEASE 2
#define MXK_QUEUE_TOTAL_SIZE  3

struct mxk_queue{
  struct mxk_qitem *q_queue;

  int q_index[MXK_QUEUE_MAP]; /* items in queue (looks like pointer to 4 types) */

  int q_fancy; /* do expensive checks */
  int q_repeat;
};

#define MXK_EVT_NULL    0x0
#define MXK_EVT_TARGET  0x1
#define MXK_EVT_SOURCE  0x2

struct mxk_watch{
  int w_sofar;
  int w_nonce;
};

#define MXK_FLASH_ON     1
#define MXK_FLASH_OFF    2
#define MXK_FLASH_WAIT   3
#define MXK_FLASH_REPEAT 4

struct mxk_flash{
  int f_id;
  int f_code;
  int f_value;
};

struct mxk_pattern{
  char *p_name;
  int p_size;
  struct mxk_flash *p_array;
  int p_used;
  int p_id;
};

struct mxk_light{
  struct mxk_pattern *l_current;
  int l_index;
  int l_count;
  int l_source_type;
};

struct mxk_find{
  int f_type; 
  char *f_name;
  int f_flags;
  int (*f_insert)(void *server, int type, int id);
};

#define MXK_WATCH_SIZE  256

#define MXK_NAME_SIZE  128
#define MXK_SOURCE_DIR  "/dev/input"
#ifndef VERSION
#define VERSION "unknown"
#endif

#ifdef DEBUG
#define MXK_CONTROL_SOCKET "socket"
#else
#define MXK_CONTROL_SOCKET "/var/run/mxk"
#endif

#include <loop.h>

struct mxk_global{
  char g_message[MXK_WATCH_SIZE]; /* message buffer written to all monitors */
  int g_length; /* size of message */
  int g_nonce; /* used to discover if message buffer new */

  int g_event; /* event link type */

  struct mxk_pattern **g_patterns; /* pattern table for light flashes */
  int g_psize;
  int g_pid; /* pattern id, increases sequentially */

  struct mxk_find *g_table; /* type table for source and target node tasks */
  int g_tsize;

  int g_verbose;
};

#include <sys/stat.h> 
#include <sys/types.h> 

#include <linux/input.h>
#include <linux/uinput.h>

#ifndef NBITS
#define NBITS(x) ((((x)-1)/(sizeof(unsigned long) * 8))+1)
#endif

/* source */
int mxk_used_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_name_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_file_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_caps_source_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_bypass_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_activate_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_lock_source_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_startup_source(struct list_loop *ls);
void mxk_shutdown_source(struct mxk_global *mg);

int mxk_forward_repeat_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_discard_repeat_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_replace_repeat_source_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_check_flaky_source_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_believe_flaky_source_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* target */
int mxk_target_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_startup_target(struct list_loop *ls);

/* light */
struct task_loop *mxk_start_light(struct list_loop *ls, char *pattern);
int mxk_stop_light(struct list_loop *ls, char *pattern);

int mxk_list_pattern_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_load_pattern_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_start_light_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_stop_light_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_startup_light(struct list_loop *ls);
void mxk_shutdown_light(struct mxk_global *mg);

/* watch */
#include <stdarg.h>
int mxk_write_watch(struct list_loop *ls, struct task_loop *tl, int tag, char *str, ...);

int mxk_start_watch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_echo_watch_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_startup_watch(struct list_loop *ls);
void mxk_shutdown_watch(struct mxk_global *mg);

/* misc */

#define MXK_MISC_OPTIONAL 1
#define MXK_MISC_REQUIRED 0

char *mxk_skip_space(char *ptr);

#include <time.h>

void mxk_milli_to_tv(struct timeval *tv, unsigned int ms);

int mxk_parse_apply(struct list_loop *ls, struct task_loop *tl, char *args, char *name, int empty, int special, int (*fn)(struct list_loop *ls, struct task_loop *tx, int special, char *args));

struct input_event *mxk_get_key(struct list_loop *ls, struct link_loop *nk, int hit);

int mxk_parse_evt(char *args, struct input_event *evt);
int mxk_string_evt(char *buffer, int len, int type, int code, int value);
int mxk_print_evt(struct control_loop *cl, int cd, int type, int code, int value);

int mxk_has_bit(unsigned long *vector, unsigned int max, unsigned int test);

int mxk_echo_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_copying_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_version_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* switch */
int mxk_send_switch(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type, int code, int value);

/* find */
int mxk_show_find_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_linkup_find_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_enroll_find(struct list_loop *ls, char *name, int flags, int count, int (*insert)(void *server, int type, int id));
struct mxk_find *mxk_test_find(struct mxk_global *mg, char *name);
void mxk_end_find(struct mxk_global *mg);
#if 0
int mxk_type_find(struct mxk_global *mg, char *name); /* replaced with lookup_task_type_loop */
#endif

/* client */
int mxk_send_client(struct list_loop *ls, struct task_loop *tl, struct mxk_client *vector, int size, int type, int code, int value);
struct mxk_client *mxk_insert_client(struct mxk_client *vector, int size, int type, int id);
void mxk_shutdown_client(struct mxk_client *vector, int size);

/* queue */
struct mxk_queue *mxk_start_queue(int size);
void mxk_stop_queue(struct mxk_queue *mq);
void mxk_update_queue(struct list_loop *ls, struct mxk_queue *mq, int code, int value);
void mxk_consume_queue(struct mxk_queue *mq, int length);

/* chord */
int mxk_start_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_startup_chord(struct list_loop *ls);
void mxk_shutdown_chord(struct mxk_global *mg);
int mxk_old_order_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_new_order_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_gap_order_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_old_set_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_new_set_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_gap_set_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_bypass_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_ignore_chord_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* degap */
struct mxk_degap *mxk_start_degap();
void mxk_stop_degap(struct mxk_degap *md);
int mxk_lookup_degap(struct mxk_degap *md, int value);
int mxk_insert_degap(struct mxk_degap *md, int value);
char *mxk_reverse_degap(struct mxk_degap *md, int compact);
#ifdef DEBUG
#include <stdio.h>
int mxk_dump_degap(struct mxk_degap *md, FILE *fp);
#endif

/* key */
struct mxk_key *mxk_start_key();
void mxk_stop_key(struct mxk_key *mk);
int mxk_insert_key(struct mxk_key *mk, int *vector, int value, int priority, int depth);
int *mxk_parse_vector_key(char *args, int offset);

/* decode */
int mxk_list_events_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_codes_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_relative_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_absolute_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* bit */
int mxk_startup_bit(struct list_loop *ls);
void mxk_shutdown_bit(struct mxk_global *mg);

int mxk_start_up_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_start_down_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_start_sticky_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_start_pair_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_default_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_key_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_light_bit_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* array */
int mxk_startup_array(struct list_loop *ls);
void mxk_shutdown_array(struct mxk_global *mg);

int mxk_start_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_light_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_press_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_release_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_quick_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_any_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_preset_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_ignore_array_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_bypass_array_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* branch */
int mxk_startup_branch(struct list_loop *ls);
void mxk_shutdown_branch(struct mxk_global *mg);

int mxk_start_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_light_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_dark_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_up_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_down_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_key_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_default_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_branch_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* stat */

int mxk_startup_stat(struct list_loop *ls);
void mxk_shutdown_stat(struct mxk_global *mg);
int mxk_start_stat_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_stat_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* mode */

int mxk_startup_mode(struct list_loop *ls);
void mxk_shutdown_mode(struct mxk_global *mg);

int mxk_start_mode_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_setup_mode_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_state_mode_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_key_mode_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_mode_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* repeat */
int mxk_startup_repeat(struct list_loop *ls);
void mxk_shutdown_repeat(struct mxk_global *mg);

int mxk_list_repeat_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_start_repeat_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_adjust_repeat_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* single */

int mxk_startup_single(struct list_loop *ls);
void mxk_shutdown_single(struct mxk_global *mg);

int mxk_start_single_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_single_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_sustained_single_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_quick_single_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_timed_single_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_threshold_single_command(struct list_loop *ls, struct task_loop *tl, char *args);


/* axis */

void mxk_shutdown_ax(struct mxk_global *mg);
int mxk_startup_ax(struct list_loop *ls);

int mxk_start_ax_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_ax_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_point_ax_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_non_zero_ax_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_return_zero_ax_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* abs to relative */
int mxk_start_ar_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_brake_ar_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_ignore_ar_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_split_ar_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_pass_ar_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_ar_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_startup_ar(struct list_loop *ls);
void mxk_shutdown_ar(struct mxk_global *mg);

/* external */
int mxk_startup_external(struct list_loop *ls);
void mxk_shutdown_external(struct mxk_global *mg);

int mxk_start_external_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_external_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* sorter */
int mxk_startup_sorter(struct list_loop *ls);
void mxk_shutdown_sorter(struct mxk_global *mg);

int mxk_bin_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_match_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_start_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_reserve_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* smit */
int mxk_startup_smit(struct list_loop *ls);
void mxk_shutdown_smit(struct mxk_global *mg);

int mxk_start_smit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_smit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_key_smit_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_default_smit_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* brasc */
int mxk_startup_brasc(struct list_loop *ls);
void mxk_shutdown_brasc(struct mxk_global *mg);

int mxk_start_brasc_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_brasc_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* timster */

int mxk_startup_timster(struct list_loop *ls);
void mxk_shutdown_timster(struct mxk_global *mg);

int mxk_start_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_trigger_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_reset_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_clear_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_arm_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_off_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);

int mxk_auto_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_single_timster_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* guess */

int mxk_startup_guess(struct list_loop *ls);
void mxk_shutdown_guess(struct mxk_global *mg);

int mxk_start_guess_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_guess_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* genb */
int mxk_startup_genb(struct list_loop *ls);
void mxk_shutdown_genb(struct mxk_global *mg);

int mxk_start_genb_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_genb_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* farm */
int mxk_startup_farm(struct list_loop *ls);
void mxk_shutdown_farm(struct mxk_global *mg);

int mxk_start_farm_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_farm_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_propose_farm_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* reduce */

int mxk_startup_reduce(struct list_loop *ls);
void mxk_shutdown_reduce(struct mxk_global *mg);

int mxk_start_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_key_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_discard_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_relay_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_self_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_replace_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_merge_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_distinct_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args);

/* bug */
int mxk_startup_bug(struct list_loop *ls);
void mxk_shutdown_bug(struct mxk_global *mg);
int mxk_start_bug_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_list_bug_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_bypass_bug_command(struct list_loop *ls, struct task_loop *tl, char *args);
int mxk_ignore_bug_command(struct list_loop *ls, struct task_loop *tl, char *args);

#include <code-name.h>

extern struct code_name mxk_event_types[];
extern struct code_name mxk_led_codes[];
extern struct code_name mxk_key_codes[];
extern struct code_name mxk_rel_codes[];
extern struct code_name mxk_abs_codes[];
extern struct code_name mxk_key_values[];

#endif
