
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <code-name.h>

#include "mxk.h"

struct code_name plain_codes[] = {
  { "`", KEY_APOSTROPHE },
  { "-", KEY_MINUS },
  { "=", KEY_EQUAL },
  { "\\", KEY_BACKSLASH },

  { "[", KEY_LEFTBRACE },
  { "]", KEY_RIGHTBRACE },

  { ";", KEY_SEMICOLON },
  { "'", KEY_APOSTROPHE },

  { ",", KEY_COMMA },
  { ".", KEY_DOT },

  { "/", KEY_SLASH },

  { " ", KEY_SPACE },

  { "1", KEY_1 },
  { "2", KEY_2 },
  { "3", KEY_3 },
  { "4", KEY_4 },
  { "5", KEY_5 },
  { "6", KEY_6 },
  { "7", KEY_7 },
  { "8", KEY_8 },
  { "9", KEY_9 },
  { "0", KEY_0 },

  { "q", KEY_Q },
  { "w", KEY_W },
  { "e", KEY_E },
  { "r", KEY_R },
  { "t", KEY_T },
  { "y", KEY_Y },
  { "u", KEY_U },
  { "i", KEY_I },
  { "o", KEY_O },
  { "p", KEY_P },
  { "a", KEY_A },
  { "s", KEY_S },
  { "d", KEY_D },
  { "f", KEY_F },
  { "g", KEY_G },
  { "h", KEY_H },
  { "j", KEY_J },
  { "k", KEY_K },
  { "l", KEY_L },
  { "z", KEY_Z },
  { "x", KEY_X },
  { "c", KEY_C },
  { "v", KEY_V },
  { "b", KEY_B },
  { "n", KEY_N },
  { "m", KEY_M },

  { NULL, -1 }
};

struct code_name shift_codes[] = {
  { "~", KEY_APOSTROPHE },
  { "_", KEY_MINUS },
  { "+", KEY_EQUAL },
  { "|", KEY_BACKSLASH },

  { "{", KEY_LEFTBRACE },
  { "}", KEY_RIGHTBRACE },

  { ":", KEY_SEMICOLON },
  { "\"", KEY_APOSTROPHE },

  { "<", KEY_COMMA },
  { ">", KEY_DOT },

  { "?", KEY_SLASH },

  { " ", KEY_SPACE },

  { "!", KEY_1 },
  { "@", KEY_2 },
  { "#", KEY_3 },
  { "$", KEY_4 },
  { "%", KEY_5 },
  { "^", KEY_6 },
  { "&", KEY_7 },
  { "*", KEY_8 },
  { "(", KEY_9 },
  { ")", KEY_0 },

  { "Q", KEY_Q },
  { "W", KEY_W },
  { "E", KEY_E },
  { "R", KEY_R },
  { "T", KEY_T },
  { "Y", KEY_Y },
  { "U", KEY_U },
  { "I", KEY_I },
  { "O", KEY_O },
  { "P", KEY_P },
  { "A", KEY_A },
  { "S", KEY_S },
  { "D", KEY_D },
  { "F", KEY_F },
  { "G", KEY_G },
  { "H", KEY_H },
  { "J", KEY_J },
  { "K", KEY_K },
  { "L", KEY_L },
  { "Z", KEY_Z },
  { "X", KEY_X },
  { "C", KEY_C },
  { "V", KEY_V },
  { "B", KEY_B },
  { "N", KEY_N },
  { "M", KEY_M },

  { NULL, -1 }
};

int mxk_print_buffer(struct input_event **ve, int size, char *string, int length)
{
  unsigned int shift, mods, bit;
  int i, j, k, sofar, mark, got;
  struct input_event *vp;
  char *tmp;
  char number[16];
  char *vector[6];

  mods = shift = 0;

  sofar = 0;
  if(length <= 0){
    return -1;
  }

  bit = 0;

  for(i = 0; i < size; i++){
    vp = ve[i];
    got = 0;
    if(vp->type == EV_KEY){
      switch(vp->code){
#if 0
        case KEY_CAPSLOCK :
          switch(vp->value){
            case 0 : cps = 0; break;
            case 1 : cps = 1; break;
            /* TODO */
          }
          break;
#endif
        case KEY_LEFTSHIFT :
        case KEY_RIGHTSHIFT :
          bit = (vp->code == KEY_LEFTSHIFT) ? 0x1 : 0x2;
          switch(vp->value){
            case 0 : shift &= ~bit; break;
            case 1 : shift |=  bit; break;
          }
          break;
        case KEY_LEFTALT :
        case KEY_RIGHTALT :
        case KEY_LEFTCTRL :
        case KEY_RIGHTCTRL :
        case KEY_LEFTMETA :
        case KEY_RIGHTMETA :
          switch(vp->code){
            case KEY_LEFTALT    : bit = 0x1; break;
            case KEY_RIGHTALT   : bit = 0x2; break;
            case KEY_LEFTCTRL   : bit = 0x4; break;
            case KEY_RIGHTCTRL  : bit = 0x8; break;
            case KEY_LEFTMETA   : bit = 0x10; break;
            case KEY_RIGHTMETA  : bit = 0x20; break;
          }
          switch(vp->value){
            case 0 : mods &= ~bit; break;
            case 1 : mods |=  bit; break;
          }

          break;

        default :
          mark = mods;
          if(vp->value == 1){
            if(shift){
              tmp = code_to_name(shift_codes, vp->code);
            } else {
              tmp = code_to_name(plain_codes, vp->code);
            }
            if(tmp == NULL){
              mark = 1;
              tmp = code_to_name(mxk_key_codes, vp->code);
              if(tmp == NULL){
                sprintf(number, "%d", vp->code);
                tmp = number;
              }
            }
            if(mark){
              vector[got++] = "<";
              if(mods & 0xc){
                vector[got++] = "ctrl-";
              }
              if(mods & 0x3){
                vector[got++] = "alt-";
              }
              if(mods & 0x30){
                vector[got++] = "meta-";
              }
            }
            vector[got++] = tmp;
            if(mark){
              vector[got++] = ">";
            }
          }
          break;
      }
    }

    for(k = 0; k < got; k++){
      for(j = 0; (sofar < length) && (vector[k][j] != '\0'); j++, sofar++){
        string[sofar] = vector[k][j];
      }
      if((sofar + 1) >= length){
        return -1;
      }
    }
  }

  string[sofar] = '\0';

  return sofar;
}
