/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#define CBU_LOAD mxk_splate
#include "cbu.h"

#include "mxk.h"

/* save and emit module - a buffer of sorts */

struct mxk_splate{
  struct cbu_anchor p_anchor;
  unsigned int p_code;
#if 0
  unsigned int p_type;
  unsigned int p_value;
#endif
};
#define CBU_ANCHOR p_anchor
#include "cbu.c"

struct mxk_smentry{
  unsigned int e_cmd;
  unsigned int e_number;
};

struct mxk_smit{
  struct mxk_smentry **s_table;
  unsigned int s_size;
  unsigned int s_base;
  struct mxk_smentry s_default;

  struct cbu_state *s_cbu;

  unsigned int s_countdown;
  unsigned int s_cmd;

  int s_press;

  struct mxk_client *s_vector;
  int s_count;
};

static void mxk_destroy_smit(struct mxk_smit *smit);
static struct mxk_smit *mxk_create_smit(int depth);
static int mxk_run_smit(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************************************************************/

#define SMIT_MASK_CMD     0x00ff
#define SMIT_CMD_NULL     0x0000  /* nothing */
#define SMIT_CMD_QUEUE    0x0001  /* fifo push */
#define SMIT_CMD_PUSH     0x0002  /* stack push */
#define SMIT_CMD_POP      0x0003  /* fifo/stack pop */
#define SMIT_CMD_TOP      0x0004  /* fifo/stack top */

#define SMIT_MASK_FLAG    0xff00
#define SMIT_FLAG_CLOBBER 0x0100  /* force discards if space limited */
#define SMIT_FLAG_TIMER   0x0200  /* do things later */
#define SMIT_FLAG_RETIMER 0x0400  /* do things later, even if something else planned */

static struct code_name smit_table[] = {
  { "queue", SMIT_CMD_QUEUE },
  { "push",  SMIT_CMD_PUSH },
  { "pop",   SMIT_CMD_POP },
  { "top",   SMIT_CMD_TOP },

  { "clobber",  SMIT_FLAG_CLOBBER },
  { "timer",    SMIT_FLAG_TIMER },
  { "retimer",  SMIT_FLAG_RETIMER },

  {  NULL,   0 }
};

/**********************************************************************/

struct mxk_smit *mxk_create_smit(int depth)
{
  struct mxk_smit *smit;

  smit = malloc(sizeof(struct mxk_smit));
  if(smit == NULL){
    return NULL;
  }

  smit->s_vector = NULL;
  smit->s_count = 0;

  smit->s_table = NULL;
  smit->s_size = 0;
  smit->s_base = UINT_MAX;

  smit->s_cbu = NULL;

  smit->s_countdown = 0;
  smit->s_cmd = SMIT_CMD_NULL;

  smit->s_press = (-1);

  smit->s_cbu = create_cbu(depth);
  if(smit->s_cbu == NULL){
    mxk_destroy_smit(smit);
    return NULL;
  }

  /* TODO - handle default */

  return smit;
}

static void mxk_destroy_smit(struct mxk_smit *smit)
{
  int i;

  if(smit == NULL){
    return;
  }

  if(smit->s_table){
    for(i = 0; i < smit->s_size; i++){
      if(smit->s_table[i]){
        free(smit->s_table[i]);
        smit->s_table[i] = NULL;
      }
    }
    free(smit->s_table);
    smit->s_table = NULL;
    smit->s_size = 0;
    smit->s_base = UINT_MAX;
  }

  destroy_cbu(smit->s_cbu);
  smit->s_cbu = NULL;

  mxk_shutdown_client(smit->s_vector, smit->s_count);

#ifdef DEBUG
  smit->s_vector = NULL;
  smit->s_count = 0;
#endif

  free(smit);
}

struct mxk_smentry *mxk_find_smit(struct mxk_smit *smit, unsigned int key)
{
  if(key < smit->s_base){ /* handles null case if base is maxint */
    return NULL;
  }

  if(key >= (smit->s_base + smit->s_size)){
    return NULL;
  }

  return smit->s_table[key - smit->s_base];
}

int mxk_add_key_smit(struct mxk_smit *smit, int special, unsigned int key, unsigned int flags, unsigned int number)
{
  struct mxk_smentry *se, **tmp;
  int grow, i;

  if(special){
    se = &(smit->s_default);
  } else {
    se = mxk_find_smit(smit, key);
  }
  if(se){
#ifdef DEBUG
fprintf(stderr, "smit: modification to existing instance for key 0x%x\n", key);
#endif
    se->e_cmd = flags;
    se->e_number = number;
    return 0;
  }

  se = malloc(sizeof(struct mxk_smentry));
  if(se == NULL){
    return -1;
  }
  se->e_cmd = flags;
  se->e_number = number;

  if(smit->s_table == NULL){ /* base case */
    smit->s_table = malloc(sizeof(struct smit_smentry *));
    if(smit->s_table == NULL){
      free(se);
      return -1;
    }

    smit->s_table[0] = se;

    smit->s_size = 1;
    smit->s_base = key;
#ifdef DEBUG
    fprintf(stderr, "smit: initial insert: key %d\n", key);
#endif
    return 0;
  }

  if(key < smit->s_base){
    grow = smit->s_base - key;
#ifdef DEBUG
    fprintf(stderr, "smit: insert at start: key=%d, base=%d, grow=%d\n", key, smit->s_base, grow);
#endif
  } else {
    grow = 1 + key - (smit->s_base + smit->s_size);
#ifdef DEBUG
    fprintf(stderr, "smit: insert at end: key=%d, base=%d, grow=%d\n", key, smit->s_base, grow);
#endif
  }

  tmp = realloc(smit->s_table, sizeof(struct smit_smentry *) * (grow + smit->s_size));
  if(tmp == NULL){
    free(se);
    return -1;
  }
  smit->s_table = tmp;

  if(key < smit->s_base){
    memmove(smit->s_table + grow, smit->s_table, smit->s_size * sizeof(struct smit_smentry *));

    smit->s_table[0] = se;
    for(i = 1; i < grow; i++){
      smit->s_table[i] = NULL;
    }
    smit->s_base = key;
  } else {
    smit->s_table[smit->s_size + grow - 1] = se;
  }

  smit->s_size += grow;

  return 0;
}

int mxk_parse_key_smit(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  char buffer[MXK_NAME_SIZE];
  int idx;
  unsigned int key, flags, flag, number;
  struct mxk_smit *smit;
  char *end;

  smit = user_loop(ls, tx);

#ifdef DEBUG
  fprintf(stderr, "smit: about to parse key entry\n");
#endif

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  key = (-1);
  if(special == 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key < 0){
      return -1;
    }
  }

  flags = 0;
  number = 0;

  while((idx = next_word(args, idx)) > 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
#ifdef DEBUG
      fprintf(stderr, "smit: load word failed\n");
#endif
      return -1;
    }
    flag = name_to_code(smit_table, buffer);
#ifdef DEBUG
    fprintf(stderr, "smit: flag (%s)->0x%x\n", buffer, flag);
#endif
    if(flag > 0){
      if(flag & SMIT_MASK_CMD){
        flags = (flags & SMIT_MASK_FLAG) | flag; /* clobber commands */
      } else {
        flags |= flag;
      }
    } else {
      number = strtoul(buffer, &end, 0);
      if((*end != '\0') && (*end != ' ')){
#ifdef DEBUG
        fprintf(stderr, "smit: bad keyword (%s)\n", buffer);
#endif
        return -1;
      }
    }
  }

  if(mxk_add_key_smit(smit, special, key, flags, number)){
    return -1;
  }

  return 0;
}

/*********************************************************************/

int mxk_key_smit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "smit", MXK_MISC_REQUIRED, 0, mxk_parse_key_smit);

  return 0;
}

int mxk_default_smit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "smit", MXK_MISC_REQUIRED, 1, mxk_parse_key_smit);

  return 0;
}

int mxk_start_smit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_smit *smit;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type, depth;
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "smit");
  if(type <= 0){
    print_control(cl, cd, 1, "smit: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "smit: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "smit: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
  if((*end != ' ') && (*end != '\0')){
    print_control(cl, cd, 1, "smit: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    depth = 1;
  } else {
    depth = strtoul(args + idx, &end, 10);
    if((*end != ' ') && (*end != '\0')){
      print_control(cl, cd, 1, "smit: bad depth\n");
      return 0;
    }
    if(depth == 0){
      print_control(cl, cd, 1, "smit: depth has to be greater than zero\n");
      return 0;
    }
  }

  smit = mxk_create_smit(depth);
  if(smit == NULL){
    print_control(cl, cd, 1, "smit: unable to allocate state\n");
    return 0;
  }

  tx = add_tag_loop(ls, -1, smit, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_smit, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "smit: unable to create task\n");
    mxk_destroy_smit(smit);
    return 0;
  }

  text_loop(ls, tx, "smit:%d depth=%d", id, depth);
  print_control(cl, cd, 1, "smit: created task\n");

  return 0;
}

static int mxk_print_flags_smentry(struct control_loop *cl, int cd, struct mxk_smentry *se)
{
  int j;

  for(j = 0; smit_table[j].name; j++){
    if((smit_table[j].code & se->e_cmd) == smit_table[j].code){
      print_control(cl, cd, 0, " %s", smit_table[j].name);
    }
  }
  print_control(cl, cd, 0, " (0x%04x)\n", se->e_cmd);

  return 0;
}

int mxk_list_smit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int i, cd, idx, id, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_smit *smit;
  struct mxk_smentry *se;
  char *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "smit");
  if(type <= 0){
    print_control(cl, cd, 1, "smit: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    smit = user_loop(ls, tx);

    if(smit){
      print_control(cl, cd, 0, "smit[%d]: base=%u, size=%u\n", tag_loop(ls, tx), smit->s_base, smit->s_size);
      for(i = 0; i < smit->s_size; i++){
        se = smit->s_table[i];
        if(se){
          ptr = code_to_name(mxk_key_codes, smit->s_base + i);
          if(ptr){
            print_control(cl, cd, 0, "%s:", ptr);
          } else {
            print_control(cl, cd, 0, "0x%x:", smit->s_base + i);
          }
          mxk_print_flags_smentry(cl, cd, se);
        }
      }
      se = &(smit->s_default);
      print_control(cl, cd, 0, "default: ");
      mxk_print_flags_smentry(cl, cd, se);
      count++;
    } else {
      print_control(cl, cd, 1, "smit: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

/**********************************************************************/

static int mxk_run_smit(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_smit *smit;
  struct mxk_smentry *se;
  struct mxk_splate *sp;
  struct input_event *vp;
  struct link_loop *nk;
  unsigned int cmd;
  int discard, clobber;

  mg = global_loop(ls);
  smit = user_loop(ls, tl);

  if(smit == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_smit(smit);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){


      if(nk->k_type != mg->g_event){
        discard_link_loop(ls, nk);
        continue;
      }

      vp = nk->k_data;
      if((vp->type != EV_KEY) || (vp->value > 1)){
#ifdef DEBUG
        if(vp->type != EV_KEY){
          fprintf(stderr, "smit: ignoring nonkey event\n");
        } else {
          fprintf(stderr, "smit: ignoring key repeat for 0x%x (value %d)\n", vp->code, vp->value);
        }
#endif
        discard_link_loop(ls, nk);
        continue;
      }

      if(vp->value == 0){
#ifdef DEBUG
        fprintf(stderr, "smit: contemplating key release (save=0x%x, got=0x%x)\n", smit->s_press, vp->code);
#endif
        if((smit->s_press >= 0) && (smit->s_press == vp->code)){ /* release whatever we saved */
          mxk_send_client(ls, tl, smit->s_vector, smit->s_count, EV_KEY, smit->s_press, 0);
          mxk_send_client(ls, tl, smit->s_vector, smit->s_count, EV_SYN, 0, 0);
          smit->s_press = (-1);
        }
        discard_link_loop(ls, nk);
        continue;
      }

      /* at this point vp has to be a key/press */
#ifdef DEBUG
      fprintf(stderr, "smit: got keypress for 0x%x\n", vp->code);
#endif

      cmd = SMIT_CMD_NULL;
      if(smit->s_countdown){
        smit->s_countdown--;
        if(smit->s_countdown == 0){
          cmd = smit->s_cmd;
          smit->s_cmd = SMIT_CMD_NULL;
        }
      }

      if(cmd != SMIT_CMD_NULL){
#ifdef DEBUG
        fprintf(stderr, "smit: timer command 0x%04x triggered\n", cmd);
#endif
        discard = 0; /* a replay command, not sure if discard needed */
      } else {
        discard = 1;
        se = mxk_find_smit(smit, vp->code);
        if(se == NULL){
          se = &(smit->s_default);
        }
        if((se->e_cmd) & SMIT_FLAG_TIMER){
#ifdef DEBUG
          fprintf(stderr, "smit: scheduling timer: cmd 0x%04x\n", se->e_cmd);
#endif
          if((smit->s_countdown == 0) || (se->e_cmd & SMIT_FLAG_RETIMER)){
            smit->s_cmd = se->e_cmd;
            smit->s_countdown = se->e_number;
            if(smit->s_countdown <= 0){
              smit->s_countdown = 1;
            }
          }
          discard_link_loop(ls, nk);
          continue;
        }
        cmd = se->e_cmd;
      }


      clobber = (cmd & SMIT_FLAG_CLOBBER) ? 1 : 0;
      sp = NULL;

#ifdef DEBUG
      fprintf(stderr, "smit: processing command 0x%04x, discard=%d, clobber=%d\n", cmd, discard, clobber);
#endif

      switch(cmd & SMIT_MASK_CMD){
        case SMIT_CMD_QUEUE :
          sp = queue_cbu(smit->s_cbu, clobber);
          break;
        case SMIT_CMD_PUSH  :
          sp = push_cbu(smit->s_cbu, clobber);
          break;
        case SMIT_CMD_POP   :
          sp = pop_cbu(smit->s_cbu);
          break;
        case SMIT_CMD_TOP   :
          sp = top_cbu(smit->s_cbu);
          break;
      }

      switch(cmd & SMIT_MASK_CMD){
        case SMIT_CMD_QUEUE :
        case SMIT_CMD_PUSH  :
          discard = 1;
          if(sp){
#ifdef DEBUG
            fprintf(stderr, "smit: saving code 0x%x\n", vp->code);
#endif
            sp->p_code = vp->code;
          }
          break;
        case SMIT_CMD_POP   :
        case SMIT_CMD_TOP   :
          if(sp){
            if(smit->s_press >= 0){
              mxk_send_client(ls, tl, smit->s_vector, smit->s_count, EV_KEY, smit->s_press, 0);
              mxk_send_client(ls, tl, smit->s_vector, smit->s_count, EV_SYN, 0, 0);
#if 0
              smit->s_press = (-1);
#endif
            }

#ifdef DEBUG
            fprintf(stderr, "smit: replaying code 0x%x\n", sp->p_code);
#endif

            mxk_send_client(ls, tl, smit->s_vector, smit->s_count, EV_KEY, sp->p_code, 1);
            mxk_send_client(ls, tl, smit->s_vector, smit->s_count, EV_SYN, 0, 0);
            smit->s_press = sp->p_code;
          }
          break;
      }

      if(discard){
#ifdef DEBUG
        fprintf(stderr, "smit: discarding code=0x%x, value=%d\n", vp->code, vp->value);
#endif
        discard_link_loop(ls, nk);
      }
    }
  }

  return 0;
}

/**********************************************************************/

static int mxk_insert_smit(void *state, int type, int id)
{
  struct mxk_smit *smit;
  struct mxk_client *mc;

  smit = state;

  mc = mxk_insert_client(smit->s_vector, smit->s_count, type, id);
  if(mc == NULL){
    return -1;
  }

  smit->s_vector = mc;
  smit->s_count++;

  return 0;
}

int mxk_startup_smit(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "smit", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_smit);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_smit(struct mxk_global *mg)
{
}
