/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

struct mxk_datum{
  int d_total;
  int d_chord;
};

struct mxk_stat{
  int s_total;
  int s_error;
  int s_other;
  int s_track;

  int s_size;
  struct mxk_datum *s_table;

  struct mxk_client *s_vector;
  int s_count;
};

static int mxk_start_stat(struct list_loop *ls, int id);
static void mxk_stop_stat(struct mxk_stat *mb);
static int mxk_run_stat(struct list_loop *ls, struct task_loop *tl, int mask);

int mxk_start_stat_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id;
  struct control_loop *cl;
  char *value;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
#ifdef DEBUG
      fprintf(stderr, "stat: id is %d\n", id);
#endif
    }
  }

  if(mxk_start_stat(ls, id)){
    print_control(cl, cd, 1, "stat: unable to start task for id %d\n", id);
  } else {
    print_control(cl, cd, 1, "stat: created id %d\n", id);
  }

  return 0;
}

int mxk_list_stat_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_stat *ms;
  struct mxk_datum *md;
  char *value, *key;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "stat");
  if(type <= 0){
    print_control(cl, cd, 1, "stat: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    ms = user_loop(ls, tx);
    if(ms){
      print_control(cl, cd, 0, "stat[%d]: total=%d, errors=%d, other=%d, track=%d\n", tag_loop(ls, tx), ms->s_total, ms->s_error, ms->s_other, ms->s_track);
      for(i = 0; i < ms->s_size; i++){
        md = &(ms->s_table[i]);
        if(md->d_total){
          key = code_to_name(mxk_key_codes, i);
          if(key){
            print_control(cl, cd, 0, "key %s:", key);
          } else {
            print_control(cl, cd, 0, "key 0x%x:", i);
          }
          print_control(cl, cd, 0, " %d/%d\n", md->d_chord, md->d_total);
        }
      }
    } else {
      print_control(cl, cd, 1, "stat:%d unavailable\n", id);
    }
    count++;
  }

  print_control(cl, cd, 1, "%d collection %s\n", count, (count == 1) ? "node" : "nodes");

  return 0;
}

/***************************************************************/

static int mxk_start_stat(struct list_loop *ls, int id)
{
  struct mxk_stat *ms;
  struct task_loop *tx;
  int type;

  type = lookup_task_type_loop(ls, "stat");
  if(type <= 0){
#ifdef DEBUG
    fprintf(stderr, "stat: type not defined\n");
#endif
    return -1;
  }

  ms = malloc(sizeof(struct mxk_stat));
  if(ms == NULL){
    return -1;
  }

  ms->s_total = 0;
  ms->s_error = 0;
  ms->s_other = 0;
  ms->s_track = 0;

  ms->s_size = 0;
  ms->s_table = NULL;

  ms->s_vector = NULL;
  ms->s_count = 0;

  tx = add_tag_loop(ls, -1, ms, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_stat, NULL, type, id);
  if(tx == NULL){
#ifdef DEBUG
    fprintf(stderr, "stat: task creation failed\n");
#endif
    mxk_stop_stat(ms);
    return -1;
  }

  text_loop(ls, tx, "stat:%d", id);

  return 0;
}

static void mxk_stop_stat(struct mxk_stat *ms)
{
  if(ms){

    mxk_shutdown_client(ms->s_vector, ms->s_count);

    if(ms->s_table){
      free(ms->s_table);
      ms->s_table = NULL;
    }
    ms->s_size = 0;

    free(ms);
  }
}

static int mxk_run_stat(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event *vp;
  struct mxk_stat *ms;
  struct mxk_global *mg;
  struct mxk_datum *md;
  struct link_loop *nk;
  int i;

  ms = user_loop(ls, tl);
  mg = global_loop(ls);

  if(mask & LO_STOP_MASK){
    mxk_stop_stat(ms);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_KEY :
            if(ms->s_size <= vp->code){
              md = realloc(ms->s_table, sizeof(struct mxk_datum) * (vp->code + 1));
              if(md == NULL){
                ms->s_error++;
                return 0;
              }
              for(i = ms->s_size; i <= vp->code; i++){
                md[i].d_total = 0;
                md[i].d_chord = 0;
              }
              ms->s_table = md;
              ms->s_size = vp->code + 1;
            }

            switch(vp->value){
              case 0 :
                ms->s_track--;
                if(ms->s_track < 0){
                  ms->s_track = 0;
                }
                md = &(ms->s_table[vp->code]);
                if(ms->s_track){
                  md->d_chord++;
                }
                md->d_total++;
                ms->s_total++;
#ifdef DEBUG
                fprintf(stderr, "stat[%d]: %d/%d (track=%d)\n", vp->code, md->d_chord, md->d_total, ms->s_track);
#endif
              break;
              case 1 :
                ms->s_track++;
              break;
            }

          break;
          default :
            ms->s_other++;
          break;
        }
        mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, vp->value);
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_stat(void *stat, int type, int id)
{
  struct mxk_stat *ms;
  struct mxk_client *mc;

  ms = stat;

  mc = mxk_insert_client(ms->s_vector, ms->s_count, type, id);
  if(mc == NULL){
    return -1;
  }

  ms->s_vector = mc;
  ms->s_count++;

  return 0;
}

int mxk_startup_stat(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "stat", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_stat);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_stat(struct mxk_global *mg)
{
}
