/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

#define OUTPUT_BUFFER 128

/* bit flaky: uses both a global message buffer with anon wake event
 * and message events ...
 */

int mxk_run_watch(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_watch *mw;
  struct link_loop *nk;
  struct input_event *vt;
  int fd, wr, len;
  char buffer[OUTPUT_BUFFER];
  int run;

  mg = global_loop(ls);
  fd = fd_loop(ls, tl);
  mw = user_loop(ls, tl);

#ifdef DEBUG
  fprintf(stderr, "watch: running\n");
#endif

  if(mask & LO_STOP_MASK){
    free(mw);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    for(run = 1; run;){
      nk = receive_link_loop(ls, tl);
      if(nk){
        if(nk->k_type == mg->g_event){
          vt = nk->k_data;

          len = mxk_string_evt(buffer, OUTPUT_BUFFER - 1, vt->type, vt->code, vt->value);
          if((len < 0) || (len >= (OUTPUT_BUFFER - 1))){
            len = OUTPUT_BUFFER - 1;
          }
          buffer[len++] = '\n';
          wr = write(fd, buffer, len);
          if(wr < len){
            if(wr < 0){
              switch(errno){
                case EAGAIN :
                case EINTR :
                  break;
                default :
                  stop_loop(ls, tl);
                  break;
              }
            }
            /* enable write mask, bomb out */
            run = 0;
          }
        } /* else if ... */
        discard_link_loop(ls, nk);
      } else {
        run = 0;
      }
    }
  }

  /* MAYBE: process READ_MASK, simply to prevent read buffers filling ? */

  if(mg->g_nonce != mw->w_nonce){
    mw->w_nonce = mg->g_nonce;
    mw->w_sofar = 0;
  }

  if(mw->w_sofar < mg->g_length){
    wr = write(fd, mg->g_message + mw->w_sofar, mg->g_length - mw->w_sofar);
    if(wr < 0){
      switch(errno){
        case EAGAIN :
        case EINTR  :
          set_mask_loop(ls, tl, LO_WRITE_MASK, 1);
          break;
        default :
          stop_loop(ls, tl);
          break;
      }
    } else {
      mw->w_sofar += wr;
      if(mw->w_sofar < mg->g_length){
        set_mask_loop(ls, tl, LO_WRITE_MASK, 1);
      } else {
        set_mask_loop(ls, tl, LO_WRITE_MASK, -1);
      }
    }
  }

  return 0;
}

int mxk_write_watch(struct list_loop *ls, struct task_loop *tl, int tag, char *str, ...)
{
  va_list ap;
  struct mxk_global *mg;
  struct task_loop *tx;
  int len, type;

  mg = global_loop(ls);
  type = lookup_task_type_loop(ls, "watch");

  mg->g_message[0] = '\r';
  mg->g_message[1] = '\n';

  va_start(ap, str);

  len = vsnprintf(mg->g_message + 2, MXK_WATCH_SIZE - 2, str, ap);
  mg->g_message[MXK_WATCH_SIZE - 1] = '\0';

  va_end(ap);

  if((len < 0) || (len >= MXK_WATCH_SIZE - 2)){
    mg->g_length = MXK_WATCH_SIZE - 1;
  } else {
    mg->g_length = len + 2;
  }
  mg->g_nonce++;

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, tag); tx; tx = find_task_tag_loop(ls, NULL, tx, type, tag)){
    send_link_loop(ls, tl, tx, 0);
  }

  return 0;
}

int mxk_start_watch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct mxk_watch *mw;
  struct mxk_global *mg;
  struct control_loop *cl;
  struct task_loop *tx;
  int fd, id, idx, type;
  char *label;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);
  mg = global_loop(ls);
  type = lookup_task_type_loop(ls, "watch");

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      label = args + idx;
      id = strtoul(label, NULL, 10);
    }
  }

  mw = malloc(sizeof(struct mxk_watch));
  if(mw == NULL){
    print_control(cl, fd, 1, "unable to allocate watch\n");
    return 0;
  }

  mw->w_nonce = 0;
  mw->w_sofar = 0;

  tx = add_tag_loop(ls, fd, mw, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_watch, NULL, type, id);

  if(tx == NULL){
    free(mw);
    print_control(cl, fd, 1, "unable to start watch task\n");
    return 0;
  }

#ifdef DEBUG
  fprintf(stderr, "watch: stole fd from control\n");
#endif

  text_loop(ls, tx, "watch:%d", id);

  set_mask_loop(ls, tl, LO_WRITE_MASK | LO_READ_MASK, -1);
  set_fd_loop(ls, tl, -1);
  stop_loop(ls, tl);

  return -1; /* WARNING: special: causes an immediate return */
}

int mxk_echo_watch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_write_watch(ls, tl, 0, "%s", args);

  return 0;
}

/*********************************************************************/

int mxk_startup_watch(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "watch", MXK_EVT_TARGET, 4, NULL);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_watch(struct mxk_global *mg)
{
}
